/******************************************************************************
* 
* Copyright (c) 2010 Freescale Semiconductor;
* All Rights Reserved                       
*
*******************************************************************************
*
* THIS SOFTWARE IS PROVIDED BY FREESCALE "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL FREESCALE OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
***************************************************************************//*!
*
* @file      d4d_base.c
*
* @author    r50233
* 
* @version   0.0.112.0
* 
* @date      Aug-9-2012
* 
* @brief     D4D driver core and base function c file 
*
******************************************************************************/

#include "d4d.h"
#include "common_files/d4d_private.h"
#include "common_files/d4d_lldapi.h"

/**************************************************************//*!
*
* Global variables
*
******************************************************************/

Byte d4d_scratchPad[D4D_SCRATCHPAD_SIZE];

/**************************************************************//*!
*
* Local variables
*
******************************************************************/

// screen activation hisotry array & index
D4D_SCREEN* d4d_screenHistory[D4D_SCREEN_HISTORY];
Byte d4d_screenHistoryIndex;

// if not NULL, keys are captured by this object
D4D_OBJECT* d4d_pKeysCapturer = NULL;

// one and only message object being used to route information to objects
D4D_MESSAGE d4d_msg;

const D4D_SIZE d4d_size_zero = {0, 0}; // zero size structure for automatic function capability
const D4D_POINT d4d_point_zero = {0, 0}; // zero point structure

// last point remembered for MoveTo, LineTo etc.
static D4D_POINT d4d_curXY;

// The buffer of input keys
static D4D_KEYS_BUFFER d4d_KeysBuff;

// The D4D system flags
D4D_SYSTEM_FLAGS d4d_systemFlags;

#ifdef D4D_LLD_TCH
  static D4D_TOUCHSCREEN_STATUS d4d_TouchScreen_Status;
  D4D_OBJECT* d4d_LastTouchedObj;
  static D4D_POINT d4d_TouchScreenCoor = {0, 0};  
  static D4D_TOUCHSCREEN_EVENT d4d_touchEvent = { 0, {0, 0}};
#endif


#ifdef __CWCC__
  #pragma mark -
  #pragma mark D4D System Functions
  #pragma mark -
#endif

/**************************************************************//*!
*
* Initialize
*
******************************************************************/

D4D_BOOL D4D_Init(D4D_SCREEN* pInitScreen)
{
  if(!D4D_LCD_Init()) // Init hardware of LCD
    return D4D_FALSE;
  
  d4d_screenHistoryIndex = 0;
  
  #ifdef D4D_LLD_TCH
    d4d_TouchScreen_Status = 0;
    d4d_LastTouchedObj = NULL;    
  #endif
  
    
  // In case that the external font is enabled, just initialize it
  #if (D4D_EXTSRC_FILE_ENABLE != D4D_FALSE) && (D4D_FNT_EXTSRC_SUPPORT != D4D_FALSE)
    // one time initialization
    if(D4D_ExtFntInit() == D4D_FALSE)
      return D4D_FALSE;
  #endif
    
  d4d_systemFlags = 0;
  
  D4D_ClearKeysBuffer();
  if(pInitScreen)
    D4D_ActivateScreen(pInitScreen, D4D_FALSE);
    
  return D4D_TRUE;  
}

/**************************************************************//*!
*
* The main poll call
*
******************************************************************/

void D4D_Poll(void)
{
    // get active screen
    D4D_SCREEN* pScreen;
    
    // handle keys (may change active screen)
    D4D_HandleKeys();
   
  #ifdef D4D_LLD_TCH
    // handle Touch screen events - if any
    if(d4d_TouchScreen_Status & D4D_TCHSCR_STATUS_EVENTS)
    {
      D4D_ManageTouchScreenEvents();
      d4d_TouchScreen_Status &= ~D4D_TCHSCR_STATUS_EVENTS;
    }
    
    pScreen = D4D_GetActiveScreen();
    if(pScreen == NULL)
      return;
    // Handle Touch screen to screen and objects
    D4D_HandleTouchScreen(pScreen);
  #endif
  
    pScreen = D4D_GetActiveScreen();
    if(pScreen == NULL)
      return;
    // call screen's main function
    if(pScreen->OnMain)
        pScreen->OnMain();    
        
    // Handle the timeticks of D4D
    pScreen = D4D_GetActiveScreen();    
    D4D_HandleTimeTicks(pScreen);
    
    // redraw all objects on active screen
    pScreen = D4D_GetActiveScreen();
    D4D_RedrawScreen(pScreen);
}

/**************************************************************//*!
*
* Clear all rest information about pushed Keys in buffer
*
******************************************************************/
void D4D_ClearKeysBuffer(void)
{
  d4d_KeysBuff.readPos = 0;
  d4d_KeysBuff.writePos = 0;    
}

/**************************************************************//*!
*
* The function sets the orientation of display
*
******************************************************************/
void D4D_SetOrientation(D4D_ORIENTATION orient)
{
  // get active screen
  D4D_SCREEN* pScreen = D4D_GetActiveScreen();
    
  D4D_LCD_SetOrientation(orient);
  
  if(pScreen)
    D4D_InvalidateScreen(pScreen, D4D_TRUE); 
}


/**************************************************************//*!
*
* Put actual information about keys
*
******************************************************************/

void D4D_TimeTickPut(void)
{
  d4d_systemFlags |= D4D_SYSTEM_F_TIMETICK;  
}


/**************************************************************//*!
*
* Put actual information about keys
*
******************************************************************/

void D4D_KeysChanged(D4D_KEYS keys)
{
  static D4D_KEYS keysLastState = 0;
  
  
  
 if(keys != keysLastState)
 {
    D4D_KEYS kChange;
    D4D_KEY_SCANCODE tmp_scanCode;
    
    kChange = (Byte)(keys ^ keysLastState); 
    
    if(kChange & D4D_KEY_UP)
    {
      tmp_scanCode = D4D_KEY_SCANCODE_UP;
      if((keys & D4D_KEY_UP) == 0) // the key is released
        tmp_scanCode |= D4D_KEY_SCANCODE_RELEASEMASK;
      
      D4D_NewKeyEvent(tmp_scanCode);      
    }
    
    if(kChange & D4D_KEY_DOWN)
    {
      tmp_scanCode = D4D_KEY_SCANCODE_DOWN;
      if((keys & D4D_KEY_DOWN) == 0) // the key is released
        tmp_scanCode |= D4D_KEY_SCANCODE_RELEASEMASK;
      
      D4D_NewKeyEvent(tmp_scanCode);      
    }
    
    if(kChange & D4D_KEY_LEFT)
    {
      tmp_scanCode = D4D_KEY_SCANCODE_LEFT;
      if((keys & D4D_KEY_LEFT) == 0) // the key is released
        tmp_scanCode |= D4D_KEY_SCANCODE_RELEASEMASK;
      
      D4D_NewKeyEvent(tmp_scanCode);      
    }
    
    if(kChange & D4D_KEY_RIGHT)
    {
      tmp_scanCode = D4D_KEY_SCANCODE_RIGHT;
      if((keys & D4D_KEY_RIGHT) == 0) // the key is released
        tmp_scanCode |= D4D_KEY_SCANCODE_RELEASEMASK;
      
      D4D_NewKeyEvent(tmp_scanCode);      
    }
    
    if(kChange & D4D_KEY_ENTER)
    {
      tmp_scanCode = D4D_KEY_SCANCODE_ENTER;
      if((keys & D4D_KEY_ENTER) == 0) // the key is released
        tmp_scanCode |= D4D_KEY_SCANCODE_RELEASEMASK;
      
      D4D_NewKeyEvent(tmp_scanCode);      
    }
    
    if(kChange & D4D_KEY_ESC)
    {
      tmp_scanCode = D4D_KEY_SCANCODE_ESC;
      if((keys & D4D_KEY_ESC) == 0) // the key is released
        tmp_scanCode |= D4D_KEY_SCANCODE_RELEASEMASK;
      
      D4D_NewKeyEvent(tmp_scanCode);      
    }
    
    keysLastState = keys;           
 }   
}

/**************************************************************//*!
*
* Put new event from keys
*
******************************************************************/

void D4D_NewKeyEvent(D4D_KEY_SCANCODE scanCode)
{

  // Is there place for new record?
  if((d4d_KeysBuff.writePos + 1 == d4d_KeysBuff.readPos) || ((d4d_KeysBuff.readPos == 0) && (d4d_KeysBuff.writePos + 1 == D4D_KEYS_BUFF_LENGTH))) 
  {
    return;
  }
  
  // place new record to buffer
  d4d_KeysBuff.buff[d4d_KeysBuff.writePos++] = scanCode;
  if(d4d_KeysBuff.writePos >= D4D_KEYS_BUFF_LENGTH)
    d4d_KeysBuff.writePos = 0;  
  
  
  
  #ifdef D4D_INPUT_EVENT_CALLBACK
    D4D_INPUT_EVENT_CALLBACK();
  #endif
  
}


/**************************************************************//*!
*
* helper to invoke OBJECT's message handler 
*
* all information need to be already prepared in d4d_msg
*
******************************************************************/

void D4D_SendMessage(D4D_MESSAGE* pMsg)
{
    D4D_OBJECT* pObject = pMsg->pObject;
    
    // screen gets the first chance
    if(pMsg->pScreen->OnObjectMsg)
    {
        // screen may discard the message by returning TRUE
        if(pMsg->pScreen->OnObjectMsg(pMsg))
            return ;
    }
    
    if(pObject)
    {      
      // now invoke the object's handler        
      if(pObject->OnUsrMessage)
      {
        if(pObject->OnUsrMessage(pMsg))
          return;   
      }
      
      if(pObject->pObjFunc->OnSysMessage)
          pObject->pObjFunc->OnSysMessage(pMsg); 
    }
}

/**************************************************************//*!
*
* helper to invoke OBJECT's message handler 
*
* all information need to be already prepared in d4d_msg
*
******************************************************************/

void D4D_SendMessageMask(D4D_MESSAGE* pMsg, D4D_OBJECT_FLAGS parentFlagsMask, D4D_OBJECT_FLAGS endFlagMask)
{
    D4D_OBJECT* pObject = pMsg->pObject;
    
    // Check if the object 
    if((pObject->pData->flags & parentFlagsMask) != parentFlagsMask)
      return;
    
    // screen gets the first chance
    if((pObject->pData->flags & endFlagMask) == endFlagMask)
      if(pMsg->pScreen->OnObjectMsg)
      {
          // screen may discard the message by returning TRUE
          if(pMsg->pScreen->OnObjectMsg(pMsg))
              return ;
      }
    
    if(pObject)
    {      
      // now invoke the object's handler        
      if((pObject->pData->flags & endFlagMask) == endFlagMask)
      {
        if(pObject->OnUsrMessage)
        {
          if(pObject->OnUsrMessage(pMsg))
            return;   
        }
        
        // Invoke to go the message to system handler of object
        if(pObject->pObjFunc->OnSysMessage)
            pObject->pObjFunc->OnSysMessage(pMsg); 
      }
      
      if(pObject->pRelations)
      {
        D4D_OBJECT** pChildObj = (D4D_OBJECT**)&pObject->pRelations[D4D_OBJECT_USR_DATA_CHILD_IX];                               
        while(*pChildObj)
        {
          pMsg->pObject = *pChildObj;
          D4D_SendMessageMask(pMsg, parentFlagsMask, endFlagMask);
          pChildObj++;
        }
      } 
      
    }
}


/**************************************************************//*!
*
* helper to invoke OBJECT's message handler 
*
* all information need to be already prepared in d4d_msg
*
******************************************************************/

void D4D_SendMessageBack(D4D_MESSAGE* pMsg)
{
  D4D_OBJECT* pObject = pMsg->pObject;
  
  if(pObject->pRelations)
  {
    D4D_OBJECT* pParentObj = (D4D_OBJECT*)pObject->pRelations[D4D_OBJECT_USR_DATA_PARENT_IX];                               
    if(pParentObj)
    {
      pMsg->pObject = pParentObj;
      D4D_SendMessageBack(pMsg);        
    }
  }
    
  pMsg->pObject = pObject;
  D4D_SendMessage(pMsg);   
}

/**************************************************************//*!
*
* Get the last state of keys
*
******************************************************************/

D4D_KEYS D4D_GetKeys(void)
{
  return (D4D_KEYS)0;
}

/**************************************************************//*!
*
* Handle the keys (part of main poll call)
*
******************************************************************/
void D4D_HandleKeys(void)
{
    // get active screen
    D4D_SCREEN* pScreen = D4D_GetActiveScreen();
    D4D_OBJECT* pFocus;
    D4D_KEY_SCANCODE scanCode;
    D4D_BOOL tmp_release;
    
    // TODO what about sending keys to Child objects??
    
    // Check the buffer of input keys changes
    if(d4d_KeysBuff.writePos == d4d_KeysBuff.readPos)
      return;
    
    // Read the latest record in the buffer of input key event
    scanCode = d4d_KeysBuff.buff[d4d_KeysBuff.readPos++];
    
    // Check the overflow of the read pointer of readed input keys
    if(d4d_KeysBuff.readPos >= D4D_KEYS_BUFF_LENGTH)
        d4d_KeysBuff.readPos = 0;
    
    tmp_release = (D4D_BOOL)(scanCode & D4D_KEY_SCANCODE_RELEASEMASK);
    scanCode &= ~D4D_KEY_SCANCODE_RELEASEMASK;

    // do we handle the keys ourselves (to navigate across the screen) ?
    if(!d4d_pKeysCapturer)
    {
        if((scanCode & D4D_KEY_SCANCODE_RELEASEMASK) == 0)
        {                
          // escape the screen?
          if(scanCode == D4D_KEY_SCANCODE_ESC)
          {            
            if(tmp_release == 0)
              D4D_EscapeScreen();
            return;
          }
          // focus previous object on the screen?
          else if(scanCode ==  D4D_KEY_FUNC_FOCUS_PREV)          
          {
            if(tmp_release == 0)
              D4D_FocusPrevObject(pScreen);
            return;
          }          
          // focus next object on the screen?
          else if(scanCode ==  D4D_KEY_FUNC_FOCUS_NEXT)
          {
            if(tmp_release == 0)
              D4D_FocusNextObject(pScreen, D4D_FALSE);
            return;
          }
        }
    }
             
    // does the object get the key events?    
    // yes, invoke the key events to the active object
    pFocus = (D4D_OBJECT*) (d4d_pKeysCapturer ? d4d_pKeysCapturer : D4D_GetFocusedObject(pScreen));
    if((pFocus->pData->flags & (D4D_OBJECT_F_VISIBLE | D4D_OBJECT_F_ENABLED | D4D_OBJECT_F_TABSTOP)) == (D4D_OBJECT_F_VISIBLE | D4D_OBJECT_F_ENABLED | D4D_OBJECT_F_TABSTOP))
    {
      if(D4D_IsEnabled(pFocus))
      {
        // prepare the message            
        d4d_msg.pScreen = pScreen;
        d4d_msg.pObject = pFocus;
            
        if(tmp_release == 0)
          d4d_msg.nMsgId = D4D_MSG_KEYDOWN;   // if key was pressed down?
        else      
          d4d_msg.nMsgId = D4D_MSG_KEYUP;     // if key was released up?
        
        d4d_msg.prm.key = (D4D_KEY_SCANCODE)(scanCode & D4D_KEY_SCANCODE_KEYMASK);

        D4D_SendMessageBack(&d4d_msg);
      }
    }
}

#ifdef __CWCC__
  #pragma mark -
  #pragma mark Touch Screen Functions
  #pragma mark -
#endif

#ifdef D4D_LLD_TCH

/**************************************************************//*!
*
* get the last TouchScreen coordinates
*
******************************************************************/

D4D_POINT D4D_GetTouchScreenCoordinates(D4D_OBJECT* pObject)
{
  return D4D_GetScreenToClientPoint(pObject, &d4d_TouchScreenCoor);
}



/**************************************************************//*!
*
* Handle the TouchScreen (part of main poll call)
*
******************************************************************/

static D4D_BOOL D4D_FindTouchedObject(D4D_OBJECT* pObject, D4D_POINT point)
{
  D4D_BOOL validCoor;      
  
  if((pObject->pData->flags & (D4D_OBJECT_F_VISIBLE | D4D_OBJECT_F_ENABLED | D4D_OBJECT_F_TOUCHENABLE)) != (D4D_OBJECT_F_VISIBLE | D4D_OBJECT_F_ENABLED | D4D_OBJECT_F_TOUCHENABLE))
    return D4D_FALSE;
  
  if(pObject->pObjFunc->CheckCoordinates == NULL)
    validCoor = D4D_ObjectCheckCoordinates(pObject, D4D_GetScreenToClientPoint(pObject, &point)); 
  else    
    // Check the coordination if its valid
    validCoor = pObject->pObjFunc->CheckCoordinates(pObject, D4D_GetScreenToClientPoint(pObject, &point)); 
  
  // if the coordinations are valid or autorepeat is enabled
  if(validCoor)
  {                        
    // This is new touch
    d4d_LastTouchedObj = pObject;
    d4d_msg.pObject = pObject;
    d4d_msg.nMsgId = D4D_MSG_TOUCHED;
    // send the D4D_MSG_TOUCHED message
    D4D_SendMessage(&d4d_msg);
    
    if(pObject->pRelations)
    {
      D4D_OBJECT** pChildren = (D4D_OBJECT**)&pObject->pRelations[D4D_OBJECT_USR_DATA_CHILD_IX];
      
      while(*pChildren && !D4D_FindTouchedObject(*pChildren, point))
      {
        pChildren++;  
      }
    }
    
    return D4D_TRUE;
  }
  
  return D4D_FALSE;
}

void D4D_HandleTouchScreen(D4D_SCREEN* pScreen)
{
  D4D_SCREEN_DATA* pData = pScreen->pData;
  D4D_BOOL validCoor;      
  D4D_OBJECT** pObjects;          
  //D4D_POINT tmpClientTouch = D4D_GetScreenToClientPoint(pScreen, &d4d_TouchScreenCoor);
    
  // prepare message
  d4d_msg.pScreen = pScreen;
  d4d_msg.prm.draw = 0;
  
  //1. Check if this is new touch or not
  if(d4d_TouchScreen_Status & D4D_TCHSCR_STATUS_NEWTOUCH)
  {        
    // new touch operation
    d4d_TouchScreen_Status &= ~(D4D_TCHSCR_STATUS_NEWTOUCH | D4D_TCHSCR_STATUS_LOSTTOUCH);
     
    // Handle Screen events
    if(pData->flags & D4D_SCR_FINT_TOUCHENABLE)
    {  
      if(pScreen->flags & D4D_SCR_F_EXIT)
      {
        if(D4D_ScrCheckExitBtnCoor(pScreen, &d4d_TouchScreenCoor))
        {        
          D4D_EscapeScreen();
          d4d_TouchScreen_Status &= ~D4D_TCHSCR_STATUS_NEWTOUCH;      
          return;
        }
      }
      
      if(D4D_ScrCheckCoor(pScreen, &d4d_TouchScreenCoor))
      {        
        d4d_msg.pObject = NULL;
        d4d_msg.nMsgId = D4D_MSG_TOUCHED;
        // send the D4D_MSG_TOUCHED message
        D4D_SendMessage(&d4d_msg);                  
      }
    }
    
    pObjects = (D4D_OBJECT**)pScreen->pObjects;
        
    // check all objects   
    while(*pObjects)
    {
      if(D4D_FindTouchedObject(*pObjects, d4d_TouchScreenCoor))
        break;
      pObjects++;
    }
    
  }else
  {
    if(d4d_LastTouchedObj)
    { 
      //Untouch operation
      if(!(d4d_TouchScreen_Status & D4D_TCHSCR_STATUS_TOUCHED))
      {
        // prepare message
        d4d_msg.nMsgId = D4D_MSG_UNTOUCHED;            
        d4d_msg.prm.draw = 0;
        
        if(pData->flags & D4D_SCR_FINT_TOUCHENABLE)
        {
          if(D4D_ScrCheckCoor(pScreen, &d4d_TouchScreenCoor))
          {
            d4d_msg.pObject = NULL;
            // send the D4D_MSG_UNTOUCHED message to screen
            D4D_SendMessage(&d4d_msg);  // To screen
          }
        }
        
        if(d4d_LastTouchedObj)
        {
          d4d_msg.pObject = d4d_LastTouchedObj;
          d4d_LastTouchedObj = NULL;
          // send the D4D_MSG_UNTOUCHED message to Object              
        }else
          return;
      }
      else
      {
        //Autorepeat operations          
        
        // Handle Screen events
        if(pData->flags & D4D_SCR_FINT_TOUCHENABLE)
        {  
          if(D4D_ScrCheckCoor(pScreen, &d4d_TouchScreenCoor))
          {        
            d4d_msg.pObject = NULL;
            d4d_msg.nMsgId = D4D_MSG_TOUCH_AUTO;
            // send the D4D_MSG_TOUCH_AUTO message
            D4D_SendMessage(&d4d_msg);  // To screen
          }
        }

        d4d_msg.pObject = d4d_LastTouchedObj;
      
        // Check the coordination if its valid
        if(d4d_LastTouchedObj->pObjFunc->CheckCoordinates)
          validCoor = d4d_LastTouchedObj->pObjFunc->CheckCoordinates(d4d_LastTouchedObj, D4D_GetScreenToClientPoint(d4d_LastTouchedObj, &d4d_TouchScreenCoor));
        else
          validCoor = D4D_ObjectCheckCoordinates(d4d_LastTouchedObj, D4D_GetScreenToClientPoint(d4d_LastTouchedObj, &d4d_TouchScreenCoor));
        
        // autorepeat touch 
        if(!validCoor && (d4d_TouchScreen_Status & D4D_TCHSCR_STATUS_LOSTTOUCH))
          return;
        
        // Check if the autorepeat touch is valid or no (LOST TOUCH EVENT)
        if(!validCoor)
        {
          d4d_TouchScreen_Status |= D4D_TCHSCR_STATUS_LOSTTOUCH;
          d4d_msg.nMsgId = D4D_MSG_TOUCH_LOST;
        }else
        {
          d4d_TouchScreen_Status &= ~D4D_TCHSCR_STATUS_LOSTTOUCH;
          d4d_msg.nMsgId = D4D_MSG_TOUCH_AUTO;
        }
      }                
        // Send prepared message
        D4D_SendMessageBack(&d4d_msg);      
      }
  }
}

/**************************************************************//*!
*
* Check  the TouchScreen - User function
*
******************************************************************/


void D4D_CheckTouchScreen(void)
{
  d4d_TouchScreen_Status |= D4D_TCHSCR_STATUS_CHECKTOUCH;
}

void D4D_PutTouchScreen(D4D_BOOL touched, D4D_COOR x, D4D_COOR y)
{
  if((d4d_TouchScreen_Status & D4D_TCHSCR_STATUS_EVENTS) == D4D_TCHSCR_STATUS_NEWEVENT)
    return;

  d4d_touchEvent.touched = touched;
  d4d_touchEvent.point.x = x;
  d4d_touchEvent.point.y = y;
  
  d4d_TouchScreen_Status |= D4D_TCHSCR_STATUS_NEWEVENT; 
  
}

void D4D_PutRawTouchScreen(D4D_BOOL touched, D4D_COOR x, D4D_COOR y)
{
  if((d4d_TouchScreen_Status & D4D_TCHSCR_STATUS_EVENTS) == D4D_TCHSCR_STATUS_NEWRAWEVENT)
    return;

  d4d_touchEvent.touched = touched;
  d4d_touchEvent.point.x = x;
  d4d_touchEvent.point.y = y;
  
  d4d_TouchScreen_Status |= D4D_TCHSCR_STATUS_NEWRAWEVENT;   
}




//
//-----------------------------------------------------------------------------
// FUNCTION:    D4D_GetTouchScreenPosition
// SCOPE:       Global
// DESCRIPTION: Reads touch screen and returns X, Y coorinates if screen 
//              touched 
// PARAMETERS:  LWord *TouchPositionX    Pointer to X coordinate
//              LWord *TouchPositionY    Pointer to Y ccordinate
// RETURNS:     0   no screen touch
//              1   screen touch, 
//-----------------------------------------------------------------------------
//
D4D_BOOL D4D_GetTouchScreen(D4D_COOR *TouchPositionX, D4D_COOR *TouchPositionY)
{
    // Declare and initialise local variables
    D4D_BOOL ScreenTouch;

    
    if(d4d_TouchScreen_Status & D4D_TCHSCR_STATUS_CHECKTOUCH)
    {    	
	    // Read raw touch position into *TouchPositionX and *TouchPositionY
	    ScreenTouch = D4D_TCH_GetRawPosition(TouchPositionX, TouchPositionY);
    }
    else
    {
    	ScreenTouch = d4d_touchEvent.touched;
    	*TouchPositionX = d4d_touchEvent.point.x;
    	*TouchPositionY = d4d_touchEvent.point.y;    	
    }
    
    
    if(ScreenTouch)
    {
      if(!(d4d_TouchScreen_Status & D4D_TCHSCR_STATUS_NEWEVENT))	
        D4D_TCH_GetCalibratedPosition(TouchPositionX, TouchPositionY);
    }
    // Return flag to indicate if screen has been touched
    return ScreenTouch;

}

void D4D_ManageTouchScreenEvents(void)
{
  D4D_COOR x,y;
  D4D_BOOL tmp_res;
  
  if(D4D_TCH_GetCalibrationStatus() == 0)
  {
    D4D_CalibrateTouchScreen();
    return; 
  }
  
  if(d4d_TouchScreen_Status & D4D_TCHSCR_STATUS_NEWTOUCH)    
    return;
  
  tmp_res = D4D_GetTouchScreen(&x, &y);
  
  if(((d4d_TouchScreen_Status & D4D_TCHSCR_STATUS_TOUCHED) ^ (tmp_res)))
  {    
    if(tmp_res)      
      d4d_TouchScreen_Status |= D4D_TCHSCR_STATUS_TOUCHED;
    else
      d4d_TouchScreen_Status &= ~D4D_TCHSCR_STATUS_TOUCHED;
    
    if(tmp_res)
      d4d_TouchScreen_Status |= D4D_TCHSCR_STATUS_NEWTOUCH;         
  }
  
  if(tmp_res)
  {
    d4d_TouchScreenCoor.x = (D4D_COOR)x;
    d4d_TouchScreenCoor.y = (D4D_COOR)y; 
    
    #ifdef D4D_INPUT_EVENT_CALLBACK
      D4D_INPUT_EVENT_CALLBACK();
    #endif   
  }
  
}

//-----------------------------------------------------------------------------
// FUNCTION:    D4D_ClearTouchScreenEvents
// SCOPE:       Global
// DESCRIPTION: Clears the all touch screen events
//              
// PARAMETERS:  none
// RETURNS:     none
//-----------------------------------------------------------------------------
void D4D_ClearTouchScreenEvents(void)
{
  d4d_TouchScreen_Status = 0;
}


/**************************************************************//*!
*
* Calibrate the TouchScreen - User function
*
******************************************************************/


void D4D_CalibrateTouchScreen(void) 
{
  D4D_TCH_Calibrate(D4D_COLOR_SYSTEM_FORE, D4D_COLOR_SYSTEM_BCKG);
  D4D_InvalidateScreen(D4D_GetActiveScreen(), D4D_TRUE);  
}


D4D_TOUCHSCREEN_CALIB D4D_GetTouchScreenCalibration(void)
{
  return D4D_TCH_GetCalibration();  
}

void D4D_SetTouchScreenCalibration(D4D_TOUCHSCREEN_CALIB newCalib)
{
  D4D_TCH_SetCalibration(newCalib);  
}




#else


#ifdef __CWCC__
  #pragma mark -
  #pragma mark Dummy Touch Screen Functions
  #pragma mark -
#endif

  D4D_POINT D4D_GetTouchScreenCoordinates(D4D_OBJECT* pObject)
  {
    D4D_UNUSED(pObject);
    D4D_POINT tmp = {0, 0};
    return tmp;
  } 
    
  void D4D_HandleTouchScreen(D4D_SCREEN* pScreen)
  {
    D4D_UNUSED(pScreen);
  }
  
  void D4D_CheckTouchScreen(void)
  {
  }
  
  D4D_TOUCHSCREEN_CALIB D4D_GetTouchScreenCalibration(void)
  {
    D4D_TOUCHSCREEN_CALIB tmp = { 0, 0, 0, 0, 0 };
    return tmp;  
  }
  
  void D4D_SetTouchScreenCalibration(D4D_TOUCHSCREEN_CALIB newCalib)
  {
    D4D_UNUSED(newCalib);
  }
  
  void D4D_CalibrateTouchScreen(void) 
  {
  
  }
  
#endif

#ifdef __CWCC__
  #pragma mark -
  #pragma mark Help Strings & Fonts Functions
  #pragma mark -
#endif

/**************************************************************//*!
*
* convert decimal number to string
*
******************************************************************/

Byte D4D_SprintDecU32(LWord val, D4D_TCHAR *pText, D4D_TCHAR fill) 
{
  LWord value;
  LWord divider;
  Byte cnt = 1;
  Byte fill_cnt = 0;
  Byte tmp_cnt;
  value = val;
  divider =  1000000000;
  
  for(tmp_cnt = 0; tmp_cnt < 9; tmp_cnt++)
  {
    if((value > divider - 1) || (cnt > 1))
    {    
      if(pText != NULL)
        *pText++ = (D4D_TCHAR)((value/divider)+'0');
      cnt++;
    } else if((pText != NULL) && (fill)) 
    {
      fill_cnt++;
      *pText++ = fill;
    }
    
    value %= divider;			
    divider /= 10;
  }
  
	if(pText != NULL)
    *pText++ = (D4D_TCHAR)((value)+'0');
  
  return (Byte)(cnt + fill_cnt);
}

/**************************************************************//*!
*
* convert signed decimal number to string
*
******************************************************************/

Byte D4D_SprintDecS32(sLWord val, D4D_TCHAR *pText, D4D_TCHAR fill) 
{
  Byte cnt = 0;
  if(val < 0)
  {
    *pText++ = '-';
    cnt++;
    val *= -1;
  }
  return (Byte)(cnt + D4D_SprintDecU32((LWord)val, pText, fill));
}

/**************************************************************//*!
*
* convert decimal number to string
*
******************************************************************/

Byte D4D_SprintDecU16(Word val, D4D_TCHAR *pText, D4D_TCHAR fill) 
{
  Word value;
  Word divider;
  Byte cnt = 1;
  Byte fill_cnt = 0;
  Byte tmp_cnt;
  value = val;
  divider =  10000;
  
  for(tmp_cnt = 0; tmp_cnt < 4; tmp_cnt++)
  {
    if((value > divider - 1) || (cnt > 1))
    {    
      if(pText != NULL)
        *pText++ = (D4D_TCHAR)((value/divider)+'0');
      cnt++;
    }else if((pText != NULL) && (fill)) 
    {
      fill_cnt++;
      *pText++ = fill;
    }
        
    value %= divider;			
    divider /= 10;
  }

	if(pText != NULL)
    *pText++ = (D4D_TCHAR)((value)+'0');
  
  return (Byte)(cnt + fill_cnt);
}

/**************************************************************//*!
*
* convert signed decimal number to string
*
******************************************************************/

Byte D4D_SprintDecS16(sWord val, D4D_TCHAR *pText, D4D_TCHAR fill) 
{
  Byte cnt = 0;
  if(val < 0)
  {
    *pText++ = '-';
    cnt++;
    val *= -1;
  }
  return (Byte)(cnt + D4D_SprintDecU16((Word)val, pText, fill));
}

/**************************************************************//*!
*
* convert decimal number to string
*
******************************************************************/

Byte D4D_SprintDecU8(Byte val, D4D_TCHAR *pText, D4D_TCHAR fill) 
{
  Byte value;
  Byte cnt = 1;
  Byte fill_cnt = 0;  
  value = val;
  
  if(value > 99)
  {    
    cnt++;
    if(pText != NULL)
      *pText++ = (D4D_TCHAR)((value/100)+'0');
    value %= 100;
    
  }else if((pText != NULL) && (fill))
  {
    fill_cnt++;
    *pText++ = fill;
  }
  
  if((value > 9) || (cnt > 1))
  {    
    cnt++;
    if(pText != NULL)
      *pText++ = (D4D_TCHAR)((value/10)+'0');
    value %= 10;
    
  }else if((pText != NULL) && (fill))
  {
    fill_cnt++;
    *pText++ = fill;
  }
      
  			
	
	if(pText != NULL)
    *pText++ = (D4D_TCHAR)((value)+'0');
  
  return (Byte)(cnt + fill_cnt);
}

/**************************************************************//*!
*
* convert signed decimal number to string
*
******************************************************************/

Byte D4D_SprintDecS8(sByte val, D4D_TCHAR *pText, D4D_TCHAR fill) 
{
  Byte cnt = 0;
  if(val < 0)
  {
    *pText++ = '-';
    cnt++;
    val *= -1;
  }
  return (Byte)(cnt + D4D_SprintDecU8((Byte)val, pText, fill));
}

/**************************************************************//*!
*
* Get curent text length
*
******************************************************************/

D4D_INDEX D4D_GetTextLength(D4D_TCHAR* pText) 
{
  D4D_INDEX cnt = 0;
  
  if(pText == NULL)
    return 0;
  
#if D4D_EXTSRC_TEXT_ENABLE != D4D_FALSE    
  if(*pText == d4d_extsrcLocale.strTranslatePrefix)
    cnt = (D4D_INDEX)d4d_extsrcLocale.D4DLOCALE_GetTranslateStrLength(pText);
  else
#endif
  {
    while (*pText++)
      cnt++;
  }
  return cnt;
}


/**************************************************************//*!
*
* Get curent text length
*
******************************************************************/
D4D_COOR D4D_GetTextWidth(D4D_FONT ix, D4D_TCHAR* pText){
  
  D4D_COOR width = 0;

  if(pText == NULL)
    return 0;
  
#if D4D_EXTSRC_TEXT_ENABLE == D4D_TRUE    

  if(*pText == d4d_extsrcLocale.strTranslatePrefix)
  {
    D4D_INDEX charCnt = 0;
    D4D_INDEX charOff = 0;
    D4D_INDEX charCntTmp; 
    
    do
    {
      charOff += charCnt;
      
      charCnt = (D4D_INDEX)d4d_extsrcLocale.D4DLOCALE_TranslateStr(pText, (D4D_TCHAR*)d4d_extsrcBuffer, charOff, D4D_EXTSRC_BUFF_SIZE);
    
      for(charCntTmp = 0; charCntTmp < charCnt; charCntTmp++)
        width += D4D_GetCharWidth(ix, ((D4D_TCHAR*)d4d_extsrcBuffer)[charCntTmp]);  
        
    }while(charCnt == D4D_EXTSRC_BUFF_SIZE);
    
  
  }else
  {
    while(*(pText) != 0)
      width += D4D_GetCharWidth(ix, *pText++); 
  }


#else  

  while(*(pText) != 0)
    width += D4D_GetCharWidth(ix, *pText++);

#endif

  return width;  
}

/**************************************************************//*!
*
* Get curent text length
*
******************************************************************/
D4D_COOR D4D_GetTextBuffWidthTab(D4D_STRING* text_buffer, D4D_TAB* pTab){
  
  D4D_COOR width = 0;
  D4D_COOR tab;
  D4D_INDEX char_ix = 0;
  D4D_TCHAR* pText = text_buffer->pText; 

  if(pText == NULL)
    return 0;
  
#if D4D_EXTSRC_TEXT_ENABLE == D4D_TRUE    

  if(*pText == d4d_extsrcLocale.strTranslatePrefix)
  {
    D4D_INDEX charCnt = 0;
    D4D_INDEX charOff = text_buffer->printOff;
    D4D_INDEX charCntTmp; 
    
    do
    {
      charOff += charCnt;
      
      charCnt = (D4D_INDEX)d4d_extsrcLocale.D4DLOCALE_TranslateStr(pText, (D4D_TCHAR*)d4d_extsrcBuffer, charOff, D4D_EXTSRC_BUFF_SIZE);
    
      for(charCntTmp = 0; charCntTmp < charCnt; charCntTmp++)
      {
        if(++char_ix > text_buffer->printLen)  
          return width;
        
        if(((D4D_TCHAR*)d4d_extsrcBuffer)[charCntTmp] == '\t')
        {
          tab = D4D_GetNextTab(pTab, width);  
          
          if(tab)
          {
            width = (D4D_COOR)(tab - (pTab->tabOffset));
            continue;  
          }
        } 
        
        width += D4D_GetCharWidth(text_buffer->fontId, ((D4D_TCHAR*)d4d_extsrcBuffer)[charCntTmp]);
      }        
    }while(charCnt == D4D_EXTSRC_BUFF_SIZE);
    
  
  }else

#endif  
  
  {
    pText += text_buffer->printOff;
    
    while(*pText != 0)
    {    
      
      if(++char_ix > text_buffer->printLen)  
        break;

      if(*pText == '\t')
      {
        tab = D4D_GetNextTab(pTab, width);  
        
        if(tab)
        {
          width = (D4D_COOR)(tab - (pTab->tabOffset));
          pText++;
          continue;  
        }
      } 
      
      width += D4D_GetCharWidth(text_buffer->fontId, *pText);
      pText++;
    }
  }
  
  
  return width;  
}
/**************************************************************//*!
*
* Find out the maximal text length that fits to maximal pixel width
*
******************************************************************/
D4D_INDEX D4D_GetTextMaxLength(D4D_STRING* pString, D4D_COOR maxWidth)
{
  D4D_TCHAR* pText;
  D4D_COOR width;
  D4D_INDEX cnt = 0;
  
  if(pString == NULL)
    return 0;
  
  pText = pString->pText;
  
  if(pText)
  {
    #if D4D_EXTSRC_TEXT_ENABLE == D4D_TRUE    
    if(*pText == d4d_extsrcLocale.strTranslatePrefix)
    {
      D4D_INDEX charCnt = 0;
      D4D_INDEX charOff = pString->printOff;
      D4D_INDEX charCntTmp; 
      
      do
      {
        charOff += charCnt;
        
        charCnt = (D4D_INDEX)d4d_extsrcLocale.D4DLOCALE_TranslateStr(pText, (D4D_TCHAR*)d4d_extsrcBuffer, charOff, D4D_EXTSRC_BUFF_SIZE);
      
        for(charCntTmp = 0; charCntTmp < charCnt; charCntTmp++)
        {
          if(((D4D_TCHAR*)d4d_extsrcBuffer)[charCntTmp] == 0)
            return cnt;
          
          if(cnt >= pString->printLen)
            return cnt;
          
          width = D4D_GetCharWidth(pString->fontId, *pText);
        
          if(width > maxWidth)
            return cnt;
          
          maxWidth -= width;
          cnt++;          
        }        
      }while(charCnt == D4D_EXTSRC_BUFF_SIZE);
      
    
    }
    else
#endif  
    {
      pText += pString->printOff;
      
      while(*pText != 0)
      {       
        if(cnt >= pString->printLen)
          break;
        
        width = D4D_GetCharWidth(pString->fontId, *pText);
        
        if(width > maxWidth)
          break;
        
        maxWidth -= width;
        pText++;
        cnt++;
      }
    }
  }
  return cnt;
}

/**************************************************************//*!
*
* change text in the buffer
*
******************************************************************/

void D4D_ChangeText(D4D_STRING* pBuff, D4D_TCHAR* pNewText, D4D_TCHAR fillChar)
{
    D4D_INDEX n = pBuff->buffSize;
    D4D_TCHAR* pDest = pBuff->pText; 
    
    // preserve one byte in the buffer
    if(pDest != pNewText)
      while(n > 1 && *pNewText)
      {
        *pDest = *pNewText;
        pNewText++;
        pDest++;
        n--;        
      } else
      {
        D4D_INDEX i;
        i = D4D_GetTextLength(pDest);
        n -= i;
        pDest += i;
      }
    
    while(n > 1 && fillChar)
    {
      *(pDest++) = fillChar;
      n--;        
    }
    
    // terminate string always
    *pDest = 0;
}

/**************************************************************//*!
*
* Set Object text 
*
******************************************************************/

void D4D_SetText(D4D_OBJECT_PTR pObject, D4D_TCHAR* pText) 
{
  D4D_STRING* p_TextBuff = NULL;
  
  if(pObject->pObjFunc->GetTextBuffer)
    p_TextBuff = pObject->pObjFunc->GetTextBuffer((D4D_OBJECT*)pObject);
  
  if(p_TextBuff)
  {
    D4D_ChangeText(p_TextBuff, pText, 0);  
    D4D_InvalidateObject(pObject, D4D_FALSE);
  }

}

/**************************************************************//*!
*
* Set Object text font properties
*
******************************************************************/

void D4D_SetFontProperties(D4D_OBJECT_PTR pObject, D4D_FONT_PROPERTIES property) 
{
  
  D4D_STRING* p_TextBuff = NULL;
  
  if(pObject->pObjFunc->GetTextBuffer)
    p_TextBuff = pObject->pObjFunc->GetTextBuffer((D4D_OBJECT*)pObject);
 
  if(p_TextBuff)
  {
    if(p_TextBuff->str_properties->font_properties == property)
      return;     // There is no change needed
    
    p_TextBuff->str_properties->font_properties = property;
    D4D_InvalidateObject(pObject, D4D_FALSE);
  }

}

/**************************************************************//*!
*
* Set Object text properties
*
******************************************************************/

void D4D_SetTextProperties(D4D_OBJECT_PTR pObject, D4D_TEXT_PROPERTIES property) 
{
  
  D4D_STRING* p_TextBuff = NULL;
  
  if(pObject->pObjFunc->GetTextBuffer)
    p_TextBuff = pObject->pObjFunc->GetTextBuffer((D4D_OBJECT*)pObject);
 
  if(p_TextBuff)
  {
    if(p_TextBuff->str_properties->text_properties == property)
      return;     // There is no change needed
    
    p_TextBuff->str_properties->text_properties = property;
    D4D_InvalidateObject(pObject, D4D_FALSE);
  }

}

#ifdef __CWCC__
  #pragma mark -
  #pragma mark Drawing Functions
  #pragma mark -
#endif


/**************************************************************//*!
*
* Basic drawing
*
******************************************************************/

void D4D_PutPixel(D4D_POINT* ppt, D4D_LINETYPE type, D4D_COLOR color) 
{
  D4D_LCD_PutPixel(ppt->x, ppt->y, type, color);
  d4d_curXY = *ppt;
}

void D4D_PutPixelTo(D4D_LINETYPE type, D4D_COLOR color) 
{
  D4D_LCD_PutPixel(d4d_curXY.x, d4d_curXY.y, type, color);  
}

void D4D_PutPixelXY(D4D_COOR x, D4D_COOR y, D4D_LINETYPE type, D4D_COLOR color) 
{
  D4D_LCD_PutPixel(x, y, type, color);
  d4d_curXY.x = x;
  d4d_curXY.y = y;
}

D4D_POINT D4D_SwapCoor(D4D_POINT point)
{
  D4D_POINT tmp_point;
  
  tmp_point.x = point.y;
  tmp_point.y = point.x;
  
  return tmp_point;
}

void D4D_MoveTo(D4D_POINT* ppt)
{
    d4d_curXY = *ppt;    
}

void D4D_MoveToXY(D4D_COOR x, D4D_COOR y)
{
  d4d_curXY.x = x;
  d4d_curXY.y = y;  
}

void D4D_ClearScreen(D4D_COLOR color) {
  D4D_LCD_Box(0, 0, D4D_DisplayWidth(), D4D_DisplayHeight(), color);
}

void D4D_LineTo(D4D_POINT* ppt, D4D_LINETYPE ltype, D4D_COLOR color)
{
    D4D_LCD_Line(d4d_curXY.x, d4d_curXY.y, ppt->x, ppt->y, ltype, color);
    
    d4d_curXY = *ppt;    
}

void D4D_LineToXY(D4D_COOR x, D4D_COOR y, D4D_LINETYPE ltype, D4D_COLOR color)
{
   D4D_LCD_Line(d4d_curXY.x, d4d_curXY.y, x, y, ltype, color);
   
   d4d_curXY.x = x;    
   d4d_curXY.y = y;
}

void D4D_Rect(D4D_POINT* ppt, D4D_SIZE* psz, D4D_LINETYPE ltype, D4D_COLOR color)
{
    if(psz->cx && psz->cy)
      D4D_LCD_Rect((D4D_COOR)ppt->x, (D4D_COOR)ppt->y, (D4D_COOR)(ppt->x+psz->cx - 1), (D4D_COOR)(ppt->y+psz->cy - 1), (D4D_LINETYPE)ltype, (D4D_COLOR)color);
    
    d4d_curXY = *ppt;    
}

void D4D_RectXY(D4D_COOR x1, D4D_COOR y1, D4D_COOR x2, D4D_COOR y2, D4D_LINETYPE ltype, D4D_COLOR color)
{
    D4D_LCD_Rect(x1, y1, x2, y2, ltype, color);
    
    d4d_curXY.x = x1;    
    d4d_curXY.y = y1;    
}

void D4D_RectTo(D4D_POINT* ppt, D4D_LINETYPE ltype, D4D_COLOR color)
{
    D4D_LCD_Rect(d4d_curXY.x, d4d_curXY.y, ppt->x, ppt->y, ltype, color);
    
    d4d_curXY = *ppt;    
}


void D4D_RectToXY(D4D_COOR x, D4D_COOR y, D4D_LINETYPE ltype, D4D_COLOR color)
{
    D4D_LCD_Rect(d4d_curXY.x, d4d_curXY.y, x, y, ltype, color);
    
    d4d_curXY.x = x;    
    d4d_curXY.y = y;    
}


void D4D_FillRect(D4D_POINT* ppt, D4D_SIZE* psz, D4D_COLOR color)
{
    if(psz->cx && psz->cy)
      D4D_LCD_Box((D4D_COOR)ppt->x, (D4D_COOR)ppt->y, (D4D_COOR)(ppt->x+psz->cx - 1), (D4D_COOR)(ppt->y+psz->cy - 1), (D4D_COLOR)color);
    
    d4d_curXY = *ppt;
}

void D4D_FillRectXY(D4D_COOR x1, D4D_COOR y1, D4D_COOR x2, D4D_COOR y2, D4D_COLOR color)
{
    D4D_LCD_Box(x1, y1, x2, y2, color);
    
    d4d_curXY.x = x1;    
    d4d_curXY.y = y1;
}

void D4D_FillRectTo(D4D_POINT* ppt, D4D_COLOR color)
{
    D4D_LCD_Box(d4d_curXY.x, d4d_curXY.y, ppt->x, ppt->y, (D4D_COLOR)color);
    
    d4d_curXY = *ppt;
}

void D4D_FillRectToXY(D4D_COOR x, D4D_COOR y, D4D_COLOR color)
{
    D4D_LCD_Box(d4d_curXY.x, d4d_curXY.y, x, y, color);
    
    d4d_curXY.x = x;    
    d4d_curXY.y = y;
}

void D4D_Box(D4D_POINT* ppt, D4D_SIZE* psz, D4D_LINETYPE ltype, D4D_COLOR colorLine, D4D_COLOR colorFill)
{    
  if(psz->cx && psz->cy)
    D4D_BoxXY(ppt->x, ppt->y, (D4D_COOR)(ppt->x+psz->cx - 1), (D4D_COOR)(ppt->y+psz->cy - 1), ltype, colorLine, colorFill);
}

void D4D_BoxXY(D4D_COOR x1, D4D_COOR y1, D4D_COOR x2, D4D_COOR y2, D4D_LINETYPE ltype, D4D_COLOR colorLine, D4D_COLOR colorFill)
{
    // fill background
    D4D_LCD_Box(x1, y1, x2, y2, colorFill);
    
    // draw rectangle around
    D4D_LCD_Rect(x1, y1, x2, y2, ltype, colorLine);
    
    d4d_curXY.x = x1;
    d4d_curXY.y = y1;   
}

void D4D_BoxTo(D4D_POINT* ppt, D4D_LINETYPE ltype, D4D_COLOR colorLine, D4D_COLOR colorFill)
{
    D4D_BoxToXY(ppt->x, ppt->y, ltype, colorLine, colorFill);
}

void D4D_BoxToXY(D4D_COOR x, D4D_COOR y, D4D_LINETYPE ltype, D4D_COLOR colorLine, D4D_COLOR colorFill)
{
    // fill background
    D4D_LCD_Box(d4d_curXY.x, d4d_curXY.y, x, y, colorFill);
    
    // draw rectangle around
    D4D_LCD_Rect(d4d_curXY.x, d4d_curXY.y, x, y, ltype, colorLine);
    
    d4d_curXY.x = x;
    d4d_curXY.y = y;   
}


void D4D_FillCircle(D4D_POINT* pCenter, D4D_COOR r, D4D_COLOR color)
{
    D4D_LCD_Circle(pCenter->x, pCenter->y, r, 1, D4D_LINE_THIN, color); 
}

void D4D_FillCircleXY(D4D_COOR x, D4D_COOR y, D4D_COOR r, D4D_COLOR color)
{
    D4D_LCD_Circle(x, y, r, 1, D4D_LINE_THIN, color); 
}

void D4D_Circle(D4D_POINT* pCenter, D4D_COOR r, D4D_LINETYPE ltype, D4D_COLOR color)
{
    D4D_LCD_Circle(pCenter->x, pCenter->y, r, 0, ltype, color); 
}

void D4D_CircleXY(D4D_COOR x, D4D_COOR y, D4D_COOR r, D4D_LINETYPE ltype, D4D_COLOR color)
{
    D4D_LCD_Circle(x, y, r, 0, ltype, color); 
}


void D4D_FillRectColorScale(D4D_POINT* ppt, D4D_SIZE* psz, D4D_COLOR startColor, D4D_COLOR endColor, D4D_DIRECTION direction)
{
  if(psz->cx && psz->cy)
    D4D_FillRectColorScaleXY(ppt->x, ppt->y, (D4D_COOR)(ppt->x+psz->cx - 1), (D4D_COOR)(ppt->y+psz->cy - 1), startColor, endColor, direction);
}

void D4D_FillRectColorScaleXY(D4D_COOR x1, D4D_COOR y1, D4D_COOR x2, D4D_COOR y2, D4D_COLOR startColor, D4D_COLOR endColor, D4D_DIRECTION direction)
{
  D4D_COOR range, i;
  D4D_COLOR color = startColor;
  
  d4d_curXY.x = x1;
  d4d_curXY.y = y1;
  
  switch(direction)
  {
    case D4D_DIR_UP:
      // switch X and Y coodinates and change direction of drawing
      startColor = endColor;
      endColor = color;
    case D4D_DIR_DOWN:
      // just switch X and Y coodinates
      range = (D4D_COOR)(y2 - y1);
      
      i = 0;
      for( ; y1 <= y2; y1++)
      {
        if(range)
          color = D4D_GetCrossColor(startColor, endColor, (Byte)D4D_MulDivUU(255, i, range));
        i++;
        
        D4D_LCD_Box(x1, y1, x2, y1, color);    
      }
      break;
    
    case D4D_DIR_LEFT:
      // switch colors
      startColor = endColor;
      endColor = color;
    case D4D_DIR_RIGHT:
      // just keep all fine
      range = (D4D_COOR)(x2 - x1);
      i = 0;
      for( ; x1 <= x2; x1++)
      {
        if(range)
          color = D4D_GetCrossColor(startColor, endColor, (Byte)D4D_MulDivUU(255, i, range));
        i++;
        
        D4D_LCD_Box(x1, y1, x1, y2, color);    
      }
      
      break;  
  }
  
  
  
  
}

void D4D_FillRectColorScaleTo(D4D_POINT* ppt, D4D_COLOR startColor, D4D_COLOR endColor, D4D_DIRECTION direction)
{
  D4D_FillRectColorScaleXY(d4d_curXY.x, d4d_curXY.y, ppt->x, ppt->y, startColor, endColor, direction);
}

void D4D_FillRectColorScaleToXY(D4D_COOR x, D4D_COOR y, D4D_COLOR startColor, D4D_COLOR endColor, D4D_DIRECTION direction)
{
  D4D_FillRectColorScaleXY(d4d_curXY.x, d4d_curXY.y, x, y, startColor, endColor, direction);
}


/**************************************************************//*!
*
* Draw text on XY coordination with TAB function
*
******************************************************************/
//void D4D_DrawTextTab(D4D_POINT* ppt, D4D_STRING* buffText, D4D_COOR* pTabTable, D4D_COLOR colorText, D4D_COLOR colorBkgd)
//{
//  D4D_DrawTextTabXY(ppt->x, ppt->y, buffText, pTabTable, colorText, colorBkgd); 
//}

/**************************************************************//*!
*
* Draw text on XY coordination with TAB function
*
******************************************************************/
void D4D_DrawTextTabXY(D4D_COOR x, D4D_COOR y, D4D_STRING* buffText, D4D_TAB* pTab, D4D_COLOR colorText, D4D_COLOR colorBkgd)
{
  D4D_PRINT_STRING tmp_strDes;
    
  tmp_strDes.x = x;
  tmp_strDes.y = y;
  tmp_strDes.pText = buffText->pText;
  tmp_strDes.pFontType = D4D_GetFont(buffText->fontId);
  tmp_strDes.pTab = pTab;
  tmp_strDes.colorText = colorText;
  tmp_strDes.colorBack = colorBkgd;
  tmp_strDes.properties = buffText->str_properties->font_properties;
  tmp_strDes.textLength = buffText->printLen;
  tmp_strDes.textOffset = buffText->printOff;
  tmp_strDes.maxWidth = 0;
  D4D_LCD_PrintStr(&tmp_strDes);  
}


/**************************************************************//*!
*
* Draw text on XY coordination
*
******************************************************************/
//void D4D_DrawText(D4D_POINT* ppt, D4D_STRING* buffText, D4D_COLOR colorText, D4D_COLOR colorBkgd)
//{
//  D4D_DrawTextTab(ppt, buffText, NULL, colorText, colorBkgd);
//}

/**************************************************************//*!
*
* Draw text on XY coordination
*
******************************************************************/
//void D4D_DrawTextXY(D4D_COOR x, D4D_COOR y, D4D_STRING* buffText, D4D_COLOR colorText, D4D_COLOR colorBkgd)
//{
//  D4D_DrawTextTabXY(x, y, buffText, NULL, colorText, colorBkgd);
//}

/**************************************************************//*!
*
* Draw text on XY coordination in logic rectangle
*
******************************************************************/
/*
void D4D_DrawTextRectTab(D4D_POINT* ppt, D4D_SIZE* psz, D4D_STRING* buffText, D4D_TAB* pTab, D4D_COLOR colorText, D4D_COLOR colorBkgd)
{
  D4D_PRINT_STRING tmp_strDes;
  D4D_SIZE tmp_size = *psz;
  D4D_TEXT_PROPERTIES tmp_txtProp;
  D4D_COOR tmp_txtWidth = D4D_GetTextBuffWidthTab(buffText, pTab);
  D4D_COOR tmp_txtHeight = D4D_GetFontHeight(buffText->fontId);
  
  tmp_strDes.pText = buffText->pText;
  tmp_strDes.pFontType = D4D_GetFont(buffText->fontId);
  tmp_strDes.pTab = pTab;
  tmp_strDes.colorText = colorText;
  tmp_strDes.colorBack = colorBkgd;
  tmp_strDes.properties.all = buffText->str_properties->font_properties.all;
  tmp_strDes.textLength = (D4D_INDEX)(buffText->buffSize - 1);
  
  if(!tmp_size.cx)
    tmp_size.cx = (D4D_COOR)(tmp_txtWidth + 1);
  
  if(!tmp_size.cy)
    tmp_size.cy = (D4D_COOR)(tmp_txtHeight + 1);
  
  tmp_txtProp.all = buffText->str_properties->text_properties.all;
  
   
  // Compute the text position in rectangle in Y axis
  switch(tmp_txtProp.bits.bAlignVertic)
  {
    case D4D_TXT_PRTY_ALIGN_V_BOTTOM:
      tmp_strDes.y = (D4D_COOR)(ppt->y + (tmp_size.cy - tmp_txtHeight));
      break;
    
    case D4D_TXT_PRTY_ALIGN_V_CENTER:
      tmp_strDes.y = (D4D_COOR)(ppt->y + ((tmp_size.cy - tmp_txtHeight) / 2));
      break;
    
    default:
    case D4D_TXT_PRTY_ALIGN_V_TOP:
      tmp_strDes.y = ppt->y;
      break;        
  }
  
  // Compute the text position in rectangle in X axis
  switch(tmp_txtProp.bits.bAlignHoriz)
  {
    case D4D_TXT_PRTY_ALIGN_H_RIGHT:
      tmp_strDes.x = (D4D_COOR)(ppt->x + (tmp_size.cx - tmp_txtWidth));
      break;
    
    case D4D_TXT_PRTY_ALIGN_H_CENTER:
      tmp_strDes.x = (D4D_COOR)(ppt->x + ((tmp_size.cx - tmp_txtWidth) / 2));
      break;
    
    default:
    case D4D_TXT_PRTY_ALIGN_H_LEFT:
      tmp_strDes.x = ppt->x;
      break;        
  }
  
  // Redraw the rest of rectangle where isn't text for normal (non trasparent) texts
  if(!tmp_strDes.properties.bits.bTransparent)
  {
    
    if(tmp_txtProp.bits.bAlignHoriz != D4D_TXT_PRTY_ALIGN_H_LEFT)
    {
      // Fill the part of rect in front of the text in full height of rect
      //if(tmp_strDes.x - ppt->x)
        D4D_FillRectXY(ppt->x, ppt->y, tmp_strDes.x, (D4D_COOR)(ppt->y + tmp_size.cy - 1), colorBkgd);
    }
    
    if(tmp_txtProp.bits.bAlignHoriz != D4D_TXT_PRTY_ALIGN_H_RIGHT)
    {
      // Fill the part of rect behind the text in full height of rect
      //if((D4D_COOR)(ppt->x + tmp_size.cx - 1) - (D4D_COOR)(tmp_strDes.x + tmp_txtWidth - 1))
        D4D_FillRectXY((D4D_COOR)(tmp_strDes.x + tmp_txtWidth - 1), ppt->y, (D4D_COOR)(ppt->x + tmp_size.cx - 1), (D4D_COOR)(ppt->y + tmp_size.cy - 1), colorBkgd);
    }
    

      
    if(tmp_txtProp.bits.bAlignVertic != D4D_TXT_PRTY_ALIGN_V_TOP)
    {
      // Fill the part of rect above the text in lenght of text
      D4D_FillRectXY(tmp_strDes.x, ppt->y, (D4D_COOR)(tmp_strDes.x + tmp_txtWidth - 1), tmp_strDes.y, colorBkgd);
    }
    
    if(tmp_txtProp.bits.bAlignVertic != D4D_TXT_PRTY_ALIGN_V_BOTTOM)
    {
      // Fill the part of rect under the text in lenght of text
      D4D_FillRectXY(tmp_strDes.x, (D4D_COOR)(tmp_strDes.y + tmp_txtHeight - 1), (D4D_COOR)(tmp_strDes.x + tmp_txtWidth - 1), (D4D_COOR)(ppt->y + tmp_size.cy - 1), colorBkgd);
    }

  }
  
  // Print the string itself
  D4D_LCD_PrintStr(&tmp_strDes);
}
  */
/**************************************************************//*!
*
* Draw text on XY coordination in logic rectangle
*
******************************************************************/
/*
void D4D_DrawTextRectTabTo(D4D_POINT* ppt, D4D_STRING* buffText, D4D_TAB* pTab, D4D_COLOR colorText, D4D_COLOR colorBkgd)
{
  D4D_SIZE tmp_size;
  
  tmp_size.cx = (D4D_COOR)(ppt->x - d4d_curXY.x);
  tmp_size.cy = (D4D_COOR)(ppt->y - d4d_curXY.y);
  
  D4D_DrawTextRectTab( &d4d_curXY, &tmp_size, buffText, pTab, colorText, colorBkgd);
}
*/
/**************************************************************//*!
*
* Draw text on XY coordination in logic rectangle
*
******************************************************************/
/*
void D4D_DrawTextRectTabXY(D4D_COOR x1, D4D_COOR y1, D4D_COOR x2, D4D_COOR y2, D4D_STRING* buffText, D4D_TAB* pTab, D4D_COLOR colorText, D4D_COLOR colorBkgd)
{
  D4D_POINT tmp_point;
  D4D_SIZE tmp_size;
  
  tmp_point.x = x1;
  tmp_point.y = y1;
  
  tmp_size.cx = (D4D_COOR)(x2 - x1);
  tmp_size.cy = (D4D_COOR)(y2 - y1);
  
  
  D4D_DrawTextRectTab( &tmp_point, &tmp_size, buffText, pTab, colorText, colorBkgd);
}
*/
/**************************************************************//*!
*
* Draw text on XY coordination in logic rectangle
*
******************************************************************/
/*
void D4D_DrawTextRectTabToXY(D4D_COOR x, D4D_COOR y, D4D_STRING* buffText, D4D_TAB* pTab, D4D_COLOR colorText, D4D_COLOR colorBkgd)
{
  D4D_SIZE tmp_size;
  
  tmp_size.cx = (D4D_COOR)(x - d4d_curXY.x);
  tmp_size.cy = (D4D_COOR)(y - d4d_curXY.y);
  
  
  D4D_DrawTextRectTab( &d4d_curXY, &tmp_size, buffText, pTab, colorText, colorBkgd);
}
*/

/**************************************************************//*!
*
* Draw text on XY coordination in logic rectangle
*
******************************************************************/
void D4D_DrawTextRectTab(D4D_POINT* ppt, D4D_SIZE* psz, D4D_STRING* buffText, D4D_TAB* pTab, D4D_COLOR colorText, D4D_COLOR colorBkgd)
{
  D4D_PRINT_STRING tmp_strDes;
  D4D_TEXT_PROPERTIES tmp_txtProp;
  D4D_COOR tmp_txtWidth = D4D_GetTextBuffWidthTabAdj(buffText, pTab);
  D4D_COOR tmp_txtHeight = D4D_GetFontHeight(buffText->fontId);
  
  tmp_strDes.pText = buffText->pText;
  tmp_strDes.pFontType = D4D_GetFont(buffText->fontId);
  tmp_strDes.pTab = pTab;
  tmp_strDes.colorText = colorText;
  tmp_strDes.colorBack = colorBkgd;
  tmp_strDes.properties = buffText->str_properties->font_properties;
  tmp_strDes.textLength = buffText->printLen;
  tmp_strDes.textOffset = buffText->printOff;
  
  if(!psz->cx || !psz->cy)
    return;
  
  if(tmp_txtHeight > psz->cy)
    return;
  
  if(tmp_txtWidth >= psz->cx)
    tmp_txtWidth = psz->cx;  
  
  tmp_strDes.maxWidth = tmp_txtWidth;
  
  tmp_txtProp = buffText->str_properties->text_properties;
  
  tmp_strDes.x = ppt->x;
  tmp_strDes.y = ppt->y;
  
  
  if(tmp_txtHeight != psz->cy)
  {
    // Compute the text position in rectangle in Y axis
    switch(tmp_txtProp & D4D_ALIGN_V_MASK)
    {
      case D4D_ALIGN_V_BOTTOM_MASK:
        tmp_strDes.y += (D4D_COOR)(psz->cy - tmp_txtHeight);
        break;
      
      case D4D_ALIGN_V_CENTER_MASK:
        tmp_strDes.y += (D4D_COOR)((psz->cy - tmp_txtHeight) / 2);
        break;
    }
  }
     
  
  if(tmp_txtWidth != psz->cx)
  {
    // Compute the text position in rectangle in X axis
    switch(tmp_txtProp & D4D_ALIGN_H_MASK)
    {
      case D4D_ALIGN_H_RIGHT_MASK:
        tmp_strDes.x += (D4D_COOR)(psz->cx - tmp_txtWidth);
        break;
      
      case D4D_ALIGN_H_CENTER_MASK:
        tmp_strDes.x += (D4D_COOR)((psz->cx - tmp_txtWidth) / 2);
        break;            
    }
  }
    
  
  // Redraw the rest of rectangle where isn't text for normal (non trasparent) texts
  if(!(tmp_strDes.properties & D4D_FNT_PRTY_TRANSPARENT_MASK))
  {
    if(tmp_txtWidth != psz->cx)
    {
      if((tmp_txtProp & D4D_ALIGN_H_MASK) != D4D_ALIGN_H_LEFT_MASK)
      {
        // Fill the part of rect in front of the text in full height of rect
          D4D_FillRectXY(ppt->x, ppt->y, tmp_strDes.x, (D4D_COOR)(ppt->y + psz->cy - 1), colorBkgd);
      }
      
      if((tmp_txtProp & D4D_ALIGN_H_MASK) != D4D_ALIGN_H_RIGHT_MASK)
      {
        // Fill the part of rect behind the text in full height of rect
          D4D_FillRectXY((D4D_COOR)(tmp_strDes.x + tmp_txtWidth - 1), ppt->y, (D4D_COOR)(ppt->x + psz->cx - 1), (D4D_COOR)(ppt->y + psz->cy - 1), colorBkgd);
      }
    }    

    if(tmp_txtHeight != psz->cy)
    {
      if((tmp_txtProp & D4D_ALIGN_V_MASK) != D4D_ALIGN_V_TOP_MASK)
      {
        // Fill the part of rect above the text in lenght of text
        D4D_FillRectXY(tmp_strDes.x, ppt->y, (D4D_COOR)(tmp_strDes.x + tmp_txtWidth - 1), tmp_strDes.y, colorBkgd);
      }
      
      if((tmp_txtProp & D4D_ALIGN_V_MASK) != D4D_ALIGN_V_BOTTOM_MASK)
      {
        // Fill the part of rect under the text in lenght of text
        D4D_FillRectXY(tmp_strDes.x, (D4D_COOR)(tmp_strDes.y + tmp_txtHeight - 1), (D4D_COOR)(tmp_strDes.x + tmp_txtWidth - 1), (D4D_COOR)(ppt->y + psz->cy - 1), colorBkgd);
      }
    }
  }
  
  // Print the string itself
  D4D_LCD_PrintStr(&tmp_strDes);
}

/**************************************************************//*!
*
* Draw text on XY coordination in logic rectangle
*
******************************************************************/
void D4D_DrawTextRectTabTo(D4D_POINT* ppt, D4D_STRING* buffText, D4D_TAB* pTab, D4D_COLOR colorText, D4D_COLOR colorBkgd)
{
  D4D_SIZE tmp_size;
  
  tmp_size.cx = (D4D_COOR)(ppt->x - d4d_curXY.x);
  tmp_size.cy = (D4D_COOR)(ppt->y - d4d_curXY.y);
  
  D4D_DrawTextRectTab( &d4d_curXY, &tmp_size, buffText, pTab, colorText, colorBkgd);
}

/**************************************************************//*!
*
* Draw text on XY coordination in logic rectangle
*
******************************************************************/
void D4D_DrawTextRectTabXY(D4D_COOR x1, D4D_COOR y1, D4D_COOR x2, D4D_COOR y2, D4D_STRING* buffText, D4D_TAB* pTab, D4D_COLOR colorText, D4D_COLOR colorBkgd)
{
  D4D_POINT tmp_point;
  D4D_SIZE tmp_size;
  
  tmp_point.x = x1;
  tmp_point.y = y1;
  
  tmp_size.cx = (D4D_COOR)(x2 - x1);
  tmp_size.cy = (D4D_COOR)(y2 - y1);
  
  
  D4D_DrawTextRectTab( &tmp_point, &tmp_size, buffText, pTab, colorText, colorBkgd);
}

/**************************************************************//*!
*
* Draw text on XY coordination in logic rectangle
*
******************************************************************/
void D4D_DrawTextRectTabToXY(D4D_COOR x, D4D_COOR y, D4D_STRING* buffText, D4D_TAB* pTab, D4D_COLOR colorText, D4D_COLOR colorBkgd)
{
  D4D_SIZE tmp_size;
  
  tmp_size.cx = (D4D_COOR)(x - d4d_curXY.x);
  tmp_size.cy = (D4D_COOR)(y - d4d_curXY.y);
  
  
  D4D_DrawTextRectTab( &d4d_curXY, &tmp_size, buffText, pTab, colorText, colorBkgd);
}


#ifdef __CWCC__
  #pragma mark -
  #pragma mark Drawing Round Corner Functions
  #pragma mark -
#endif

#if D4D_ROUND_CORNER_ENABLE == D4D_TRUE

void D4D_RRect(D4D_POINT* ppt, D4D_SIZE* psz, D4D_LINETYPE ltype, D4D_COLOR color, D4D_COOR radius)
{
  if(psz->cx && psz->cy)
    D4D_LCD_RRect((D4D_COOR)ppt->x, (D4D_COOR)ppt->y, (D4D_COOR)(ppt->x+psz->cx - 1), (D4D_COOR)(ppt->y+psz->cy - 1), ltype, color, radius);
    
    d4d_curXY = *ppt;    
}

void D4D_RRectXY(D4D_COOR x1, D4D_COOR y1, D4D_COOR x2, D4D_COOR y2, D4D_LINETYPE ltype, D4D_COLOR color, D4D_COOR radius)
{
    D4D_LCD_RRect(x1, y1, x2, y2, ltype, color,radius);
    
    d4d_curXY.x = x1;    
    d4d_curXY.y = y1;    
}

void D4D_RRectTo(D4D_POINT* ppt, D4D_LINETYPE ltype, D4D_COLOR color, D4D_COOR radius)
{
    D4D_LCD_RRect(d4d_curXY.x, d4d_curXY.y, ppt->x, ppt->y, ltype, color, radius);
    
    d4d_curXY = *ppt;    
}


void D4D_RRectToXY(D4D_COOR x, D4D_COOR y, D4D_LINETYPE ltype, D4D_COLOR color, D4D_COOR radius)
{
    D4D_LCD_RRect(d4d_curXY.x, d4d_curXY.y, x, y, ltype, color, radius);
    
    d4d_curXY.x = x;    
    d4d_curXY.y = y;    
}


void D4D_FillRRect(D4D_POINT* ppt, D4D_SIZE* psz, D4D_COLOR color, D4D_COOR radius)
{
  if(psz->cx && psz->cy)  
    D4D_LCD_RBox((D4D_COOR)ppt->x, (D4D_COOR)ppt->y, (D4D_COOR)(ppt->x + psz->cx - 1), (D4D_COOR)(ppt->y + psz->cy - 1), color, radius);
    d4d_curXY = *ppt;
}

void D4D_FillRRectXY(D4D_COOR x1, D4D_COOR y1, D4D_COOR x2, D4D_COOR y2, D4D_COLOR color, D4D_COOR radius)
{
    D4D_LCD_RBox(x1, y1, x2, y2, color, radius);
    
    d4d_curXY.x = x1;    
    d4d_curXY.y = y1;
}

void D4D_FillRRectTo(D4D_POINT* ppt, D4D_COLOR color, D4D_COOR radius)
{
    D4D_LCD_RBox(d4d_curXY.x, d4d_curXY.y, ppt->x, ppt->y, (D4D_COLOR)color, radius);
    
    d4d_curXY = *ppt;
}

void D4D_FillRRectToXY(D4D_COOR x, D4D_COOR y, D4D_COLOR color, D4D_COOR radius)
{
    D4D_LCD_RBox(d4d_curXY.x, d4d_curXY.y, x, y, color, radius);
    
    d4d_curXY.x = x;    
    d4d_curXY.y = y;
}

void D4D_RBox(D4D_POINT* ppt, D4D_SIZE* psz, D4D_LINETYPE ltype, D4D_COLOR colorLine, D4D_COLOR colorFill, D4D_COOR radius)
{    
  if(psz->cx && psz->cy)  
    D4D_RBoxXY(ppt->x, ppt->y, (D4D_COOR)(ppt->x + psz->cx - 1), (D4D_COOR)(ppt->y + psz->cy - 1), ltype, colorLine, colorFill, radius);
}

void D4D_RBoxXY(D4D_COOR x1, D4D_COOR y1, D4D_COOR x2, D4D_COOR y2, D4D_LINETYPE ltype, D4D_COLOR colorLine, D4D_COLOR colorFill, D4D_COOR radius)
{
    // fill background
    D4D_LCD_RBox(x1, y1, x2, y2, colorFill, radius);
    
    // draw rectangle around
    D4D_LCD_RRect(x1, y1, x2, y2, ltype, colorLine, radius);
    
    d4d_curXY.x = x1;
    d4d_curXY.y = y1;   
}

void D4D_RBoxTo(D4D_POINT* ppt, D4D_LINETYPE ltype, D4D_COLOR colorLine, D4D_COLOR colorFill, D4D_COOR radius)
{
    D4D_RBoxToXY(ppt->x, ppt->y, ltype, colorLine, colorFill, radius);
}

void D4D_RBoxToXY(D4D_COOR x, D4D_COOR y, D4D_LINETYPE ltype, D4D_COLOR colorLine, D4D_COLOR colorFill, D4D_COOR radius)
{
    // fill background
    D4D_LCD_RBox(d4d_curXY.x, d4d_curXY.y, x, y, colorFill, radius);
    
    // draw rectangle around
    D4D_LCD_RRect(d4d_curXY.x, d4d_curXY.y, x, y, ltype, colorLine, radius);
    
    d4d_curXY.x = x;
    d4d_curXY.y = y;   
}


void D4D_FillQuadrant(D4D_POINT* pCenter, D4D_COOR radius, D4D_COLOR color, D4D_QUADRANT quadrant)
{
    D4D_LCD_QCircle(pCenter->x, pCenter->y, radius, D4D_TRUE, D4D_LINE_THIN, color, quadrant); 
}

void D4D_FillQuadrantXY(D4D_COOR x, D4D_COOR y, D4D_COOR radius, D4D_COLOR color, D4D_QUADRANT quadrant)
{
    D4D_LCD_QCircle(x, y, radius, D4D_TRUE, D4D_LINE_THIN, color, quadrant); 
}

void D4D_Quadrant(D4D_POINT* pCenter, D4D_COOR radius, D4D_LINETYPE ltype, D4D_COLOR color, D4D_QUADRANT quadrant)
{
    D4D_LCD_QCircle(pCenter->x, pCenter->y, radius, D4D_FALSE, ltype, color, quadrant); 
}

void D4D_QuadrantXY(D4D_COOR x, D4D_COOR y, D4D_COOR radius, D4D_LINETYPE ltype, D4D_COLOR color, D4D_QUADRANT quadrant)
{
    D4D_LCD_QCircle(x, y, radius, D4D_FALSE, ltype, color, quadrant); 
}

void D4D_FillRRectColorScale(D4D_POINT* ppt, D4D_SIZE* psz, D4D_COLOR startColor, D4D_COLOR endColor, D4D_DIRECTION direction, D4D_COOR radius)
{
  if(psz->cx && psz->cy)
    D4D_FillRRectColorScaleXY(ppt->x, ppt->y, (D4D_COOR)(ppt->x+psz->cx - 1), (D4D_COOR)(ppt->y+psz->cy - 1), startColor, endColor, direction, radius);
}

void D4D_FillRRectColorScaleXY(D4D_COOR x1, D4D_COOR y1, D4D_COOR x2, D4D_COOR y2, D4D_COLOR startColor, D4D_COLOR endColor, D4D_DIRECTION direction, D4D_COOR radius)
{
  D4D_COOR range, i, non_print_len;
  D4D_COLOR color = startColor;
  
  if(radius)
  {
    
    d4d_curXY.x = x1;
    d4d_curXY.y = y1;

    switch(direction)
    {
    case D4D_DIR_UP:
      // switch X and Y coodinates and change direction of drawing
      startColor = endColor;
      endColor = color;
    case D4D_DIR_DOWN:
      // just switch X and Y coodinates
      range = (D4D_COOR)(y2 - y1);
      
      i = 0;
      for( ; y1 <= y2; y1++)
      {
        if(range)
          color = D4D_GetCrossColor(startColor, endColor, (Byte)D4D_MulDivUU(255, i, range));
        

        if(i < radius)
          non_print_len = D4D_RndCornerGetNonPrintedPxl(radius, i); // get len of non printed pixels
        else if((range - i) < radius)
          non_print_len = D4D_RndCornerGetNonPrintedPxl(radius, (D4D_COOR)(range - i)); // get len of non printed pixels
        else
          non_print_len = 0;
      
        i++;
        D4D_LCD_Box((D4D_COOR)(x1 + non_print_len), y1, (D4D_COOR)(x2 - non_print_len), y1, color);    
      }
      break;

    case D4D_DIR_LEFT:
      // switch colors
      startColor = endColor;
      endColor = color;
    case D4D_DIR_RIGHT:
      // just keep all fine
      range = (D4D_COOR)(x2 - x1);
      
      i = 0;
      for( ; x1 <= x2; x1++)
      {
        if(range)
          color = D4D_GetCrossColor(startColor, endColor, (Byte)D4D_MulDivUU(255, i, range));
        

        if(i < radius)
          non_print_len = D4D_RndCornerGetNonPrintedPxl(radius, i); // get len of non printed pixels
        else if((range - i) < radius)
          non_print_len = D4D_RndCornerGetNonPrintedPxl(radius, (D4D_COOR)(range - i)); // get len of non printed pixels
        else
          non_print_len = 0;
      
        i++;
        D4D_LCD_Box(x1, (D4D_COOR)(y1 + non_print_len), x1, (D4D_COOR)(y2 - non_print_len), color);    
      }
      
      break;  
    }
  }else
  {
    D4D_FillRectColorScaleXY(x1, y1, x2, y2, startColor, endColor, direction);  
  }
  
}

void D4D_FillRRectColorScaleTo(D4D_POINT* ppt, D4D_COLOR startColor, D4D_COLOR endColor, D4D_DIRECTION direction, D4D_COOR radius)
{
  D4D_FillRRectColorScaleXY(d4d_curXY.x, d4d_curXY.y, ppt->x, ppt->y, startColor, endColor, direction, radius);
}

void D4D_FillRRectColorScaleToXY(D4D_COOR x, D4D_COOR y, D4D_COLOR startColor, D4D_COLOR endColor, D4D_DIRECTION direction, D4D_COOR radius)
{
  D4D_FillRRectColorScaleXY(d4d_curXY.x, d4d_curXY.y, x, y, startColor, endColor, direction, radius);
}


D4D_COOR D4D_RndCornerGetNonPrintedPxl(D4D_COOR radius, D4D_COOR line)
{
  D4D_INDEX_DELTA error =  (D4D_INDEX_DELTA)(-1 * (D4D_INDEX_DELTA)radius);
  D4D_COOR x = radius;
  D4D_COOR y = 0;
  D4D_COOR x1_bf,y1_bf;
  D4D_BOOL match = D4D_FALSE;    
 
  line = (D4D_COOR)(radius - line);
 
  if(line == 0) return radius;
  
  while (x >= y)
  {    
    if ((x != 0) && (y != 0)) //Valid point
    {
    	if(y == line) return (D4D_COOR)(radius - x);
    	
    	if(x == line)
    	{
    		match = D4D_TRUE;
    		x1_bf = (Byte)x;
    		y1_bf = (Byte)y;
    	}
      
      if(match  && (x != x1_bf)) 
        return (D4D_COOR)(radius - y1_bf);    	
    }
    
    error += y;
    ++y;
    error += y;

    if (error >= 0)
    {
      if((radius - x) >= line)
        return (D4D_COOR)(radius - y);
      
      x--;
      error -= x;
      error -= x;
    }
  }
  
  return (D4D_COOR)(radius - x);
}

/**************************************************************//*!
*
* Draw text on XY coordination in logic rectangle
*
******************************************************************/
void D4D_DrawTextRRectTab(D4D_POINT* ppt, D4D_SIZE* psz, D4D_STRING* buffText, D4D_TAB* pTab, D4D_COLOR colorText, D4D_COLOR colorBkgd, D4D_COOR radius)
{
  D4D_PRINT_STRING tmp_strDes;
  //D4D_SIZE tmp_size = *psz;
  D4D_TEXT_PROPERTIES tmp_txtProp;
  D4D_COOR tmp_txtWidth;
  D4D_COOR tmp_txtHeight;
  //D4D_POINT tmp_Pos;
  
  if(!radius)
  {
    D4D_DrawTextRectTab(ppt, psz, buffText, pTab, colorText, colorBkgd);
    return;
  }
  
  tmp_txtWidth = D4D_GetTextBuffWidthTabAdj(buffText, pTab);
  tmp_txtHeight = D4D_GetFontHeight(buffText->fontId);
    
  tmp_strDes.pText = buffText->pText;
  tmp_strDes.pFontType = D4D_GetFont(buffText->fontId);
  tmp_strDes.pTab = pTab;
  tmp_strDes.colorText = colorText;
  tmp_strDes.colorBack = colorBkgd;
  tmp_strDes.properties = buffText->str_properties->font_properties;
  tmp_strDes.textLength = buffText->printLen;
  tmp_strDes.textOffset = buffText->printOff;
        
  if(tmp_txtWidth + radius >= psz->cx)
  {
    if(psz->cx - radius > 0)
      tmp_txtWidth = (D4D_COOR)(psz->cx - radius);  
    else
      tmp_txtWidth = 0;
  }
  
  tmp_strDes.maxWidth = tmp_txtWidth;
    
  tmp_txtProp = buffText->str_properties->text_properties;

  tmp_strDes.x = ppt->x;
  tmp_strDes.y = ppt->y;
   
  // Compute the text position in rectangle in Y axis
  switch(tmp_txtProp & D4D_ALIGN_V_MASK)
  {
    case D4D_ALIGN_V_BOTTOM_MASK:
      tmp_strDes.y += (D4D_COOR)(psz->cy - tmp_txtHeight - radius / 2);
      break;
    
    case D4D_ALIGN_V_CENTER_MASK:
      tmp_strDes.y += (D4D_COOR)((psz->cy - tmp_txtHeight) / 2);
      break;
    
    default:
    case D4D_ALIGN_V_TOP_MASK:
      tmp_strDes.y += (D4D_COOR)(radius / 2);
      break;        
  }
  
  // Compute the text position in rectangle in X axis
  switch(tmp_txtProp & D4D_ALIGN_H_MASK)
  {
    case D4D_ALIGN_H_RIGHT_MASK:
      tmp_strDes.x += (D4D_COOR)(psz->cx - tmp_txtWidth - radius / 2);
      break;
    
    case D4D_ALIGN_H_CENTER_MASK:
      tmp_strDes.x += (D4D_COOR) ((psz->cx - tmp_txtWidth) / 2);
      break;
    
    default:
    case D4D_ALIGN_H_LEFT_MASK:
      tmp_strDes.x += (D4D_COOR)(radius / 2);
      break;        
  }
  
  // Redraw the rest of rectangle where isn't text for normal (non trasparent) texts
  if(!(tmp_strDes.properties & D4D_FNT_PRTY_TRANSPARENT_MASK))
  {
    D4D_FillRRect(ppt, psz, colorBkgd, radius);
#if 0
    
    
    tmp_Pos = *ppt;
    
    tmp_Pos.x += radius;
    tmp_Pos.y += radius;
    D4D_FillQuadrant(&tmp_Pos, radius, colorBkgd, D4D_QUAD_4);    
    tmp_Pos.x += psz->cx - radius * 2 - 1;
    D4D_FillQuadrant(&tmp_Pos, radius, colorBkgd, D4D_QUAD_1);
    tmp_Pos.y += psz->cy - radius * 2 - 1;
    D4D_FillQuadrant(&tmp_Pos, radius, colorBkgd, D4D_QUAD_2);
    tmp_Pos.x -= (psz->cx - radius * 2 - 1);
    D4D_FillQuadrant(&tmp_Pos, radius, colorBkgd, D4D_QUAD_3);
    
    
    if((tmp_txtProp & D4D_ALIGN_H_MASK) != D4D_ALIGN_H_LEFT_MASK)
    {
      // Fill the part of rect in front of the text in full height of rect
        D4D_FillRectXY(ppt->x, (D4D_COOR)(ppt->y + radius), tmp_strDes.x, (D4D_COOR)(ppt->y + psz->cy - 1 - radius), colorBkgd);
    }
    
    if((tmp_txtProp & D4D_ALIGN_H_MASK) != D4D_ALIGN_H_RIGHT_MASK)
    {
      // Fill the part of rect behind the text in full height of rect        if()
        D4D_FillRectXY((D4D_COOR)(tmp_strDes.x + tmp_txtWidth - 1), (D4D_COOR)(ppt->y), (D4D_COOR)(ppt->x + psz->cx - 1 - radius), (D4D_COOR)(ppt->y + psz->cy - 1), colorBkgd);
    }
      
    if((tmp_txtProp & D4D_ALIGN_V_MASK) != D4D_ALIGN_V_TOP_MASK)
    {
      // Fill the part of rect above the text in lenght of text
      D4D_FillRectXY((D4D_COOR)(ppt->x + radius), ppt->y, (D4D_COOR)(ppt->x + psz->cx - 1 - radius), (D4D_COOR)(tmp_strDes.y + 1), colorBkgd);
    }
    
    if((tmp_txtProp & D4D_ALIGN_V_MASK) != D4D_ALIGN_V_BOTTOM_MASK)
    {
      // Fill the part of rect under the text in lenght of text
      D4D_FillRectXY((D4D_COOR)(ppt->x + radius), (D4D_COOR)(tmp_strDes.y + tmp_txtHeight - 1), (D4D_COOR)(ppt->x + psz->cx - 1 - radius), (D4D_COOR)(ppt->y + psz->cy - 1), colorBkgd);
    }
#endif

  }
  
  // Print the string itself
  D4D_LCD_PrintStr(&tmp_strDes);
}
/**************************************************************//*!
*
* Draw text on XY coordination in logic rectangle
*
******************************************************************/
void D4D_DrawTextRRectTabTo(D4D_POINT* ppt, D4D_STRING* buffText, D4D_TAB* pTab, D4D_COLOR colorText, D4D_COLOR colorBkgd, D4D_COOR radius)
{
  D4D_SIZE tmp_size;
  
  tmp_size.cx = (D4D_COOR)(ppt->x - d4d_curXY.x);
  tmp_size.cy = (D4D_COOR)(ppt->y - d4d_curXY.y);
  
  D4D_DrawTextRRectTab( &d4d_curXY, &tmp_size, buffText, pTab, colorText, colorBkgd, radius);
}

/**************************************************************//*!
*
* Draw text on XY coordination in logic rectangle
*
******************************************************************/
void D4D_DrawTextRRectTabXY(D4D_COOR x1, D4D_COOR y1, D4D_COOR x2, D4D_COOR y2, D4D_STRING* buffText, D4D_TAB* pTab, D4D_COLOR colorText, D4D_COLOR colorBkgd, D4D_COOR radius)
{
  D4D_POINT tmp_point;
  D4D_SIZE tmp_size;
  
  tmp_point.x = x1;
  tmp_point.y = y1;
  
  tmp_size.cx = (D4D_COOR)(x2 - x1);
  tmp_size.cy = (D4D_COOR)(y2 - y1);
  
  
  D4D_DrawTextRRectTab( &tmp_point, &tmp_size, buffText, pTab, colorText, colorBkgd, radius);
}

/**************************************************************//*!
*
* Draw text on XY coordination in logic rectangle
*
******************************************************************/
void D4D_DrawTextRRectTabToXY(D4D_COOR x, D4D_COOR y, D4D_STRING* buffText, D4D_TAB* pTab, D4D_COLOR colorText, D4D_COLOR colorBkgd, D4D_COOR radius)
{
  D4D_SIZE tmp_size;
  
  tmp_size.cx = (D4D_COOR)(x - d4d_curXY.x);
  tmp_size.cy = (D4D_COOR)(y - d4d_curXY.y);
  
  
  D4D_DrawTextRRectTab( &d4d_curXY, &tmp_size, buffText, pTab, colorText, colorBkgd, radius);
}
#endif

/**************************************************************//*!
*
* Draw the bevel of object
*
******************************************************************/
void D4D_Bevel(D4D_POINT* ppt, D4D_SIZE* psz, D4D_COLOR color, D4D_BEVEL type, D4D_COOR radius)
{
  D4D_COLOR fore, bck;
  D4D_POINT shPos;
  D4D_SIZE shSize;
  
  #if (D4D_ROUND_CORNER_ENABLE == D4D_FALSE) //if no Round corners
    D4D_UNUSED(radius)
  #endif
  
  switch(type)
  {
    case D4D_BEVEL_NONE:
      // Draw nothing :-)
      return;
    case D4D_BEVEL_RAISED:
      fore = D4D_ChangeColorIntesity(color, D4D_BEVEL_FORE_INTENSITY);
      bck = D4D_ChangeColorIntesity(color, D4D_BEVEL_BCKG_INTENSITY);
      break;
    case D4D_BEVEL_DENTED:
      fore = D4D_ChangeColorIntesity(color, D4D_BEVEL_BCKG_INTENSITY);
      bck = D4D_ChangeColorIntesity(color, D4D_BEVEL_FORE_INTENSITY);
      break;
    case D4D_BEVEL_RESERVED:
      // Draw nothing :-)
      return;  
    default:
      return;    
  }
  
  if(psz->cx < (D4D_BEVEL_WIDTH * 2))
    return;
  
  if(psz->cy < (D4D_BEVEL_WIDTH * 2))
    return;
  
  // Draw Bevel    
  shPos = *ppt;    
  shSize.cx = psz->cx;
  shSize.cy = D4D_BEVEL_WIDTH;
 
#if (D4D_ROUND_CORNER_ENABLE != D4D_FALSE) //if Round corners  
  shPos.x += radius;
  shPos.y += radius;
  D4D_Quadrant(&shPos, radius, D4D_LINE_THICK, fore, D4D_QUAD_4);    
  shPos.x += psz->cx - radius * 2 - 2;
  D4D_Quadrant(&shPos, radius, D4D_LINE_THICK, fore, D4D_QUAD_1);
  shPos.y += psz->cy - radius * 2 - 2;
  D4D_Quadrant(&shPos, radius, D4D_LINE_THICK, bck, D4D_QUAD_2);
  shPos.x -= (psz->cx - radius * 2 - 2);
  D4D_Quadrant(&shPos, radius, D4D_LINE_THICK, bck, D4D_QUAD_3);
  
  shPos = *ppt;
  shPos.x += radius;
  shSize.cx -= radius * 2;
#endif 
  
  D4D_FillRect(&shPos, &shSize, fore);
  
  shPos.y += (psz->cy - D4D_BEVEL_WIDTH);     
  D4D_FillRect(&shPos, &shSize, bck);
  
  shPos = *ppt; 
  
  // change pos
  shSize.cx = D4D_BEVEL_WIDTH;
  shSize.cy = psz->cy; 
  
#if (D4D_ROUND_CORNER_ENABLE != D4D_FALSE) //if Round corners
  shPos.y += radius;
  shSize.cy -= radius * 2; 
#endif  
  D4D_FillRect(&shPos, &shSize, fore);
  
  shPos.x += (psz->cx - D4D_BEVEL_WIDTH);
  
  D4D_FillRect(&shPos, &shSize, bck);

}


#ifdef __CWCC__
  #pragma mark -
  #pragma mark Help Functions
  #pragma mark -
#endif

/**************************************************************//*!
*
* Compute cross color between two basic color in 256 steps  
*
******************************************************************/
D4D_COLOR D4D_GetCrossColor(D4D_COLOR startColor, D4D_COLOR endColor, Byte value)
{
  Byte r, g, b;
  Byte c_s, c_e;
  
  //Red color computing
  c_s = D4D_COLOR_GET_R(startColor);
  c_e = D4D_COLOR_GET_R(endColor);
  
  if(c_s < c_e)
    r = (Byte)(c_s + D4D_MulDivUU8(value, (Byte)(c_e - c_s), 255));
  else
    r = (Byte)(c_s - D4D_MulDivUU8(value, (Byte)(c_s - c_e), 255));
  
  //Green color computing
  c_s = D4D_COLOR_GET_G(startColor);
  c_e = D4D_COLOR_GET_G(endColor);
  
  if(c_s < c_e)
    g = (Byte)(c_s + D4D_MulDivUU8(value, (Byte)(c_e - c_s), 255));
  else
    g = (Byte)(c_s - D4D_MulDivUU8(value, (Byte)(c_s - c_e), 255));
  
  //Blue color computing
  c_s = D4D_COLOR_GET_B(startColor);
  c_e = D4D_COLOR_GET_B(endColor);
  
  if(c_s < c_e)
    b = (Byte)(c_s + D4D_MulDivUU8(value, (Byte)(c_e - c_s), 255));
  else
    b = (Byte)(c_s - D4D_MulDivUU8(value, (Byte)(c_s - c_e), 255));
  
  return D4D_COLOR_RGB(r, g, b);
  
}

/**************************************************************//*!
*
* Compute cross color between two basic color in 256 steps  
*
******************************************************************/
D4D_COLOR D4D_ChangeColorIntesity(D4D_COLOR color, sByte intensity)
{
  sWord r, g, b;
  Byte absIntesity;
  
  if(intensity > 100)
    intensity = 100;
  
  if(intensity < -100)
    intensity = -100;

  absIntesity = D4D_Abs(intensity);
  
  
  r = D4D_MulDivUU8(absIntesity, D4D_COLOR_GET_R(color), 100);
  g = D4D_MulDivUU8(absIntesity, D4D_COLOR_GET_G(color), 100);
  b = D4D_MulDivUU8(absIntesity, D4D_COLOR_GET_B(color), 100);

  if(intensity > 0)
  {
    r += D4D_COLOR_GET_R(color);
    if(r > 255)
      r = 255;
    
    g += D4D_COLOR_GET_G(color);
    if(g > 255)
      g = 255;
    
    b += D4D_COLOR_GET_B(color);
    if(b > 255)
      b = 255;
  }else
  {
    r = (Byte)(D4D_COLOR_GET_R(color) - r);
    g = (Byte)(D4D_COLOR_GET_G(color) - g);
    b = (Byte)(D4D_COLOR_GET_B(color) - b);
  } 
  
  return D4D_COLOR_RGB(r, g, b);  
}


/**************************************************************//*!
*
* Compute the grayscale color  
*
******************************************************************/
D4D_COLOR D4D_GetGreyScale(D4D_COLOR color)
{
  return D4D_LCD_GetGreyScale(color);
}

/**************************************************************//*!
*
* Return the longer side of size structure
*
******************************************************************/
D4D_COOR D4D_GetLongerSide(D4D_SIZE* pSz)
{
  if(pSz->cx > pSz->cy)
    return pSz->cx;
  
  return pSz->cy;
}

void D4D_FlushOutput(void)
{
  D4D_LLD_LCD.D4DLCD_FlushBuffer(D4DLCD_FLSH_FORCE);
}
