/******************************************************************************
* 
* Copyright (c) 2010 Freescale Semiconductor;
* All Rights Reserved                       
*
*******************************************************************************
*
* THIS SOFTWARE IS PROVIDED BY FREESCALE "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL FREESCALE OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
***************************************************************************//*!
*
* @file      d4d_screen.c
*
* @author    b01119
* 
* @version   0.0.38.0
* 
* @date      Aug-22-2012
* 
* @brief     D4D driver screen functions c file 
*
******************************************************************************/

#include "d4d.h"
#include "common_files/d4d_private.h"
#include "common_files/d4d_lldapi.h"


static D4D_COOR D4D_GetScrHeaderSize(D4D_SCREEN* pScreen);
static D4D_COOR D4D_GetScrHeaderTitleOffset(D4D_SCREEN* pScreen);
static D4D_BOOL D4D_GetScrHeaderExitBtnCoor(D4D_SCREEN* pScreen, D4D_POINT* resultPos, D4D_SIZE* resultSize);
static void D4D_DrawScreenNC(D4D_SCREEN* pScreen, D4D_BOOL active);
static D4D_OBJECT* D4D_FindObject(D4D_SCREEN* pScreen, D4D_OBJECT_PTR pObject);
//static Byte D4D_GetObjectCount(D4D_SCREEN* pScreen);


D4D_MARGIN d4d_screenMarginDflt = 
{
  0,    //left
  0,    //top
  0,    //right
  0     //bottom  
};

/**************************************************************//*!
*
* Get active screen
*
******************************************************************/

D4D_SCREEN* D4D_GetActiveScreen(void)
{
    // no screen yet shown
    if(d4d_screenHistoryIndex <= 0)
        return NULL;
    
    return d4d_screenHistory[d4d_screenHistoryIndex - 1];
}

/**************************************************************//*!
*
* Change the active screen
*
******************************************************************/
static void D4D_ChangeScreen(D4D_SCREEN* pNewScreen, D4D_SCREEN* pOldScreen)
{    
    D4D_SCREEN_DATA* pData;
    D4D_MESSAGE tmp_msg;

    if(pNewScreen == NULL)
      return;
    
    if(pOldScreen == pNewScreen)
      return;
    
    tmp_msg.pScreen = pOldScreen;
    
    if(pOldScreen != NULL)
    {
      tmp_msg.nMsgId = D4D_MSG_KILLFOCUS;
      tmp_msg.pObject = pOldScreen->pData->focusedObject;
      D4D_SendMessage(&tmp_msg);
              
      // Draw NC screen area as an inactivate
      D4D_DrawScreenNC(pOldScreen, D4D_FALSE);
      
      // call de-activate event
      if(pOldScreen->OnDeactivate != NULL)
          pOldScreen->OnDeactivate();       
    }

    // invalidate the new screen (global complete redraw, not individual objects)
    D4D_InvalidateScreen(pNewScreen, D4D_TRUE);
    
    // if this is the first time activating 
    pData = pNewScreen->pData;  
    
    // Init the screen
    D4D_InitScreen(pNewScreen);
    
    // Send to the object Focus message
    tmp_msg.pScreen = pNewScreen;
    tmp_msg.nMsgId = D4D_MSG_SETFOCUS;
    tmp_msg.pObject = pData->focusedObject;
    D4D_SendMessage(&tmp_msg);          
  
    // inform the screen it has been activated
    if(pNewScreen->OnActivate)
        pNewScreen->OnActivate();
    
    D4D_ClearKeysBuffer();
  
  #ifdef D4D_LLD_TCH  
    d4d_LastTouchedObj = NULL;
  #endif
}


/**************************************************************//*!
*
* Enter new screen, remember the history if not only replacing
*
******************************************************************/
void D4D_ActivateScreen(D4D_SCREEN* pNewScreen, Byte bReplaceCurrent)
{    
  D4D_SCREEN* pOldScreen;

  // can not activate screen (no space left in history array)
  if(d4d_screenHistoryIndex >= D4D_SCREEN_HISTORY && !bReplaceCurrent)
      return ;
  
  if(!pNewScreen)
    return;
  
  // cancel key capture
  D4D_CaptureKeys(NULL);

  // deactivate current screen (if any)
  pOldScreen = D4D_GetActiveScreen();

  if(pOldScreen != NULL)
  {
      // replacing this screen in history?
      if(bReplaceCurrent)
          d4d_screenHistoryIndex--; // note that we are sure this is >0
  }

  // set the new screen as the active one
  d4d_screenHistory[d4d_screenHistoryIndex] = pNewScreen;
  d4d_screenHistoryIndex++;

  D4D_ChangeScreen(pNewScreen, pOldScreen);    
}

 

/**************************************************************//*!
*
* Return to previous screen recorded in the history
*
******************************************************************/
void D4D_EscapeScreen(void)
{
    D4D_SCREEN* pOldScreen;
    
    // can not escape current screen, it is the top one 
    if(d4d_screenHistoryIndex <= 1)
      return;
    
    // cancel key capture
    D4D_CaptureKeys(NULL);

    pOldScreen = D4D_GetActiveScreen();
    
    // pop the history stack
    if(d4d_screenHistoryIndex > 1)
      d4d_screenHistoryIndex--;
    
    D4D_ChangeScreen(D4D_GetActiveScreen(), pOldScreen);

}


/**************************************************************//*!
*
* Return to previous screen recorded in the history
*
******************************************************************/
void D4D_EscapeToBaseScreen(void)
{
    D4D_SCREEN* pOldScreen;

    // can not escape current screen, it is the top one 
    if(d4d_screenHistoryIndex == 0)
        return;
    
    // cancel key capture
    D4D_CaptureKeys(NULL);

    
    // I can be sure this is not NULL
    pOldScreen = D4D_GetActiveScreen();

    // pop the history stack
    d4d_screenHistoryIndex = 1;    
    
    D4D_ChangeScreen(D4D_GetActiveScreen(), pOldScreen);
}

/**************************************************************//*!
*
* Init screen and all objects on given screen
*
******************************************************************/

void D4D_InitScreen(D4D_SCREEN* pScreen)
{
  D4D_SCREEN_DATA* pData = pScreen->pData;
  D4D_OBJECT** pObj;
  D4D_MESSAGE localMsg;


  if(!pScreen)
    return;  

  // prepare message
  localMsg.pScreen = pScreen;
  localMsg.nMsgId = D4D_MSG_ONINIT;
  
  pObj = (D4D_OBJECT**) pScreen->pObjects;
  
  // init objects   
  while(*pObj != NULL)
  {  
    
    localMsg.pObject = *pObj;
    
    // initialize the pointers on this screen in all screen objects
    D4D_SetObjectScreenPointer(*pObj, pScreen);
    
    // send the ON INIT message
    D4D_SendMessageMask(&localMsg, 0, D4D_OBJECT_F_NOTINIT);    
    
    pObj++;
  }
  
  if(!(pData->flags & D4D_SCR_FINT_INITDONE))
  {
    pData->flags |= D4D_SCR_FINT_INITDONE;
    
    D4D_FocusNextObject(pScreen, D4D_TRUE);
    
    // user's screen initialization
    if(pScreen->OnInit)
        pScreen->OnInit();
    
    // Enable natively touch scren for screens if exit button is enabled
    if(pScreen->flags & D4D_SCR_F_EXIT)
      pData->flags |= D4D_SCR_FINT_TOUCHENABLE;
    
    
  }
  
  
}


/**************************************************************//*!
*
* Mark all object on screen as "redraw pennding"
*
******************************************************************/

void D4D_InvalidateScreen(D4D_SCREEN* pScreen, Byte bComplete)
{    
    if(!pScreen)
      return;

    pScreen->pData->flags |= D4D_SCR_FINT_CHECKOBJECTS;
    
    if(bComplete)
    {
        pScreen->pData->flags |= D4D_SCR_FINT_REDRAWC;
    }
    else
    {
      D4D_OBJECT** pObj = (D4D_OBJECT**)pScreen->pObjects;
      while(*pObj != NULL)
      {
        D4D_SetObjectFlags(*pObj,  D4D_OBJECT_F_REDRAW);
        pObj++;
      }
    }
}

/**************************************************************//*!
*
* Gety the count of object in given screen
*
******************************************************************/
/*
static Byte D4D_GetObjectCount(D4D_SCREEN* pScreen)
{
    Byte ix = 0;
    
    if(!pScreen)
      return 0;
    
    while(pScreen->pObjects[ix] != NULL)
        ix++;
    
    return ix;
}
*/
/**************************************************************//*!
*
* Gety the count of object in given screen
*
******************************************************************/

D4D_OBJECT* D4D_GetLastObject(D4D_SCREEN* pScreen)
{
    D4D_OBJECT** pObject;
    
    if(!pScreen)
      return NULL;
    
    pObject = (D4D_OBJECT**)pScreen->pObjects;
    
    while(*(pObject + 1))
      pObject++;
    
    return *pObject;
}


/**************************************************************//*!
*
* Get focused object in a given screen
*
******************************************************************/

D4D_OBJECT_PTR D4D_GetFocusedObject(D4D_SCREEN* pScreen)
{
  if(!pScreen)
    return NULL;

  // just return the object on a focus index
  return pScreen->pData->focusedObject;
}





 /**************************************************************//*!
*
* Focus next object in given screen
*
******************************************************************/

void D4D_FocusNextObject(D4D_SCREEN* pScreen, D4D_BOOL bInitialSearch)
{
    D4D_SCREEN_DATA* pData = pScreen->pData;
    const D4D_OBJECT* const* pObjects = pScreen->pObjects;
    D4D_OBJECT* pFocusedObj = pData->focusedObject;
    
    if(!pScreen)
      return;
    
    // sanity check of list of objects - contains Screen any object?
    if(*pObjects == NULL)
      return;

    // currently focused object already has a tabstop    
    if(bInitialSearch)
    {
      //SetUp start object
      pData->focusedObject = (D4D_OBJECT*)pObjects[0];
      pFocusedObj = (D4D_OBJECT*)pObjects[0];
      if((pObjects[0]->pData->flags & (D4D_OBJECT_F_TABSTOP | D4D_OBJECT_F_ENABLED)) == (D4D_OBJECT_F_TABSTOP | D4D_OBJECT_F_ENABLED))
        return ;
    }
    
    do
    {
        
        // get next object
        pFocusedObj = D4D_FindNextObject(pFocusedObj, (pFocusedObj->pData->flags & D4D_OBJECT_F_ENABLED));
               
        // object with focus enabled?
        if((pFocusedObj->pData->flags & (D4D_OBJECT_F_TABSTOP | D4D_OBJECT_F_ENABLED)) == (D4D_OBJECT_F_TABSTOP | D4D_OBJECT_F_ENABLED))
            break;
        
      // avoid endless loop if no focused object can be found
    }while(((D4D_OBJECT*)pFocusedObj) != pData->focusedObject);
    
    
    
    if(((D4D_OBJECT*)pFocusedObj) != pData->focusedObject)
    {        
      // invalidate object which is loosing focus
      D4D_InvalidateObject(pData->focusedObject, D4D_FALSE);

      
      // prepare message KILLFOCUS
      d4d_msg.pScreen = pScreen;
      d4d_msg.nMsgId = D4D_MSG_KILLFOCUS;
      d4d_msg.pObject = pData->focusedObject;
      D4D_SendMessage(&d4d_msg);
      
      
      // invalidate object which is getting focus
      D4D_InvalidateObject(((D4D_OBJECT*)pFocusedObj), D4D_FALSE);
      
      // move the focus
      pData->focusedObject = ((D4D_OBJECT*)pFocusedObj);
      
      // prepare message
      d4d_msg.pScreen = pScreen;
      d4d_msg.nMsgId = D4D_MSG_SETFOCUS;
      d4d_msg.pObject = ((D4D_OBJECT*)pFocusedObj);
      D4D_SendMessage(&d4d_msg);
    }
}

/**************************************************************//*!
*
* Focus previous object in given screen
*
******************************************************************/

void D4D_FocusPrevObject(D4D_SCREEN* pScreen)
{
    D4D_SCREEN_DATA* pData = pScreen->pData;
    const D4D_OBJECT* const* pObjects = pScreen->pObjects;
    D4D_OBJECT* pFocusedObj = pData->focusedObject;
    
    if(!pScreen)
      return;
    
    // sanity check of list of objects - contains Screen any object?
    if(*pObjects == NULL)
      return;
    
    do
    {
        
        // just get previous object
        pFocusedObj = D4D_FindPreviousObject(pFocusedObj, (pFocusedObj->pData->flags & D4D_OBJECT_F_ENABLED));
        
        // object with focus enabled?
        if((pFocusedObj->pData->flags & (D4D_OBJECT_F_TABSTOP | D4D_OBJECT_F_ENABLED)) == (D4D_OBJECT_F_TABSTOP | D4D_OBJECT_F_ENABLED))
            break;
        
      // avoid endless loop if no focused object can be found
    }while(pFocusedObj != pData->focusedObject);
    
    if(pFocusedObj != pData->focusedObject)
    {        
      // invalidate object which is loosing focus
      D4D_InvalidateObject(pData->focusedObject, D4D_FALSE);
      
      // prepare message KILLFOCUS
      d4d_msg.pScreen = pScreen;
      d4d_msg.nMsgId = D4D_MSG_KILLFOCUS;
      d4d_msg.pObject = pData->focusedObject;
      D4D_SendMessage(&d4d_msg);
      
      
      // invalidate object which is getting focus
      D4D_InvalidateObject(pFocusedObj, D4D_FALSE);
      
      // move the focus
      pData->focusedObject = pFocusedObj;
      
      // prepare message
      d4d_msg.pScreen = pScreen;
      d4d_msg.nMsgId = D4D_MSG_SETFOCUS;
      d4d_msg.pObject = pFocusedObj;
      D4D_SendMessage(&d4d_msg);
    }
}

/**************************************************************//*!
*
* Find object's index
*
******************************************************************/

static D4D_OBJECT* D4D_FindObject(D4D_SCREEN* pScreen, D4D_OBJECT_PTR pObject)
{
    D4D_OBJECT* pLocObject = pScreen->pData->focusedObject;
    
    if(!pScreen || !pLocObject)
      return NULL;
    
    do
    {
      // get next object
      pLocObject = D4D_FindNextObject(pLocObject, D4D_TRUE);
      
      if(pLocObject == pObject)
        return pLocObject;
            
    }while(pLocObject != pScreen->pData->focusedObject);
            
    // not found
    return NULL;
}

/**************************************************************//*!
*
* Focus selected object in given screen
*
******************************************************************/

void D4D_FocusSet(D4D_SCREEN* pScreen, D4D_OBJECT_PTR pObject)
{
    D4D_SCREEN_DATA* pData = pScreen->pData;
    D4D_OBJECT* pFocusedObj = pData->focusedObject;
    D4D_OBJECT* pNewFocus;
    
    if(pScreen == NULL)
      return;
    
    // check if object is really item of the current screen
    pNewFocus = D4D_FindObject(pScreen, pObject);
    if(pNewFocus == NULL)
      return;
    
    if(pNewFocus == pFocusedObj) // is selected object same as focused?
      return;
  
   if((pNewFocus->pData->flags & (D4D_OBJECT_F_TABSTOP | D4D_OBJECT_F_ENABLED)) != (D4D_OBJECT_F_TABSTOP | D4D_OBJECT_F_ENABLED)) // is this object selectable?
      return;
    
    if(!D4D_IsEnabled(pNewFocus))
      return;
    
    // invalidate object which is loosing focus
    D4D_InvalidateObject(pFocusedObj, D4D_FALSE);
    
    // prepare message
    d4d_msg.pScreen = pScreen;
    d4d_msg.nMsgId = D4D_MSG_KILLFOCUS;
    d4d_msg.pObject = pFocusedObj;
    D4D_SendMessage(&d4d_msg);
    
    // invalidate object which is getting focus
    D4D_InvalidateObject(pNewFocus, D4D_FALSE);
    
    // move the focus
    pData->focusedObject = pNewFocus;
    
    // prepare message
    d4d_msg.pScreen = pScreen;
    d4d_msg.nMsgId = D4D_MSG_SETFOCUS;
    d4d_msg.pObject = pNewFocus;
    D4D_SendMessage(&d4d_msg);
        
}


/**************************************************************//*!
*
* enable or diasble touch pad capatibilities for an SCREEN
*
******************************************************************/
#ifdef D4D_LLD_TCH
  void D4D_EnableScrTouchScreen(const D4D_SCREEN* pScr, Byte bEnable)
  {
    if(bEnable)  
      pScr->pData->flags |= D4D_SCR_FINT_TOUCHENABLE;    
    else
      pScr->pData->flags &= ~D4D_SCR_FINT_TOUCHENABLE;    
  }
#else
  void D4D_EnableScrTouchScreen(const D4D_SCREEN* pScr, Byte bEnable)
  {
    D4D_UNUSED(pScr);
    D4D_UNUSED(bEnable);
  }
#endif  


/**************************************************************//*!
*
* Set Screen text font properties
*
******************************************************************/

void D4D_SetScreenFontProperties(D4D_SCREEN* pScreen, D4D_FONT_PROPERTIES property) 
{
  if(pScreen->textBuff.str_properties->font_properties == property)
    return;     // There is no change needed
  
  pScreen->textBuff.str_properties->font_properties = property;
  D4D_InvalidateScreen(pScreen, D4D_FALSE);
}

/**************************************************************//*!
*
* Set Object text properties
*
******************************************************************/

void D4D_SetScreenTextProperties(D4D_SCREEN* pScreen, D4D_TEXT_PROPERTIES property) 
{
  if(pScreen->textBuff.str_properties->text_properties == property)
    return;     // There is no change needed
  
  pScreen->textBuff.str_properties->text_properties = property;
  D4D_InvalidateScreen(pScreen, D4D_FALSE);

}



/**************************************************************//*!
*
* Compute size of screen header 
*
******************************************************************/

static D4D_COOR D4D_GetScrHeaderSize(D4D_SCREEN* pScreen)
{
  D4D_COOR tmp_size = 0;
  
  if(pScreen == NULL)
    return 0;
  
  if((pScreen->flags & D4D_SCR_F_TITLEBAR) || pScreen->textBuff.pText)
  {
    tmp_size = (D4D_COOR)(D4D_SCR_TITLE_OFF_Y * 2 + D4D_GetFontHeight(pScreen->textBuff.fontId));
  }

  if(pScreen->pIcon != NULL)
    if(tmp_size < (D4D_GetBmpHeight(pScreen->pIcon) + D4D_SCR_TITLE_OFF_Y * 2 + 1))
      tmp_size = (D4D_COOR)(D4D_GetBmpHeight(pScreen->pIcon) + D4D_SCR_TITLE_OFF_Y * 2 + 1);          
  
  if(pScreen->flags & D4D_SCR_F_TITLEBAR) 
    if(tmp_size < D4D_SCR_HEADER_SIZE_MIN_SIZE)
      tmp_size = D4D_SCR_HEADER_SIZE_MIN_SIZE;   
        
  return tmp_size;  
}

/**************************************************************//*!
*
* Compute offset of screen header title
*
******************************************************************/

static D4D_COOR D4D_GetScrHeaderTitleOffset(D4D_SCREEN* pScreen)
{
  D4D_COOR tmp_off = D4D_SCR_TITLE_OFF_X;
  
  if(pScreen == NULL)
    return 0;
  
  if(pScreen->pIcon != NULL)    
    tmp_off = (D4D_COOR)(D4D_GetBmpWidth(pScreen->pIcon) + D4D_SCR_TITLE_OFF_X * 2);          
  
  #if D4D_ROUND_CORNER_ENABLE == D4D_TRUE
    tmp_off += (pScreen->radius * 2);
  #endif  
      
  return tmp_off;  
}


/**************************************************************//*!
*
* Compute position of exit button on screen header title
*
******************************************************************/

static D4D_BOOL D4D_GetScrHeaderExitBtnCoor(D4D_SCREEN* pScreen, D4D_POINT* resultPos, D4D_SIZE* resultSize)
{        
  D4D_COOR tmp_coor = D4D_GetScrHeaderSize(pScreen);
  
  resultPos->x = 0;
  resultPos->y = 0;
  
  resultSize->cx = 0;
  resultSize->cy = 0;
  
  if(pScreen == NULL)
    return D4D_FALSE;
  
  if((tmp_coor - D4D_SCR_TITLE_EXITBTN_OFFSET * 2) <= 0)
    return D4D_FALSE;

  tmp_coor -= D4D_SCR_TITLE_EXITBTN_OFFSET * 2;
  
  if(tmp_coor < D4D_SCR_TITLE_EXITBTN_MIN_SIZE)
    return D4D_FALSE;
  
  resultSize->cx = tmp_coor;
  resultSize->cy = tmp_coor;
   
  resultPos->x = (D4D_COOR)(pScreen->position.x + pScreen->size.cx - tmp_coor - D4D_SCR_TITLE_EXITBTN_OFFSET);
  resultPos->y = (D4D_COOR)(pScreen->position.y + D4D_SCR_TITLE_EXITBTN_OFFSET);
  
  return D4D_TRUE;
}

/**************************************************************//*!
*
* Redraw non client screen area - Header, background, title, etc.
*
******************************************************************/

static void D4D_DrawScreenNC(D4D_SCREEN* pScreen, D4D_BOOL active)
{
  D4D_CLR_SCHEME *pScheme = D4D_ScreenGetScheme(pScreen);
  D4D_COLOR clr;
  
#if D4D_ROUND_CORNER_ENABLE == D4D_TRUE  
  
  if(active)
  {    
    if(pScreen->flags & D4D_SCR_F_BCKG)
    {    
      D4D_FillRRect(&pScreen->position, &pScreen->size, pScheme->screen.desktop, pScreen->radius);
    }
    
    if(pScreen->flags & D4D_SCR_F_OUTLINE)
    {
      D4D_RRect(&(pScreen->position), &(pScreen->size), D4D_LINE_THIN, pScheme->screen.outline, pScreen->radius);
    }
  }

  if(pScreen->flags & D4D_SCR_F_TITLEBAR)
  {
    clr = pScheme->screen.title_bar;
    
    if(!active)
      clr = D4D_GetGreyScale(clr);    
    
    D4D_FillRRectXY((D4D_COOR)(pScreen->position.x + 1), (D4D_COOR)(pScreen->position.y + 1), (D4D_COOR)(pScreen->size.cx + pScreen->position.x - 1), (D4D_COOR)(D4D_GetScrHeaderSize(pScreen) + pScreen->position.y - 1), clr, pScreen->radius);    
  }
  
  if(pScreen->pIcon != NULL)
  {
    D4D_DrawRBmpXY((D4D_COOR)(pScreen->position.x + D4D_SCR_TITLE_OFF_X), (D4D_COOR)(pScreen->position.y + (D4D_GetScrHeaderSize(pScreen) - D4D_GetBmpHeight(pScreen->pIcon)) / 2), pScreen->pIcon, (D4D_BOOL)!active, pScreen->radius);  
  }
  
  if(pScreen->textBuff.pText)
  {        
    clr = pScheme->screen.title_text;
    
    if(!active)
      clr = D4D_GetGreyScale(clr);    
    
    D4D_DrawTextXY((D4D_COOR)(pScreen->position.x + D4D_GetScrHeaderTitleOffset(pScreen)), (D4D_COOR)(pScreen->position.y + (D4D_GetScrHeaderSize(pScreen) + pScreen->radius - D4D_GetFontHeight(pScreen->textBuff.fontId)) / 2), 
        &pScreen->textBuff, clr, 0);
    }
  
  if(pScreen->flags & D4D_SCR_F_EXIT)
  {
    D4D_POINT tmp_pos;
    D4D_SIZE tmp_size;
    clr = pScheme->screen.exitBtnFore;
    
    if(!active)
      clr = D4D_GetGreyScale(clr);
    
    if(D4D_GetScrHeaderExitBtnCoor(pScreen, &tmp_pos, &tmp_size))
    {         
      D4D_RBox(&tmp_pos, &tmp_size, D4D_LINE_THIN, clr, pScheme->screen.exitBtnBckg, pScreen->radius);
      
      tmp_pos.x += (D4D_COOR)(D4D_SCR_EXITBTN_CROSS_SIZE + pScreen->radius / 2);
      tmp_pos.y += (D4D_COOR)(D4D_SCR_EXITBTN_CROSS_SIZE + pScreen->radius / 2);    
      
      D4D_MoveTo(&tmp_pos);
      tmp_pos.x += tmp_size.cx - (2 * D4D_SCR_EXITBTN_CROSS_SIZE  + pScreen->radius);
      tmp_pos.y += tmp_size.cy - (2 * D4D_SCR_EXITBTN_CROSS_SIZE  + pScreen->radius);    
      D4D_LineTo(&tmp_pos, D4D_LINE_THIN, clr);
      
      tmp_pos.y -= tmp_size.cy - (2 * D4D_SCR_EXITBTN_CROSS_SIZE  + pScreen->radius);    
      
      D4D_MoveTo(&tmp_pos);
      tmp_pos.x -= tmp_size.cx - (2 * D4D_SCR_EXITBTN_CROSS_SIZE  + pScreen->radius);
      tmp_pos.y += tmp_size.cy - (2 * D4D_SCR_EXITBTN_CROSS_SIZE  + pScreen->radius);    
      D4D_LineTo(&tmp_pos, D4D_LINE_THIN, clr);
    }
  }
  
#else

  if(active)
  {    
    if(pScreen->flags & D4D_SCR_F_BCKG)
    {    
      D4D_FillRect(&pScreen->position, &pScreen->size, pScheme->screen.desktop);
    }
    
    if(pScreen->flags & D4D_SCR_F_OUTLINE)
    {
      D4D_Rect(&(pScreen->position), &(pScreen->size), D4D_LINE_THIN, pScheme->screen.outline);
    }
  }

  if(pScreen->flags & D4D_SCR_F_TITLEBAR)
  {
    clr = pScheme->screen.title_bar;
    
    if(!active)
      clr = D4D_GetGreyScale(clr);    
    
    D4D_FillRectXY((D4D_COOR)(pScreen->position.x + 1), (D4D_COOR)(pScreen->position.y + 1), (D4D_COOR)(pScreen->size.cx + pScreen->position.x - 1), (D4D_COOR)(D4D_GetScrHeaderSize(pScreen) + pScreen->position.y - 1), clr);    
  }
  
  if(pScreen->pIcon != NULL)
  {
    D4D_DrawBmpXY((D4D_COOR)(pScreen->position.x + D4D_SCR_TITLE_OFF_X), (D4D_COOR)(pScreen->position.y + (D4D_GetScrHeaderSize(pScreen) - D4D_GetBmpHeight(pScreen->pIcon)) / 2), pScreen->pIcon, (D4D_BOOL)!active);  
  }
  
  if(pScreen->textBuff.pText)
  {        
    clr = pScheme->screen.title_text;
    
    if(!active)
      clr = D4D_GetGreyScale(clr);    
    
    D4D_DrawTextXY((D4D_COOR)(pScreen->position.x + D4D_GetScrHeaderTitleOffset(pScreen)), (D4D_COOR)(pScreen->position.y + (D4D_GetScrHeaderSize(pScreen) - D4D_GetFontHeight(pScreen->textBuff.fontId)) / 2), 
        &pScreen->textBuff, clr, 0);
    }
  
  if(pScreen->flags & D4D_SCR_F_EXIT)
  {
    D4D_POINT tmp_pos;
    D4D_SIZE tmp_size;
    clr = pScheme->screen.exitBtnFore;
    
    if(!active)
      clr = D4D_GetGreyScale(clr);
    
    if(D4D_GetScrHeaderExitBtnCoor(pScreen, &tmp_pos, &tmp_size))
    {         
      D4D_Box(&tmp_pos, &tmp_size, D4D_LINE_THIN, clr, pScheme->screen.exitBtnBckg);
      
      tmp_pos.x += D4D_SCR_EXITBTN_CROSS_SIZE;
      tmp_pos.y += D4D_SCR_EXITBTN_CROSS_SIZE;    
      
      D4D_MoveTo(&tmp_pos);
      tmp_pos.x += tmp_size.cx - 2 * D4D_SCR_EXITBTN_CROSS_SIZE;
      tmp_pos.y += tmp_size.cy - 2 * D4D_SCR_EXITBTN_CROSS_SIZE;    
      D4D_LineTo(&tmp_pos, D4D_LINE_THIN, clr);
      
      tmp_pos.y -= tmp_size.cy - 2 * D4D_SCR_EXITBTN_CROSS_SIZE;    
      
      D4D_MoveTo(&tmp_pos);
      tmp_pos.x -= tmp_size.cx - 2 * D4D_SCR_EXITBTN_CROSS_SIZE;
      tmp_pos.y += tmp_size.cy - 2 * D4D_SCR_EXITBTN_CROSS_SIZE;    
      D4D_LineTo(&tmp_pos, D4D_LINE_THIN, clr);
    }
  }
#endif  
    
}
 
/**************************************************************//*!
*
* Redraw all objects that need it on given screen
*
******************************************************************/

static void D4D_RedrawScreenObject(D4D_OBJECT* pObject, D4D_BOOL complete)
{
  D4D_MESSAGE msg;
  
  if(!(pObject->pData->flags & D4D_OBJECT_F_VISIBLE))
    return;
  
  msg.pScreen = pObject->pData->pScreen;
  msg.pObject = pObject;
  msg.prm.draw = 0;    
  
  /* forced complete redraw of all objects or object wants to be completely redrawn*/
  if(complete || (pObject->pData->flags & D4D_OBJECT_F_REDRAWC))
    msg.prm.draw |= D4D_OBJECT_DRAWFLAGS_COMPLETE;
  
  if(msg.prm.draw || (pObject->pData->flags & D4D_OBJECT_F_REDRAW)) /* object does need to redraw */
  {
    // draw special?
    if((pObject == msg.pScreen->pData->focusedObject) && (pObject->pData->flags & D4D_OBJECT_F_TABSTOP))
    {
        msg.prm.draw |= D4D_OBJECT_DRAWFLAGS_FOCUSED;
  
        if(pObject == d4d_pKeysCapturer)
            msg.prm.draw |= D4D_OBJECT_DRAWFLAGS_CAPTURING;
    }
    
    // will draw now            
    pObject->pData->flags &= ~(D4D_OBJECT_F_REDRAW | D4D_OBJECT_F_REDRAWC);  
    // send the DRAW message
    msg.nMsgId = D4D_MSG_DRAW;
    D4D_SendMessage(&msg);
    // send the DRAWDONE message
    msg.nMsgId = D4D_MSG_DRAWDONE;
    D4D_SendMessage(&msg);
  }
           
  
  if(pObject->pRelations)
  {
    D4D_OBJECT** pChild = (D4D_OBJECT**)&(pObject->pRelations[D4D_OBJECT_USR_DATA_CHILD_IX]);
    
    while(*pChild)
    {
      D4D_RedrawScreenObject(*pChild, complete);
      pChild++;
    }    
  }      
}



void D4D_RedrawScreen(D4D_SCREEN* pScreen)
{
  D4D_SCREEN_DATA data = *pScreen->pData;
  D4D_OBJECT** pObj;
  
  pScreen->pData->flags &= ~(D4D_SCR_FINT_REDRAWC | D4D_SCR_FINT_CHECKOBJECTS);
  
  // prepare message
  d4d_msg.pScreen = pScreen;
  
  // redraw screen non client area if necessary
  if(data.flags & D4D_SCR_FINT_REDRAWC)
  {
    D4D_LLD_LCD.D4DLCD_FlushBuffer(D4DLCD_FLSH_SCR_START);
    
    d4d_msg.pObject = NULL;
    d4d_msg.nMsgId = D4D_MSG_DRAW;
    D4D_SendMessage(&d4d_msg);
    
    D4D_DrawScreenNC(pScreen, D4D_TRUE);    
  }
  
  if(data.flags & D4D_SCR_FINT_CHECKOBJECTS)
  {
    // redraw objects   
    pObj = (D4D_OBJECT**)pScreen->pObjects;
    // Go through the all top level objects
    while(*pObj != NULL)
    {
      // redraw all top level objects includes it children
      D4D_RedrawScreenObject(*pObj, data.flags & D4D_SCR_FINT_REDRAWC);
      pObj++;          
    }
    
    d4d_msg.pObject = NULL;
    d4d_msg.nMsgId = D4D_MSG_DRAWDONE;
    D4D_SendMessage(&d4d_msg);
    
    if(!(data.flags & D4D_SCR_FINT_REDRAWC))
      D4D_LLD_LCD.D4DLCD_FlushBuffer(D4DLCD_FLSH_FORCE);
  }
  
  if(data.flags & D4D_SCR_FINT_REDRAWC)
    D4D_LLD_LCD.D4DLCD_FlushBuffer(D4DLCD_FLSH_SCR_END);
}

/**************************************************************//*!
*
* Handle the timeticks events of eGUI
*
******************************************************************/
void D4D_HandleTimeTicks(D4D_SCREEN* pScreen)
{
  if(d4d_systemFlags & D4D_SYSTEM_F_TIMETICK)
  {
    D4D_OBJECT** pObjects = (D4D_OBJECT**)pScreen->pObjects;
    d4d_systemFlags &= ~D4D_SYSTEM_F_TIMETICK;
    
    d4d_msg.pScreen = pScreen;
    d4d_msg.nMsgId = D4D_MSG_TIMETICK;
    
    // Go through the all top level objects
    while(*pObjects != NULL)
    {
      d4d_msg.pObject = *pObjects;
      // send timeticks to all top level objects includes it children
      D4D_SendMessageMask(&d4d_msg, D4D_OBJECT_F_VISIBLE , D4D_OBJECT_F_TIMETICK);
      pObjects++;          
    }
    
  }
}

/**************************************************************//*!
*
* Convert Client point into the global screen point
*
******************************************************************/

D4D_POINT D4D_GetClientToScreenPoint(D4D_OBJECT* pObject, D4D_POINT* nClientPoint)
{
  D4D_POINT tmp_point;
  D4D_SCREEN* pScreen = pObject->pData->pScreen;
  
  
  if(pScreen == NULL)
  {
    return *nClientPoint;  
  }
  
  tmp_point.x = (D4D_COOR)(pScreen->position.x + nClientPoint->x);
  tmp_point.y = (D4D_COOR)(pScreen->position.y + nClientPoint->y);
  
  if(pScreen->flags & D4D_SCR_F_OUTLINE)
  	tmp_point.x++;	
  
  tmp_point.y += D4D_GetScrHeaderSize(pScreen); 
  
  if(pObject->pRelations)
  {
    D4D_OBJECT* pParent = (D4D_OBJECT*)pObject->pRelations[D4D_OBJECT_USR_DATA_PARENT_IX];
  
    while(pParent)
    {
      tmp_point.x += pParent->position.x;
      tmp_point.y += pParent->position.y;
      
      if(pParent->pMargin)
      {
        tmp_point.x += pParent->pMargin->left;
        tmp_point.y += pParent->pMargin->top;        
      }
      
      pParent = (D4D_OBJECT*)pParent->pRelations[D4D_OBJECT_USR_DATA_PARENT_IX];
    }
  }
  
  return tmp_point;
}

/**************************************************************//*!
*
* Convert global screen point into the client point
*
******************************************************************/

D4D_POINT D4D_GetScreenToClientPoint(D4D_OBJECT* pObject, D4D_POINT* nScreenPoint)
{
  D4D_POINT tmp_point;
  D4D_SCREEN* pScreen = pObject->pData->pScreen;
  
  if(pScreen == NULL)
  {
    return *nScreenPoint;  
  }
  
  tmp_point.x = (D4D_COOR)(nScreenPoint->x - pScreen->position.x);
  tmp_point.y = (D4D_COOR)(nScreenPoint->y - pScreen->position.y);
  
  if(pScreen->flags & D4D_SCR_F_OUTLINE)
  	tmp_point.x--;
  
  tmp_point.y -= D4D_GetScrHeaderSize(pScreen); 
  
  if(pObject->pRelations)
  {
    D4D_OBJECT* pParent = (D4D_OBJECT*)pObject->pRelations[D4D_OBJECT_USR_DATA_PARENT_IX];
  
    while(pParent)
    {
      tmp_point.x -= pParent->position.x;
      tmp_point.y -= pParent->position.y;
      if(pParent->pMargin)
      {
        tmp_point.x -= pParent->pMargin->left;
        tmp_point.y -= pParent->pMargin->top;        
      }
      pParent = (D4D_OBJECT*)pParent->pRelations[D4D_OBJECT_USR_DATA_PARENT_IX];
    }
  }
  
  
  return tmp_point;
}

/**************************************************************//*!
*
* Get real client screen size
*
******************************************************************/

D4D_SIZE D4D_GetClientScreenSize(D4D_OBJECT* pObject)
{
  D4D_SIZE tmp_size = {0, 0};
  D4D_SCREEN* pScreen = pObject->pData->pScreen;
  
  
  if(pObject->pRelations)
  {
    D4D_OBJECT* pParent = (D4D_OBJECT*)pObject->pRelations[D4D_OBJECT_USR_DATA_PARENT_IX];
  
    if(pParent)
    {
      tmp_size = pParent->size;
      if(pParent->pMargin)
      {
        tmp_size.cx -= (D4D_COOR)(pParent->pMargin->left + pParent->pMargin->right);
        tmp_size.cy -= (D4D_COOR)(pParent->pMargin->top + pParent->pMargin->bottom);        
      }
      return tmp_size;
    }
  }
  
  if(pScreen == NULL)
  {
    return tmp_size;  
  }
  
  tmp_size.cx = pScreen->size.cx;
  tmp_size.cy = pScreen->size.cy;
  
  if(pScreen->flags & D4D_SCR_F_OUTLINE)
  {
    tmp_size.cx -= 2;
    tmp_size.cy -= 2;
  }
  
  tmp_size.cy -= D4D_GetScrHeaderSize(pScreen); 
        
  return tmp_size;
}


/**************************************************************//*!
*
* Check if the coordination are in range of exit button
*
******************************************************************/

D4D_BOOL D4D_ScrCheckExitBtnCoor(D4D_SCREEN* pScreen, D4D_POINT* point)
{
  D4D_POINT tmp_pos;
  D4D_SIZE tmp_size;
  
  
  if(D4D_GetScrHeaderExitBtnCoor(pScreen, &tmp_pos, &tmp_size))
  {    
    if((point->x >= tmp_pos.x) && (point->y >= tmp_pos.y))
    { 
      if((point->x <= (tmp_pos.x + tmp_size.cx)) && (point->y <= (tmp_pos.y + tmp_size.cy)))
        return D4D_TRUE;
    }
  }
  return D4D_FALSE;
}

/**************************************************************//*!
*
* Check if the coordination are in range of screen
*
******************************************************************/

D4D_BOOL D4D_ScrCheckCoor(D4D_SCREEN* pScreen, D4D_POINT* point)
{
  if((point->x >= pScreen->position.x) && (point->y >= (pScreen->position.y + D4D_GetScrHeaderSize(pScreen))))
  { 
    if((point->x <= (pScreen->position.x + pScreen->size.cx)) && (point->y <= (pScreen->position.y + pScreen->size.cy)))
      return D4D_TRUE;
  }
  return D4D_FALSE;
}