/**************************************************************************
*
* $FileName: wave_task.c$
* $Version : 0.0.2.0$
* $Date    : Feb-9-2012$
*
* Comments: Provide wave files processing on external SD card.
*
*END************************************************************************/


#include <string.h>
#include <mqx.h>

#include <bsp.h>
#include <fio.h>
#include <event.h>
#include <lwmsgq.h>
#include "sai.h"
#include "sgtl5000.h"

#include "wave_task.h"


/******************************************************************************
* Local functions prototype                                              *
******************************************************************************/
static void WAVE_Init(void);
static void WAVE_Stop(signed short* pBufferLast);
static void WAVE_Play(char* pFileName);
static void WAVE_Record(char* pFileName);
static void WAVE_Pause(void);
static void WAVE_Restore(void);
static void WAVE_KeepPlaying(short* pOutBuff);
static void WAVE_StoreRecord(short* pInBuff);

static void WAVE_RequestFillUpOutput(void);
static void WAVE_ChangeState(WAVE_MSG_STATE new_state);

/********************************************************************/
/* State machine                                                    */
/********************************************************************/
static WAVE_MSG_STATE wave_state = WAVEST_NOTINIT;

/********************************************************************/
/* End of state machine						    */
/********************************************************************/

static char* playerStateStr[] = 
{
  "Not Init",
  "Ready",
  "Recording",
  "Playing",
  "Playing hold",
  "Pause"  
};

static uint_32 songDuration = 0;

static uint_8 first_time = 1;

// wave player variables
static MQX_FILE_PTR  pFileSong;
static uint_32 nRead; 


WAV_ID wav_id;
#define WAV_BUFFER_SIZE 1024
#if WAV_BUFFER_SIZE > SAI_BUFFER_SIZE
  #error  WAV_BUFFER_SIZE Cannot bigger than SAI_BUFFER_SIZE!!
#endif    

//MQX Messages
uint_32 wave_queue[sizeof(LWMSGQ_STRUCT)/sizeof(uint_32) + NUM_MESSAGES * MSG_SIZE];

#if WAVE_TASK_BUFFER_REC_MODE
  uint_16 * pRecordBuffer = NULL;
  uint_16 * pRecordBufferWork = NULL;
#endif


//-----------------------------------------------------------------------------
// FUNCTION:    WAVE_task
// SCOPE:       Wave Player  public related function
// DESCRIPTION: This is the main task function for Wave Player
//              
// PARAMETERS:  temp - initialization data - not used
//              
// RETURNS:     none
//-----------------------------------------------------------------------------  
void WAVE_task(unsigned long temp) {
  WAVE_MSG message;

  // Init the main message handling loop
  if(_lwmsgq_init((pointer)wave_queue, NUM_MESSAGES, MSG_SIZE) == MQX_OK)
  {  
    // force as first message Init decoder
    message.type = WAVE_INIT;
    _lwmsgq_send((pointer)wave_queue, (uint_32*)&message, 0);
 
    while(1)
    {
      // Read meesage from buffer
      if(_lwmsgq_receive((pointer)wave_queue, (uint_32*)&message, LWMSGQ_RECEIVE_BLOCK_ON_EMPTY, 0, 0) == MQX_OK)
      {      
        // And do action
        switch(message.type)
        {
        //Play song
        case  WAVE_PLAY:
          if(wave_state == WAVEST_PAUSE)
            WAVE_Restore();
          else if(wave_state == WAVEST_READY)
            WAVE_Play((char*)message.pointer);        
          break;
        
        //Record song
        case  WAVE_REC:
          if(wave_state == WAVEST_PAUSE)
            WAVE_Restore();
          else if(wave_state == WAVEST_READY)
            WAVE_Record((char*)message.pointer);        
          break;
          
        // Pause playing  
        case  WAVE_PAUSE:
          if(wave_state == WAVEST_PLAYING)
            WAVE_Pause();  
          else if(wave_state == WAVEST_PAUSE)
            WAVE_Restore(); 
          break;
        
        // Stop playing  
        case  WAVE_STOP:
          if(wave_state > WAVEST_READY)
            WAVE_Stop(NULL);        
          break;
        
        // Init decoder  
        case  WAVE_INIT:
          if(wave_state == WAVEST_NOTINIT)
            WAVE_Init();  
          break;
        
        // deinit decoder  
        case  WAVE_DEINIT:      
          if(wave_state > WAVEST_READY)
          {
            // stop the song
            WAVE_Stop(NULL);
          }
          
          WAVE_ChangeState(WAVEST_NOTINIT);
          printf("Wave player is unallocated.");
          break;
          
        // Top up the audio aoutput buffers  
        case  WAVE_NEEDDATA:
          if(wave_state > WAVEST_READY)
            WAVE_KeepPlaying((short*)message.pointer);
          break;
        
        case  WAVE_HAVEDATA:
          if(wave_state > WAVEST_READY)
            WAVE_StoreRecord((short*)message.pointer);
          break;
          
        default:
          // do nothing :-)
          break;
        }        
      }
    }
  }
  
  printf("The Wave player task failed");
}

//-----------------------------------------------------------------------------
// FUNCTION:    WAVE_GetState
// SCOPE:       Wave Player public related function
// DESCRIPTION: Functions return state of Wave Player
//              
// PARAMETERS:  none
//              
// RETURNS:     Wave Player state
//-----------------------------------------------------------------------------  
WAVE_MSG_STATE WAVE_GetState(void)
{
  return wave_state;
}

//-----------------------------------------------------------------------------
// FUNCTION:    WAVE_ChangeState
// SCOPE:       Wave Player private related function
// DESCRIPTION: Functions managing requests to change the state of Wave Player
//              And also ntify the Parent PEG application about this change
//              
// PARAMETERS:  new_state - new requested state
//              
// RETURNS:     none
//----------------------------------------------------------------------------- 
static void WAVE_ChangeState(WAVE_MSG_STATE new_state)
{  
  if(new_state == wave_state)
    return;
  
  wave_state = new_state;  
  printf("Wave player change state to: %s. \n", playerStateStr[new_state]);
  
}


//-----------------------------------------------------------------------------
// FUNCTION:    WAVE_Stop
// SCOPE:       Wave Player private related function
// DESCRIPTION: Function stop the playing of song
//              
// PARAMETERS:  pBufferLast - NULL - for immediate stop the song
//                            pointer on last buffer to play out the song          
//              
// RETURNS:     none
//----------------------------------------------------------------------------- 
static void WAVE_Stop(signed short* pBufferLast)
{
  switch(SAI_GetState())
  {
    case SAI_ST_TX:
      // Disable Audio HW IRQ
      if(pBufferLast)
        SAI_LastBlockLoaded(pBufferLast);
      else  
        SAI_DisableTx();  
      break;
          
    case SAI_ST_RX:
    {  
      uint_32 size;
      // Disable Audio HW IRQ
      SAI_DisableRx();  
      
      #if WAVE_TASK_BUFFER_REC_MODE
        //Write the buffered data to the file
        size = pRecordBufferWork - pRecordBuffer;
        if(fwrite(pRecordBuffer, 1, size, pFileSong) != size)
          printf("Error with writing wave file.\n");      
      #else
        
      #endif            
      size = pFileSong->SIZE - 8;
      fseek(pFileSong, 4, IO_SEEK_SET);
      if(fwrite(&size, 1, 4, pFileSong) != 4)
        printf("Error with updating ChunkSize.\n");      
      break;
    }
  }
  // close the file
  if(fclose(pFileSong) < 0)
    printf("Error: File closing failed\n");
  
  pFileSong = NULL;
  
  WAVE_ChangeState(WAVEST_READY);
}

//-----------------------------------------------------------------------------
// FUNCTION:    WAVE_Play
// SCOPE:       Wave Player private related function
// DESCRIPTION: Function start playing of song
//              
// PARAMETERS:  pFileName - pointer on file name (complete path + name + extension)          
//              
// RETURNS:     none
//----------------------------------------------------------------------------- 
static void WAVE_Play(char* pFileName)
{  
  
  songDuration = WAVE_GetSongDuration(pFileName);
  
  if(!songDuration)
  {
    printf("Error: File %s has zero length \n", pFileName);
    return;
  }
  
  pFileSong = fopen(pFileName, "rb");
  if (pFileSong == NULL)
  {
    printf("Error: File %s not found \n", pFileName);
    return;
  }
  
  first_time = 1;
        
  WAVE_RequestFillUpOutput();

  printf("Start playing %s\n", pFileName);  
  
  WAVE_ChangeState(WAVEST_PLAYING_HOLD);
}

//-----------------------------------------------------------------------------
// FUNCTION:    WAVE_Record
// SCOPE:       Wave Player private related function
// DESCRIPTION: Function start recording of sound
//              
// PARAMETERS:  pFileName - pointer on file name (complete path + name + extension)          
//              
// RETURNS:     none
//----------------------------------------------------------------------------- 
static void WAVE_Record(char* pFileName)
{
  WAV_ID wav_id;
  
  pFileSong = fopen(pFileName, "wb");
  
  if (pFileSong == NULL)
  {
    printf("Error: File %s is not created \n", pFileName);
    return;
  }
  
  
  wav_id.ChunkID[0] = 'R';wav_id.ChunkID[1] = 'I';wav_id.ChunkID[2] = 'F';wav_id.ChunkID[3] = 'F';
  wav_id.ChunkSize = 0;  // file size - 8 
  wav_id.Format[0] = 'W';wav_id.Format[1] = 'A';wav_id.Format[2] = 'V';wav_id.Format[3] = 'E'; 
  wav_id.Subchunk1ID[0] = 'f';wav_id.Subchunk1ID[1] = 'm';wav_id.Subchunk1ID[2] = 't';wav_id.Subchunk1ID[3] = ' ';  
  wav_id.Subchunk1Size = 18; 
  wav_id.AudioFormat = 1; 
  wav_id.NumChannels = 2; 
  wav_id.SampleRate = 44100; 
  wav_id.ByteRate = 176400;
  wav_id.BlockAlign = 4;   
  wav_id.BitsPerSample = 16; 
  wav_id.Subchunk2ID[0] = '\0';wav_id.Subchunk2ID[1] = '\0';wav_id.Subchunk2ID[2] = 'd';wav_id.Subchunk2ID[3] = 'a';  
  wav_id.Subchunk2Size = 2885706100; 
        
        
  if(fwrite((uint_8*)&wav_id, 1, sizeof(WAV_ID), pFileSong) != sizeof(WAV_ID)) // Write the wave data
  {
    printf("Error - Write File. Record Stops");
    WAVE_Stop(NULL);
    return;
  }          
  
  
#if WAVE_TASK_BUFFER_REC_MODE
  pRecordBufferWork = pRecordBuffer;
#else
        
#endif            
          
  printf("Start recording %s\n", pFileName);  
  
  SAI_EnableRx();
            
  WAVE_ChangeState(WAVEST_RECORDING);
}

//-----------------------------------------------------------------------------
// FUNCTION:    WAVE_Pause
// SCOPE:       Wave Player private related function
// DESCRIPTION: Function pause playing of song
//              
// PARAMETERS:  none          
//              
// RETURNS:     none
//----------------------------------------------------------------------------- 
static void WAVE_Pause(void)
{   
  switch(SAI_GetState())
  {
    case SAI_ST_TX:
      SAI_PauseTx();
      break;
          
    case SAI_ST_RX:
      SAI_PauseRx();
      break;
  }  
  
  WAVE_ChangeState(WAVEST_PAUSE);
  printf("\nPause\n");  
}

//-----------------------------------------------------------------------------
// FUNCTION:    WAVE_Restore
// SCOPE:       Wave Player private related function
// DESCRIPTION: Function restore playing of song
//              
// PARAMETERS:  none          
//              
// RETURNS:     none
//-----------------------------------------------------------------------------
static void WAVE_Restore(void)
{
  switch(SAI_GetState())
  {
    case SAI_ST_TX:
      SAI_ResumeTx();   
      WAVE_ChangeState(WAVEST_PLAYING);
      break;
          
    case SAI_ST_RX:
      SAI_ResumeRx();   
      WAVE_ChangeState(WAVEST_RECORDING);
      break;
  }
  
}

//-----------------------------------------------------------------------------
// FUNCTION:    WAVE_GetSongDuration
// SCOPE:       Wave Player public related function
// DESCRIPTION: Functions compute song duration
//              
// PARAMETERS:  pFileName - song file name (with full path)
//              
// RETURNS:     song duration (0 could means error)
//----------------------------------------------------------------------------- 
uint_32 WAVE_GetSongDuration(char* pFileName)
{
  MQX_FILE_PTR  pFileWave = NULL;

  uint_32  seconds = 0;
  uint_8 *pReadBuff = NULL;
   
  #define READ_BUFF_SIZE sizeof(WAV_ID)
  
  pFileWave = fopen(pFileName, "rb");
  if (pFileWave == NULL)
  {
    printf("Error: File %s not found \n", pFileName);
    return 0;
  }
  
  pReadBuff = (uint_8*)_mem_alloc(READ_BUFF_SIZE);
  if(pReadBuff == NULL)
  {   
    fclose(pFileWave);
    return 0;
  }  
  (void)fread(pReadBuff, sizeof(WAV_ID), 1, pFileWave);
  
  seconds = (int)(pFileWave->SIZE - sizeof(WAV_ID)) / ((WAV_ID*)pReadBuff)->ByteRate;

  _mem_free(pReadBuff);
  fclose(pFileWave);
  
  return seconds;
}

//-----------------------------------------------------------------------------
// FUNCTION:    WAVE_Init
// SCOPE:       Wave Player private related function
// DESCRIPTION: Function initialize the Song decoder 
//              
// PARAMETERS:  none          
//              
// RETURNS:     none
//----------------------------------------------------------------------------
static void WAVE_Init(void)
{
  SAI_Init();
  InitCodec();
  //sgtl5000_Init();
  //Sgtl500_Set_Volume(75,75,100); // left, right, steps
      
#if WAVE_TASK_BUFFER_REC_MODE
  pRecordBuffer = (uint_16*)_mem_alloc(WAVE_TASK_BUFFER_REC_SIZE);
  if(pRecordBuffer == NULL)
  {
    printf("Cannot alloc the recording buffer. System is going down.");
    _mqx_exit(0);
  }
#endif
  PORTE_PCR9 = PORT_PCR_MUX(1) | PORT_PCR_DSE_MASK;
  GPIOE_PDDR = 1<<9;
  WAVE_ChangeState(WAVEST_READY); 
}

//-----------------------------------------------------------------------------
// FUNCTION:    WAVE_KeepPlaying
// SCOPE:       Wave Player private related function
// DESCRIPTION: Function fill up the uoput buffer with new audio stream data 
//              
// PARAMETERS:  pOutBuff - pointer on uotput buffer          
//              
// RETURNS:     none
//----------------------------------------------------------------------------
static void WAVE_KeepPlaying(short* pOutBuff)
{
  if (first_time)
  {
    // read the wave descriptor
    (void)fread(&wav_id, sizeof(WAV_ID), 1, pFileSong); // Read 44-bytes of wav header
  }
  
  nRead = fread(pOutBuff, 1, WAV_BUFFER_SIZE * 2, pFileSong); // Read the wave data
  
  if(nRead != (WAV_BUFFER_SIZE * 2))
    WAVE_Stop(pOutBuff);  

  if (first_time)
  {
    first_time = 0;
    
    if((SAI_SetSampleRate(wav_id.SampleRate) < 0) || (wav_id.NumChannels != 2))
    {
      WAVE_Stop(NULL);
      return;
    }  
      
    

    if(SAI_EnableTx(WAV_BUFFER_SIZE) < 0)
    {
      WAVE_Stop(NULL);
      return;
    }
    WAVE_ChangeState(WAVEST_PLAYING);
    
    printf("File Info\n---------\nBit rate: %d kbps\nSample rate: %d Hz\n"\
    , (wav_id.ByteRate/200), wav_id.SampleRate);
    printf("Size of file: %d B\t\n", pFileSong->SIZE);
    printf("Total time: %d s\t\t\n", (pFileSong->SIZE - sizeof(WAV_ID))/(wav_id.ByteRate));    
    printf("Channel count: %d \t\n", wav_id.NumChannels);
    printf("Bits per Sample: %d \t\n", wav_id.BitsPerSample);    
  }
  
  if(wave_state == WAVEST_PLAYING_HOLD)
      WAVE_Restore();
  
}

//-----------------------------------------------------------------------------
// FUNCTION:    WAVE_StoreRecord
// SCOPE:       Wave Player private related function
// DESCRIPTION: Function store piece of record to the file 
//              
// PARAMETERS:  pInBuff - pointer on input buffer          
//              
// RETURNS:     none
//----------------------------------------------------------------------------
static void WAVE_StoreRecord(short* pInBuff)
{
  GPIOE_PDOR |= 1<<9;
#if WAVE_TASK_BUFFER_REC_MODE

  
  memcpy(pRecordBufferWork, pInBuff, WAV_BUFFER_SIZE * 2);
  
  pRecordBufferWork += WAV_BUFFER_SIZE;
  
  // End of record buffer?
  if((pRecordBufferWork - pRecordBuffer) >= (WAVE_TASK_BUFFER_REC_SIZE-(WAV_BUFFER_SIZE * 2)))
    WAVE_Stop(NULL);
  
  
#else
  if(fwrite(pInBuff, 1, WAV_BUFFER_SIZE * 2, pFileSong) != WAV_BUFFER_SIZE * 2) // Write the wave data
  {
    printf("Error - Write File. Record Stops.\n\r");
    WAVE_Stop(NULL);
  }
#endif
  GPIOE_PDOR &= ~(1<<9);
}

//-----------------------------------------------------------------------------
// FUNCTION:    WAVE_RequestFillUpOutput
// SCOPE:       Wave Player private related function
// DESCRIPTION: Functions requests fill up all audio output buffers 
//              
// PARAMETERS:  none
//              
// RETURNS:     none
//----------------------------------------------------------------------------- 
static void WAVE_RequestFillUpOutput(void)
{
  WAVE_MSG message;
  uint_32 i,i_max = SAI_GetOutBuffCnt();
 
  message.type = WAVE_NEEDDATA;
  
  for(i = 0; i<i_max;i++)
  {
    message.pointer = SAI_GetOutBuff((unsigned char)i);
    _lwmsgq_send((pointer)wave_queue, (uint_32*)&message, 0);
  }

}
/******************************************************************************
 * Public CallBack functions definitions                                               *
 ******************************************************************************/

//-----------------------------------------------------------------------------
// FUNCTION:    SAI_FreeBufferCB
// SCOPE:       Wave Player call back from SAI peripheral related function
// DESCRIPTION: Functions notify the decoder that the next output buffer is ready to load
//              
// PARAMETERS:  ix - index of buffer
//              
// RETURNS:     none
//----------------------------------------------------------------------------- 
void SAI_FreeBufferCB(int ix)
{
  WAVE_MSG message;
  message.type = WAVE_NEEDDATA;
  message.pointer = SAI_GetOutBuff((unsigned char)ix);
  
  _lwmsgq_send((pointer)wave_queue, (uint_32*)&message, 0); 
}

//-----------------------------------------------------------------------------
// FUNCTION:    SAI_FullBufferCB
// SCOPE:       Wave Player call back from SAI peripheral related function
// DESCRIPTION: Functions notify the wave task that the RX buffer is full
//              
// PARAMETERS:  ix - index of buffer
//              
// RETURNS:     none
//----------------------------------------------------------------------------- 
void SAI_FullBufferCB(int ix)
{
  WAVE_MSG message;
  message.type = WAVE_HAVEDATA;
  message.pointer = SAI_GetOutBuff((unsigned char)ix);
  
  _lwmsgq_send((pointer)wave_queue, (uint_32*)&message, 0); 
}

/******************************************************************************
 * End of task_init module                                                    *
 ******************************************************************************/ 
