/***************************************************************************//*!
*
* @file      file_management.c
*
* @author    
* 
* @version   
* 
* @date      
* 
* @brief     File management 
*
******************************************************************************/

#include "file_management.h"
#include <mfs.h>
#include <fio.h>
#include <shell.h>
#include <mqx.h>

#define BUFSIZE   512


/** Extracts the drive name from absolute file path.
* @param[out] drive   Pointer to a preallocated memory where to store the name.
* @param[in]  absPath Absolute path to the file, e.g. "A:\abs\path\file.ext".
* @return             Zero on success, 1 for invalid argument, 2 for extraction
*                     failure. */
unsigned char MFS_getDrive(char *drive, const char *absPath) {
  if(!drive || !absPath) return 1;
  do {
    *drive++ = *absPath++;
  } while(*absPath && (*absPath!='\\'));
  *drive = 0;
  return (*absPath=='\\' ? 0 : 2);
}

/** Checks whether a file is a directory.
* @param[in]  absPath   Absolute path to the file, e.g. "A:\abs\path\file.ext".
* @return               Non-zero value for any directory, zero for other files. */
boolean MFS_isDir(char *absPath) {
  unsigned char attributes = 0;
  MFS_getFileAttrs(absPath, &attributes);
  return (attributes & MFS_ATTR_DIR_NAME);
}

/** Gets file attributes as bit array. See MFS_ATTR_??? macros.
* @param[in]  absPath   Absolute path to the file, e.g. "A:\abs\path\file.ext".
* @param[out] attrs     File attributes will be stored here.
* @return               0 on success, 1 for invalid arguments, 2 if ioctl fails. */
unsigned char MFS_getFileAttrs(char *absPath, unsigned char *attrs) {
  MQX_FILE_PTR        fs_ptr=NULL;
  MFS_FILE_ATTR_PARAM attr;
  uint_32             error_code;
  char                sDrive[FILENAME_SHORT_SIZE];
  
  if(!absPath || !attrs) return 1;

  attr.ATTRIBUTE_PTR  = attrs;
  attr.PATHNAME       = absPath;
  
  MFS_getDrive(sDrive, absPath);
  fs_ptr = _io_get_fs_by_name(sDrive);

  /* Get the attribute: */
  error_code = ioctl(fs_ptr, IO_IOCTL_GET_FILE_ATTR, (uint_32_ptr)&attr);
  if(error_code == MFS_NO_ERROR) return 0;
  else return 2;
}

/******************************************************************************/
boolean MFS_copy_files(char _PTR_ in_MFS, char _PTR_ in_dir, char _PTR_ in_file, 
                       char _PTR_ out_MFS, char _PTR_ out_dir,
                       char _PTR_ out_file)
{
  MQX_FILE_PTR in_fd=NULL, out_fd=NULL; 
  char _PTR_ buffer=NULL;
  int_32 size = 0, wsize = 0, copysize;
  boolean  temp;
  char_ptr in_abs_path, out_abs_path;
  
  if (MFS_alloc_path(&in_abs_path) != MFS_NO_ERROR) {
    printf("Error, unable to allocate memory for paths\n" );
    return FALSE;
  }
  
    if (MFS_alloc_path(&out_abs_path) != MFS_NO_ERROR) {
    printf("Error, unable to allocate memory for paths\n" );
    return FALSE;
  }
  
  _io_get_dev_for_path(in_abs_path,&temp,PATHNAME_SIZE, in_file, in_MFS);
  _io_rel2abs(in_abs_path, in_dir, in_file, PATHNAME_SIZE, in_MFS);
  _io_get_dev_for_path(out_abs_path,&temp,PATHNAME_SIZE, out_file, out_MFS);
  _io_rel2abs(out_abs_path, out_dir, out_file, PATHNAME_SIZE, out_MFS);
  
  if(MFS_isDir(in_abs_path)) {
    MFS_make_dir(out_MFS, out_dir, in_file);
    return TRUE;
    
  } else {
    //Copy files 
    
    out_fd = fopen(out_abs_path, "w");    
    if(out_fd==NULL) {
      return FALSE;
    }
    
    in_fd = fopen(in_abs_path, "r");    
    if(in_fd==NULL) {
      fclose(out_fd);
      return FALSE;
    }
    
    MFS_free_path(in_abs_path);
    MFS_free_path(out_abs_path);    
    
    buffer = _mem_alloc_system(in_fd->SIZE+1);    
    if(!buffer) {
      if(_task_get_error()!=MQX_OK) {
        _task_set_error(MQX_OK);
      }
      copysize = BUFSIZE;
      buffer = _mem_alloc(copysize+1);
    } else {
      copysize = in_fd->SIZE;
    }
    
    do {
      size = read(in_fd, buffer, copysize);
      if (size > 0) {
        wsize = write(out_fd, buffer, size);
      }
      else
        break;
    } while (wsize == size);
    
    if (buffer) {
      _mem_free(buffer);
    }
    if (in_fd) {
      fclose(in_fd);
    } 
    if (out_fd) {
      fclose(out_fd);
    }
    
    return TRUE; 
  }
}
  
boolean MFS_delete_file(char * file_system, char *dir, char * file_name)
{
  boolean  temp;
  int_32 error = 0;
  MQX_FILE_PTR fs_ptr;
  char_ptr abs_path;
  
  if (MFS_alloc_path(&abs_path) != MFS_NO_ERROR) {
    printf("Error, unable to allocate memory for paths\n" );
    return FALSE;
  }

  _io_get_dev_for_path(abs_path,&temp,PATHNAME_SIZE, file_name, file_system);
  fs_ptr = _io_get_fs_by_name(abs_path);
  if (fs_ptr == NULL) {
    printf("Error, file system not mounted\n" );
    return FALSE;
  } else { 
    error = _io_rel2abs(abs_path, dir, file_name, PATHNAME_SIZE, file_system);
    if(!error) {
      error = ioctl(fs_ptr, IO_IOCTL_DELETE_FILE, (pointer)abs_path);
    }
    MFS_free_path(abs_path);
  }
    
  return (error==0);
}

boolean MFS_make_dir(char * file_system, char *dir, char *dir_name)
{
  
   boolean temp;
   int_32 error = 0;
   MQX_FILE_PTR fs_ptr;
   char_ptr abs_path = NULL;
   

         if (MFS_alloc_path(&abs_path) != MFS_NO_ERROR) {
            printf("Error, unable to allocate memory for paths\n" );
            return FALSE;
         }
         else
         {
            _io_get_dev_for_path(abs_path,&temp,PATHNAME_SIZE,dir_name,file_system);
            fs_ptr = _io_get_fs_by_name(abs_path);
            /* check if filesystem is mounted */
            if (fs_ptr == NULL)  {
               printf("Error, file system not mounted\n" );
               return FALSE;
            } else  {
               error = _io_rel2abs(abs_path,dir,dir_name,PATHNAME_SIZE,file_system);
               if(!error)
               {
                  error = ioctl(fs_ptr, IO_IOCTL_CREATE_SUBDIR, (pointer) abs_path);
               }
               if (error)  {
                  printf("Error creating directory %s\n", dir_name);
                  return FALSE;
                
               }
            }
          MFS_free_path(abs_path);  
         }
    return TRUE;
}

void printf_dir(char * fs_name)
{
  MQX_FILE_PTR fs_ptr;
  char_ptr path_ptr, mode_ptr;
  pointer dir_ptr;
  char_ptr buffer;
  int_32 error;
  

    buffer = NULL;
    error = 0;
    path_ptr  ="*.*";
    mode_ptr = "m";
    
    fs_ptr=_io_get_fs_by_name(fs_name);
    /* check if filesystem is mounted */ 
    if (fs_ptr == NULL)  {
      printf("Error, file system not mounted\n");
    } else  {

      buffer = _mem_alloc(50);
      
      if (buffer && !error) {
        
        dir_ptr = _io_mfs_dir_open(fs_ptr, path_ptr, mode_ptr ); //important
        
        if (dir_ptr == NULL)  {
          printf("File not found.\n");
        } else {
          printf("\n");
          while ((_io_is_fs_valid(fs_ptr)) && (/*len = */_io_mfs_dir_read(dir_ptr, buffer, 50)) > 0) {
            printf(buffer);
          } 
          _io_mfs_dir_close(dir_ptr);
        }
        _mem_free(buffer);
      } else {
        if(buffer == NULL){
          printf("Error, unable to allocate space.\n" );
        } else {
          printf("Error, directory does not exist.\n" );
        }
      }
    }
}