/******************************************************************************
* 
* Copyright (c) 2012 Elnico s.r.o.
* All Rights Reserved                       
*
*******************************************************************************
*
* THIS SOFTWARE IS PROVIDED BY ELNICO "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL ELNICO OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************//*!
*
* @file      d4d_screen_filemngr.c
*
* @author    Petr Kubiznak
* 
* @version   1.0
* 
* @date      Sep-08-2012
* 
* @brief     Simple filemanager for eGui.
*
*******************************************************************************/

#include <d4d.h>
#include <lwmsgq.h>
#include <string.h>

#include "d4d_screen_filemngr.h"
#include "d4d_screen_audioplr.h"
#include "strings.h"
#include "fonts.h"
#include "../utils.h"
#include "../file_list.h"
#include "../file_management/file_management.h"
#include "../usb/USB_File.h"

/******************************************************************************
*
*                       local definitions
*
******************************************************************************/
#define DRIVES_CNT                      3

typedef struct panel_data {
  D4D_OBJECT  *pBtnDrive;
  D4D_OBJECT  *pLblDir;
  D4D_OBJECT  *pLstFiles;
  char        cDriveIndex;
  short       iMaxCnt;
  FILE_LIST   **ppFileLists;
} PANEL_DATA, *PANEL_DATA_PTR;

typedef struct drive_handle {
  MQX_FILE_PTR  pMFS_FD;
  D4D_CHAR      sIdentifier[FILENAME_SHORT_SIZE];
} DRIVE_HANDLE, *DRIVE_HANDLE_PTR;

/******************************************************************************
*
*                       locals variables declarations
*
******************************************************************************/

extern MQX_FILE_PTR filesystem_handle;
extern MQX_FILE_PTR ftp_filesystem_handle;
extern MQX_FILE_PTR usbdrive_filesystem_handle;
//extern MQX_FILE_PTR usb_fs_handle;

static DRIVE_HANDLE gpDriveHandles[DRIVES_CNT];

static D4D_TCHAR gsDriveL[4]            = D4D_DEFSTR("???");
static D4D_TCHAR gsDriveR[4]            = D4D_DEFSTR("???");
static D4D_TCHAR gsDirL[PATHNAME_SIZE]  = D4D_DEFSTR("???");
static D4D_TCHAR gsDirR[PATHNAME_SIZE]  = D4D_DEFSTR("???");

static PANEL_DATA gpPanelData[2];
static PANEL_DATA *gpPanelDataActive = NULL;

/****************************************************************************
*       Local functions declarations
***************************************************************************/
extern void ScrFileviewer_Show(const D4D_CHAR *fileName);

static void btnDrive_OnClick(D4D_OBJECT *pThis);
static void lstFiles_OnChange(D4D_OBJECT *pThis, D4D_EVENTID eventId);
static Byte lstFiles_OnObjectMsg(D4D_MESSAGE* pMsg);

static void btnOpen_OnClick(D4D_OBJECT* pThis);
static void btnCopy_OnClick(D4D_OBJECT* pThis);
static void btnRemove_OnClick(D4D_OBJECT* pThis);
static void btnExit_OnClick(D4D_OBJECT* pThis);

/** ****************************************************************************
* Refreshes the objects connected to given panel data.
* @param[in]  pPanelData  Initialized panel data.
* @return									0 on success, 1 for invalid argument, 2 for memory
*                         allocation failure.
*******************************************************************************/
static short panelRefresh(PANEL_DATA_PTR pPanelData);

/** ****************************************************************************
* Opens a file selected in the given panel data.
* @param[in]  pPanelData  Initialized panel data.
* @return									0 on success, 1 for invalid argument.
*******************************************************************************/
static short fileOpen(PANEL_DATA_PTR pPanelData);

/** ****************************************************************************
* Sets the active panel and focuses its listbox.
* @param[in]  pPanelData  Panel data to activate. If NULL, the active panel
*                         is not changed, only the listbox is focused.
*******************************************************************************/
static void focusPanel(PANEL_DATA_PTR pPanelData);

/** ****************************************************************************
* Redraws a frame arround the listbox connected to given panel data. Depending
* on the active panel (gpPanelDataActive), the frame is either drawn or erased.
* @param[in]  pPanelData  Panel data to to redraw its listbox. If NULL,
*                         frames of listboxes of both left and right panel are
*                         redrawn.
*******************************************************************************/
static void redrawPanelFrames(PANEL_DATA_PTR pPanelData);

/** ****************************************************************************
* Returns a currently used filelist.
* @param[in]  pPanelData  Initialized panel data.
* @return                 Pointer to the filelist or NULL on failure.
*******************************************************************************/
static FILE_LIST *getCurrentFileList(const PANEL_DATA_PTR pPanelData);

/** ****************************************************************************
* Returns a name of currently selected drive.
* @param[in]  pPanelData  Initialized panel data.
* @return                 Pointer to the drive name or NULL on failure.
*******************************************************************************/
static const D4D_CHAR *getCurrentDrive(const PANEL_DATA_PTR pPanelData);

/** ****************************************************************************
* Returns a name of currently selected file/directory.
* @param[in]  pPanelData  Initialized panel data.
* @return                 Pointer to the filename or NULL on failure.
*******************************************************************************/
static D4D_CHAR *getSelectedFilename(const PANEL_DATA_PTR pPanelData);

/** ****************************************************************************
* Filters out the "drive" part of the filepath.
* @param[in]  str String to locate the start of path in.
* @return         Pointer to the first letter of str not part of the drive
*                 specifier, i.e. pointer to the first '\' character. NULL
*                 for invalid argument or no backslash in the path.
*******************************************************************************/
static const D4D_CHAR *extractFilePath(const D4D_CHAR *str);

/******************************************************************************
*
*                       Graphic objects declarations
*
******************************************************************************/

#if ((D4D_SCREEN_SIZE_LONGER_SIDE >= 800) && (D4D_SCREEN_SIZE_SHORTER_SIDE >= 480))
# define  LCD_800X480                   1
#else
# define  LCD_800X480                   0
#endif

#define PADDING_X                       4
#define PADDING_Y                       2
#define RADIUS                          5

#if LCD_800X480
# define WIDGET_MINSIZE_X               100
# define WIDGET_MINSIZE_Y               50
#else /* LCD_800X480 */
# define WIDGET_MINSIZE_X               50
# define WIDGET_MINSIZE_Y               36
#endif /* LCD_800X480 */

#define SCREEN_MARGIN_X                 0
#define SCREEN_MARGIN_Y                 0
#define SCREEN_POS_X                    (SCREEN_MARGIN_X)
#define SCREEN_POS_Y                    (SCREEN_MARGIN_Y)
#define SCREEN_SIZE_X                   ((D4D_SCREEN_SIZE_LONGER_SIDE)-2*(SCREEN_MARGIN_X))
#define SCREEN_SIZE_Y                   ((D4D_SCREEN_SIZE_SHORTER_SIDE)-2*(SCREEN_MARGIN_Y))

#define PNLBOTTOM_SIZE_X                (SCREEN_SIZE_X)
#define PNLBOTTOM_SIZE_Y                (WIDGET_MINSIZE_Y + 2*(PADDING_Y))
#define PNLBOTTOM_POS_X                 0
#define PNLBOTTOM_POS_Y                 (SCREEN_SIZE_Y-PNLBOTTOM_SIZE_Y)

#define PANEL_SIZE_X                    ((SCREEN_SIZE_X)/2)
#define PANEL_SIZE_Y                    (SCREEN_SIZE_Y - PNLBOTTOM_SIZE_Y)
#define PANEL_POS_X(panel)              ((panel)*PANEL_SIZE_X)
#define PANEL_POS_Y                     0

#define DRIVE_SIZE_X                    (WIDGET_MINSIZE_X)
#define DRIVE_SIZE_Y                    (WIDGET_MINSIZE_Y)
#define DRIVE_POS_X                     (0)
#define DRIVE_POS_Y                     (0)

#define DIR_SIZE_X                      (PANEL_SIZE_X - DRIVE_SIZE_X - 3*PADDING_X)
#define DIR_SIZE_Y                      (WIDGET_MINSIZE_Y)
#define DIR_POS_X                       (DRIVE_POS_X + DRIVE_SIZE_X + PADDING_X)
#define DIR_POS_Y                       (0)

#define FILES_POS_X                     (0)
#define FILES_POS_Y                     (DRIVE_POS_Y + DRIVE_SIZE_Y + PADDING_Y)
#define FILES_SIZE_X                    (PANEL_SIZE_X - 2*PADDING_X)
#define FILES_SIZE_Y                    (PANEL_SIZE_Y - FILES_POS_Y - 2*PADDING_Y -16)
#define LISTBOX_ITEM_SIZE_Y             (WIDGET_MINSIZE_Y-2)
#define LISTBOX_ITEMS_PER_PAGE          ((FILES_SIZE_Y-4)/LISTBOX_ITEM_SIZE_Y)

#define COLUMN_SIZE_X                   (PNLBOTTOM_SIZE_X/4)

#define BUTTON_SIZE_X                   (COLUMN_SIZE_X - 2*PADDING_X)
#define BUTTON_SIZE_Y                   (WIDGET_MINSIZE_Y)
#define BUTTON_POS_X(col)               ((col)*COLUMN_SIZE_X + PADDING_X)
#define BUTTON_POS_Y                    (PNLBOTTOM_POS_Y + PADDING_Y)


#define FILES_LISTBOX_FLAGS (D4D_OBJECT_F_VISIBLE | D4D_OBJECT_F_ENABLED | D4D_OBJECT_F_TABSTOP | D4D_OBJECT_F_TOUCHENABLE)


/*******************************************
* Control buttons panel
*******************************************/
// Buttons
D4D_DECLARE_TXT_RBUTTON(sFilemngr_btnOpen,   DEF_EXTSTR(STR_OPEN),   BUTTON_POS_X(0), BUTTON_POS_Y, BUTTON_SIZE_X, BUTTON_SIZE_Y, RADIUS, FONT_SYSTEM, btnOpen_OnClick)
D4D_DECLARE_TXT_RBUTTON(sFilemngr_btnCopy,   DEF_EXTSTR(STR_COPY),   BUTTON_POS_X(1), BUTTON_POS_Y, BUTTON_SIZE_X, BUTTON_SIZE_Y, RADIUS, FONT_SYSTEM, btnCopy_OnClick)
D4D_DECLARE_TXT_RBUTTON(sFilemngr_btnRemove, DEF_EXTSTR(STR_REMOVE), BUTTON_POS_X(2), BUTTON_POS_Y, BUTTON_SIZE_X, BUTTON_SIZE_Y, RADIUS, FONT_SYSTEM, btnRemove_OnClick)
D4D_DECLARE_TXT_RBUTTON(sFilemngr_btnExit,   DEF_EXTSTR(STR_EXIT),   BUTTON_POS_X(3), BUTTON_POS_Y, BUTTON_SIZE_X, BUTTON_SIZE_Y, RADIUS, FONT_SYSTEM, btnExit_OnClick)


/*******************************************
* Left panel
*******************************************/
// Relations
D4D_EXTERN_OBJECT(sFilemngr_gpbLeft);
D4D_DECLARE_OBJECT_RELATIONS_BEGIN(gpbLeft_rel, &sFilemngr_gpbLeft)
D4D_DECLARE_OBJECT_RELATIONS_END

// Drive selection
//_D4D_DECLARE_BUTTON(type, name, text, x, y, cx, cy, radius, pMargin, pRelations, flags, pbmpN, pbmpF, pScheme, fontId, pUser, onclick, pOnUsrMsg)
_D4D_DECLARE_BUTTON(D4D_NO_CONST, sFilemngr_btnDriveL, gsDriveL, DRIVE_POS_X, DRIVE_POS_Y, DRIVE_SIZE_X, DRIVE_SIZE_Y, RADIUS, NULL, gpbLeft_rel, (D4D_BTN_F_DEFAULT), NULL, NULL, NULL, FONT_SYSTEM, &gpPanelData[0], btnDrive_OnClick, NULL)

//_D4D_DECLARE_LABEL(type, name, text, x, y, cx, cy, radius, pMargin, pRelations, flags, pScheme, fontId, pUser, pOnUsrMsg)
_D4D_DECLARE_LABEL(D4D_NO_CONST, sFilemngr_lblDirL, gsDirL, DIR_POS_X, DIR_POS_Y, DIR_SIZE_X, DIR_SIZE_Y, 0, NULL, gpbLeft_rel, (D4D_OBJECT_F_VISIBLE | D4D_OBJECT_F_ENABLED), NULL, FONT_SYSTEM_SMALL, &gpPanelData[0], NULL)

// Files listbox
//_D4D_DECLARE_LIST_BOX_BEGIN(type, name, x, y, cx, cy, radius, pMargin, pParent, flags, pScheme, itemsFontId, posCnt, listBoxItemsOff, pUser, pOnChange, pOnUsrMsg)
_D4D_DECLARE_LIST_BOX_BEGIN(D4D_NO_CONST, sFilemngr_lstFilesL, FILES_POS_X, FILES_POS_Y, FILES_SIZE_X, FILES_SIZE_Y, RADIUS, (D4D_MARGIN*)&listBox_marginDefault, &sFilemngr_gpbLeft, FILES_LISTBOX_FLAGS, NULL, FONT_SYSTEM, LISTBOX_ITEMS_PER_PAGE, LISTBOX_ITEM_SIZE_Y, &gpPanelData[0], lstFiles_OnChange, lstFiles_OnObjectMsg)
  //content will be filled automatically
D4D_DECLARE_LIST_BOX_END()

// Groupbox
D4D_DECLARE_STD_GROUP_BOX_BEGIN(sFilemngr_gpbLeft, PANEL_POS_X(0), PANEL_POS_Y, PANEL_SIZE_X, PANEL_SIZE_Y, RADIUS, NULL, DEF_EXTSTR(STR_LEFT), FONT_SYSTEM_SMALL)
  D4D_DECLARE_GROUP_BOX_OBJECT(sFilemngr_btnDriveL)
  D4D_DECLARE_GROUP_BOX_OBJECT(sFilemngr_lblDirL)
  D4D_DECLARE_GROUP_BOX_OBJECT(sFilemngr_lstFilesL)
D4D_DECLARE_GROUP_BOX_OBJECT_END


/*******************************************
* Right panel
*******************************************/
// Relations
D4D_EXTERN_OBJECT(sFilemngr_gpbRight);
D4D_DECLARE_OBJECT_RELATIONS_BEGIN(gpbRight_rel, &sFilemngr_gpbRight)
D4D_DECLARE_OBJECT_RELATIONS_END

// Drive selection
//_D4D_DECLARE_BUTTON(type, name, text, x, y, cx, cy, radius, pMargin, pRelations, flags, pbmpN, pbmpF, pScheme, fontId, pUser, onclick, pOnUsrMsg)
_D4D_DECLARE_BUTTON(D4D_NO_CONST, sFilemngr_btnDriveR, gsDriveR, DRIVE_POS_X, DRIVE_POS_Y, DRIVE_SIZE_X, DRIVE_SIZE_Y, RADIUS, NULL, gpbRight_rel, (D4D_BTN_F_DEFAULT), NULL, NULL, NULL, FONT_SYSTEM, &gpPanelData[1], btnDrive_OnClick, NULL)

//_D4D_DECLARE_LABEL(type, name, text, x, y, cx, cy, radius, pMargin, pRelations, flags, pScheme, fontId, pUser, pOnUsrMsg)
_D4D_DECLARE_LABEL(D4D_NO_CONST, sFilemngr_lblDirR, gsDirR, DIR_POS_X, DIR_POS_Y, DIR_SIZE_X, DIR_SIZE_Y, 0, NULL, gpbRight_rel, (D4D_OBJECT_F_VISIBLE | D4D_OBJECT_F_ENABLED), NULL, FONT_SYSTEM_SMALL, &gpPanelData[1], NULL)

// Files listbox
//_D4D_DECLARE_LIST_BOX_BEGIN(type, name, x, y, cx, cy, radius, pMargin, pParent, flags, pScheme, itemsFontId, posCnt, listBoxItemsOff, pUser, pOnChange, pOnUsrMsg)
_D4D_DECLARE_LIST_BOX_BEGIN(D4D_NO_CONST, sFilemngr_lstFilesR, FILES_POS_X, FILES_POS_Y, FILES_SIZE_X, FILES_SIZE_Y, RADIUS, (D4D_MARGIN*)&listBox_marginDefault, &sFilemngr_gpbRight, FILES_LISTBOX_FLAGS, NULL, FONT_SYSTEM, LISTBOX_ITEMS_PER_PAGE, LISTBOX_ITEM_SIZE_Y, &gpPanelData[1], lstFiles_OnChange, lstFiles_OnObjectMsg)
  //content will be filled automatically
D4D_DECLARE_LIST_BOX_END()

// Groupbox
D4D_DECLARE_STD_GROUP_BOX_BEGIN(sFilemngr_gpbRight, PANEL_POS_X(1), PANEL_POS_Y, PANEL_SIZE_X, PANEL_SIZE_Y, RADIUS, NULL, DEF_EXTSTR(STR_RIGHT), FONT_SYSTEM_SMALL)
  D4D_DECLARE_GROUP_BOX_OBJECT(sFilemngr_btnDriveR)
  D4D_DECLARE_GROUP_BOX_OBJECT(sFilemngr_lblDirR)
  D4D_DECLARE_GROUP_BOX_OBJECT(sFilemngr_lstFilesR)
D4D_DECLARE_GROUP_BOX_OBJECT_END


/*******************************************
* Screen
*******************************************/
D4D_DECLARE_SCREEN_BEGIN(screen_filemngr, ScreenFileMngr_, 0, 0, (D4D_COOR)(D4D_SCREEN_SIZE_LONGER_SIDE), (D4D_COOR)(D4D_SCREEN_SIZE_SHORTER_SIDE), NULL, 0, NULL, D4D_SCR_F_BCKG, NULL)
  // left panel
  D4D_DECLARE_SCREEN_OBJECT(sFilemngr_gpbLeft)
  // left panel
  D4D_DECLARE_SCREEN_OBJECT(sFilemngr_gpbRight)
  // control buttons
  D4D_DECLARE_SCREEN_OBJECT(sFilemngr_btnOpen)
  D4D_DECLARE_SCREEN_OBJECT(sFilemngr_btnCopy)
  D4D_DECLARE_SCREEN_OBJECT(sFilemngr_btnRemove)
  D4D_DECLARE_SCREEN_OBJECT(sFilemngr_btnExit)
D4D_DECLARE_SCREEN_END()    


/****************************************************************************
*       D4D standard screen functions definitions
****************************************************************************/

// One time called screen function in screen initialization proces
static void ScreenFileMngr_OnInit() {
  short int i,j;
  D4D_CHAR  sDrive[FILENAME_SHORT_SIZE];
  
  gpDriveHandles[0].pMFS_FD = filesystem_handle;
  strcpy(gpDriveHandles[0].sIdentifier, "A:");
  gpDriveHandles[1].pMFS_FD = ftp_filesystem_handle;
  strcpy(gpDriveHandles[1].sIdentifier, "F:");
  gpDriveHandles[2].pMFS_FD = usbdrive_filesystem_handle; //((USB_FILESYSTEM_STRUCT_PTR)usb_fs_handle)->FS_FD_PTR;
  strcpy(gpDriveHandles[2].sIdentifier, "B:");

  
  // left panel data initialization
  gpPanelData[0].pBtnDrive  = &sFilemngr_btnDriveL;
  gpPanelData[0].pLblDir    = &sFilemngr_lblDirL;
  gpPanelData[0].pLstFiles  = &sFilemngr_lstFilesL;
  
  // right panel data initialization
  gpPanelData[1].pBtnDrive  = &sFilemngr_btnDriveR;
  gpPanelData[1].pLblDir    = &sFilemngr_lblDirR;
  gpPanelData[1].pLstFiles  = &sFilemngr_lstFilesR;
  
  // ListFiles initialization for both left and right panel data
  for(i=0; i<2; ++i) {
    D4D_SetTextProperties(gpPanelData[i].pLblDir, (D4D_TXT_PRTY_ALIGN_H_LEFT_MASK | D4D_TXT_PRTY_ALIGN_V_CENTER_MASK));
    gpPanelData[i].cDriveIndex  = 0;
    gpPanelData[i].iMaxCnt      = 0;
    gpPanelData[i].ppFileLists  = (FILE_LIST**)D4D_MemAlloc(DRIVES_CNT * sizeof(FILE_LIST*));
    for(j=0; j<DRIVES_CNT; ++j) {
      D4D_StrCopy (sDrive, gpDriveHandles[j].sIdentifier);
      D4D_StrCat  (sDrive, "\\");
      gpPanelData[i].ppFileLists[j] = FileList_Install(gpDriveHandles[j].pMFS_FD, sDrive);
    }
    panelRefresh(&gpPanelData[i]);
  }
  
  // make left panel active
  focusPanel(&gpPanelData[0]);
}

// Screen on Activate function called with each screen activation
static void ScreenFileMngr_OnActivate() {
  focusPanel(NULL);                                         //select the proper listbox
}

// Screen "Main" function called periodically in each D4D_poll runs
static void ScreenFileMngr_OnMain() {
  /* nothing */
}


// Screen on DeActivate function called with each screen deactivation
static void ScreenFileMngr_OnDeactivate() {
  /* nothing */
}

// Screen on message function called with each internal massage for this screen
static Byte ScreenFileMngr_OnObjectMsg(D4D_MESSAGE* pMsg) {
  D4D_UNUSED(pMsg);  
  return D4D_MSG_NOSKIP;
}

/****************************************************************************
*       D4D widgets static events functions definitions
****************************************************************************/
// On click event of "Drive" button
static void btnDrive_OnClick(D4D_OBJECT *pThis) {
  PANEL_DATA_PTR pPanelData = (PANEL_DATA_PTR)pThis->userData;
  if(!pPanelData) return;
  
  // cyclicly increment the drive index
  if(++(pPanelData->cDriveIndex) >= DRIVES_CNT) pPanelData->cDriveIndex=0;
  // load current directory and refresh gui
  panelRefresh(pPanelData);
  
  // select the proper list
  focusPanel(pPanelData);
  redrawPanelFrames(NULL);
}

/******************************************************************************/
// On Change Event of "Files" listbox. We care of clicks only.
static void lstFiles_OnChange(D4D_OBJECT* pThis, D4D_EVENTID eventId) {
  // Just OnClick events are wanted
  if(eventId != D4D_EVENT_ONCLICK) return;
  
  // Try to open the selected file
  fileOpen(pThis->userData);
}

/******************************************************************************/
// Listbox OnMessage function called with each internal massage for this object
static Byte lstFiles_OnObjectMsg(D4D_MESSAGE* pMsg) {
  switch(pMsg->nMsgId) {
  case D4D_MSG_SETFOCUS:
    // sets current panel data to be the active one (gpPanelDataActive)
    focusPanel((PANEL_DATA_PTR)pMsg->pObject->userData);
    break;
    
  case D4D_MSG_DRAWDONE:
    // add/remove frame to the listbox
    redrawPanelFrames((PANEL_DATA_PTR)pMsg->pObject->userData);
    break;
  }
  return D4D_MSG_NOSKIP;
}
        
/******************************************************************************/
// On Click Event of "Open" button
static void btnOpen_OnClick(D4D_OBJECT* pThis) {
  fileOpen(gpPanelDataActive);
}

/******************************************************************************/
// On Click Event of "Copy" button
static void btnCopy_OnClick(D4D_OBJECT* pThis) {
  PANEL_DATA_PTR  pPanelDataPassive = (gpPanelDataActive == &gpPanelData[0] ?
                                       &gpPanelData[1] : &gpPanelData[0]);
  
  boolean ret = MFS_copy_files((char*)getCurrentDrive(gpPanelDataActive),       //src drive
                               (char*)extractFilePath(FileList_GetCurrentPath(getCurrentFileList(gpPanelDataActive))),  //src path
                               getSelectedFilename(gpPanelDataActive),          //src filename
                               (char*)getCurrentDrive(pPanelDataPassive),       //dst drive
                               (char*)extractFilePath(FileList_GetCurrentPath(getCurrentFileList(pPanelDataPassive))),  //dst path
                               getSelectedFilename(gpPanelDataActive));         //dst filename = src filename
  
  printf(ret ? "File copied successfully.\n" : "File copying failed!\n");
  
  // refresh the panels content
  panelRefresh(&gpPanelData[0]);
  panelRefresh(&gpPanelData[1]);
  focusPanel(NULL);                                         // switch the focus back to the active panel listbox
}

/******************************************************************************/
// On Click Event of "Remove" button
static void btnRemove_OnClick(D4D_OBJECT* pThis) {
  boolean ret = MFS_delete_file((char*)getCurrentDrive(gpPanelDataActive),
                  (char*)extractFilePath(FileList_GetCurrentPath(getCurrentFileList(gpPanelDataActive))),
                  getSelectedFilename(gpPanelDataActive));
  
  printf(ret ? "File deleted successfully.\n" : "File removal failed!\n");
  
  // refresh the panel content
  panelRefresh(gpPanelDataActive);
  focusPanel(NULL);                                         // switch the focus back to the active panel listbox
}

/******************************************************************************/
// On Click Event of "Exit" button
static void btnExit_OnClick(D4D_OBJECT* pThis) {
  D4D_EscapeScreen();
}

/** ****************************************************************************
* Refreshes the objects connected to given panel data.
* @param[in]  pPanelData  Initialized panel data.
* @return									0 on success, 1 for invalid argument, 2 for memory
*                         allocation failure.
*******************************************************************************/
static short panelRefresh(PANEL_DATA_PTR pPanelData) {
  LWord             i, dircnt, filescnt;
  D4D_WCHAR         buf[PATHNAME_SIZE];
  D4D_TCHAR         *pText;
  FILE_LIST         *pCurrentFileList = NULL;
  D4D_LIST_BOX_ITEM *pLstItems = NULL;

  if(!pPanelData) return 1;

  pCurrentFileList = pPanelData->ppFileLists[pPanelData->cDriveIndex];
  FileList_RefreshData(pCurrentFileList);

  // drive button text
  D4D_StrCopyAscii2Unicode(buf, getCurrentDrive(pPanelData));
  D4D_SetText(pPanelData->pBtnDrive,  buf);
  
  // directory label text
  D4D_StrCopyAscii2Unicode(buf, FileList_GetCurrentPath(pCurrentFileList));
  D4D_SetText(pPanelData->pLblDir,    buf);
  
  // directory content list
  dircnt    = FileList_GetDirCnt(pCurrentFileList);
  filescnt  = FileList_GetFilesCnt(pCurrentFileList);
  pLstItems = D4D_ListBoxGetItemList(pPanelData->pLstFiles);
  // -- allocate memory
  if(dircnt+filescnt > pPanelData->iMaxCnt) {
    pPanelData->iMaxCnt = dircnt+filescnt;
    D4D_MemFree((void*)pLstItems);
    //allocate all needed memory at once (array of structures and array of strings=filenames)
    pLstItems = (D4D_LIST_BOX_ITEM*)D4D_MemAlloc( (sizeof(D4D_LIST_BOX_ITEM) + (sizeof(D4D_TCHAR)*FILENAME_SHORT_SIZE)) * (pPanelData->iMaxCnt + 1) );
    if(!pLstItems) return 2;
  }
  pText = (D4D_TCHAR*)&(pLstItems[(dircnt+filescnt+1)]);
  // -- insert directories
  for(i=0; i<dircnt; ++i) {
    pLstItems[i].pUser = NULL;
    pLstItems[i].pText = pText;
    D4D_StrCopyAscii2Unicode(pText, FileList_GetDirName(pCurrentFileList, i));
    pText += FILENAME_SHORT_SIZE;
  }
  // -- insert files
  for(i=0; i<filescnt; ++i) {
    pLstItems[dircnt+i].pUser = NULL;
    pLstItems[dircnt+i].pText = pText;
    D4D_StrCopyAscii2Unicode(pText, FileList_GetFileName(pCurrentFileList, i));
    pText += FILENAME_SHORT_SIZE;
  }
  // -- terminate the list
  pLstItems[dircnt+i].pUser = NULL;
  pLstItems[dircnt+i].pText = NULL;
  // -- replace the list in the Listbox
  D4D_ListBoxChangeItemList(pPanelData->pLstFiles, pLstItems);

  return 0;
}

/** ****************************************************************************
* Opens a file selected in the given panel data.
* @param[in]  pPanelData  Initialized panel data.
* @return									0 on success, 1 for invalid argument.
*******************************************************************************/
static short fileOpen(PANEL_DATA_PTR pPanelData) {
  D4D_CHAR    pFilename[PATHNAME_SIZE] = "";
  D4D_CHAR    *pExt;
  FILE_LIST   *pCurrentFileList = NULL;
  int         ix = 0;
  int         iDirCnt = 0;

  if(!pPanelData) return 1;
  pCurrentFileList = pPanelData->ppFileLists[pPanelData->cDriveIndex];

  // Get the selected file
  ix = D4D_ListBoxGetIndex(pPanelData->pLstFiles);
  iDirCnt = FileList_GetDirCnt(pCurrentFileList);
  if(ix < iDirCnt) {                                        // Directory is selected -> change the directory
    FileList_ChangeDirectory(pCurrentFileList, ix);
    panelRefresh(pPanelData);
    focusPanel(NULL);

  } else {                                                  // Another file is selected -> open the file
    // Create the full path name
    D4D_StrCopy(pFilename, FileList_GetCurrentPath(pCurrentFileList));
    D4D_StrCat(pFilename, FileList_GetFileName(pCurrentFileList, ix-iDirCnt));

    // According to the file type...
    pExt = D4D_GetFileExtension((D4D_CHAR*)pFilename);

    // -- switch to the audio player
    if(!D4D_CompareStrings(pExt, "WAV")) {
      ScrAudioplr_Show(pFilename);
      
    // -- switch to the file viewer
    } else {
      ScrFileviewer_Show(pFilename);
    }
  }
  
  return 0;
}


/** ****************************************************************************
* Sets the active panel and focuses its listbox.
* @param[in]  pPanelData  Panel data to activate. If NULL, the active panel
*                         is not changed, only the listbox is focused.
*******************************************************************************/
static void focusPanel(PANEL_DATA_PTR pPanelData) {
  D4D_SCREEN *pScreen;
  
  if(pPanelData) gpPanelDataActive = pPanelData;
  
  pScreen = gpPanelDataActive->pLstFiles->pData->pScreen;
  if((D4D_OBJECT*)D4D_GetFocusedObject(pScreen) != gpPanelDataActive->pLstFiles) {
    D4D_FocusSet(pScreen, gpPanelDataActive->pLstFiles);
  }
}


/** ****************************************************************************
* Redraws a frame arround the listbox connected to given panel data. Depending
* on the active panel (gpPanelDataActive), the frame is either drawn or erased.
* @param[in]  pPanelData  Panel data to to redraw its listbox. If NULL,
*                         frames of listboxes of both left and right panel are
*                         redrawn.
*******************************************************************************/
static void redrawPanelFrames(PANEL_DATA_PTR pPanelData) {
  D4D_COLOR   color;
  D4D_POINT   pt;
  D4D_SIZE    sz;
  D4D_OBJECT  *pObject;
  char        i,low,high;
  
  low   = (pPanelData == &gpPanelData[1] ? 1 : 0);
  high  = (pPanelData == &gpPanelData[0] ? 0 : 1);

  // Runs once (i=0:0) for gpPanelData[0], once (i=1:1) for gpPanelData[1],
  // twice (i=0:1) for NULL, i.e. redraw both frames
  for(i=low; i<=high; ++i) {
    color = (&gpPanelData[i] == gpPanelDataActive) ?
      D4D_ScreenGetScheme((D4D_SCREEN*)&screen_filemngr)->fore :
      D4D_ScreenGetScheme((D4D_SCREEN*)&screen_filemngr)->bckg;
    pObject = gpPanelData[i].pLstFiles;
    pt  = D4D_GetClientToScreenPoint(pObject, &pObject->position);
    sz  = pObject->size;
    sz.cx--; sz.cy--;
    D4D_RRect(&pt, &sz, D4D_LINE_THICK, color, RADIUS);
  }
}


/** ****************************************************************************
* Returns a currently used filelist.
* @param[in]  pPanelData  Initialized panel data.
* @return                 Pointer to the filelist or NULL on failure.
*******************************************************************************/
static FILE_LIST *getCurrentFileList(const PANEL_DATA_PTR pPanelData) {
  if(!pPanelData) return NULL;
  else return pPanelData->ppFileLists[pPanelData->cDriveIndex];
}

/** ****************************************************************************
* Returns a name of currently selected drive.
* @param[in]  pPanelData  Initialized panel data.
* @return                 Pointer to the drive name or NULL on failure.
*******************************************************************************/
static const D4D_CHAR *getCurrentDrive(const PANEL_DATA_PTR pPanelData) {
  if(!pPanelData) return NULL;
  else return gpDriveHandles[pPanelData->cDriveIndex].sIdentifier;
}

/** ****************************************************************************
* Returns a name of currently selected file/directory.
* @param[in]  pPanelData  Initialized panel data.
* @return                 Pointer to the filename or NULL on failure.
*******************************************************************************/
static D4D_CHAR *getSelectedFilename(const PANEL_DATA_PTR pPanelData) {
  FILE_LIST *pFileList = getCurrentFileList(pPanelData);
  int ix, iDirCnt;
  
  if(!pFileList) return NULL;
  
  ix = D4D_ListBoxGetIndex(pPanelData->pLstFiles);
  iDirCnt = FileList_GetDirCnt(pFileList);
  if(ix < iDirCnt) return FileList_GetDirName(pFileList, ix);
  else return FileList_GetFileName(pFileList, ix-iDirCnt);
}

/** ****************************************************************************
* Filters out the "drive" part of the filepath.
* @param[in]  str String to locate the start of path in.
* @return         Pointer to the first letter of str not part of the drive
*                 specifier, i.e. pointer to the first '\' character. NULL
*                 for invalid argument or no backslash in the path.
*******************************************************************************/
static const D4D_CHAR *extractFilePath(const D4D_CHAR *str) {
  if(!str) return NULL;
  while(*str && *str!='\\') ++str;
  return (*str ? str : NULL);
}