/******************************************************************************
* 
* Copyright (c) 2012 Elnico s.r.o.
* All Rights Reserved                       
*
*******************************************************************************
*
* THIS SOFTWARE IS PROVIDED BY ELNICO "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL ELNICO OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************//*!
*
* @file      lwcfg.c
*
* @author    Petr Kubiznak
* 
* @version   1.0
* 
* @date      Aug-24-2012
* 
* @brief     Lightweight configuration files access library.
*
*******************************************************************************/


/*******************************************************************************
*       Includes header files
*******************************************************************************/
#include "lwcfg.h"
#include "utils.h"
#include "stdlib.h"

/*******************************************************************************
*       Functions declarations
*******************************************************************************/

/*******************************************************************************
* Initializes the record structure, i.e. allocates its memory and initializes
* the member variables to NULLs.
* @param[out]	record	  Structure variable to initialize.
* @param[in]  keyLen    Number of bytes to allocate for the 'key' variable.
* @param[in]  valueLen  Number of bytes to allocate for the 'value' variable.
*******************************************************************************/
static void initRecord(LWCFG_RECORD_PTR *record, unsigned short keyLen,
                       unsigned short valueLen);
/*******************************************************************************
* Frees the memory allocated for given record and its variables.
* @param[out]	record	  Structure variable to free. It is set to NULL after.
* @return               Pointer to the next record ( (*record)->next ).
*******************************************************************************/
static LWCFG_RECORD_PTR freeRecord(LWCFG_RECORD_PTR *record);
/*******************************************************************************
* Parses buffered text to the record structures.
* @param[out]	lwcfgFile		Initialized config file structure.
* @param[in]	buffer			Text to parse.
* @return									Number of records parsed. -1 in case of uninitialized
*													parameters.
*******************************************************************************/
static int parseText(LWCFG_FILE_PTR *lwcfgFile, D4D_CHAR *buffer);
/*******************************************************************************
* Stringifies the record structure and stores it to text buffer.
* @param[in]	lwcfgFile		Initialized config file structure.
* @param[out]	buffer			Pointer to UNinitialized memory. It is allocated in
*                         this function when needed memory space is known.
*                         User has to free the memory after use.
* @return									Number of output stored to the buffer,
*                         -1 for uninitialized input argument,
*                         -2 for invalid output argument. If a negative value
*                         is returned, buffer is not allocated, user should not
*                         try to free it.
*******************************************************************************/
static int stringifyData(LWCFG_FILE_PTR lwcfgFile, D4D_CHAR **buffer);
/*******************************************************************************
* Finds location of the record specified by given key.
* @param[in]	lwcfgFile		Initialized config file structure.
* @param[in]	key					Key to locate its record.
* @return									Address of pointer to the found record. If the record
*                         is not found (unused key), address of the ending
*                         record pointer in the chain is returned, i.e. address
*                         of 'next' variable of the last allocated record
*                         is returned. Since for this pointer next==NULL, this
*                         might be easily recognizable by check *ret==NULL.
*                         Hence three situations may occur:
*                         1. [ret && (*ret)]  -> record found successfully
*                         2. [ret && !(*ret)] -> record not found
*                         3. [!ret]           -> invalid argument
*******************************************************************************/
static LWCFG_RECORD_PTR *getRecordAddr(LWCFG_FILE_PTR lwcfgFile, D4D_CHAR *key);

/*******************************************************************************
******* Functions definitions **************************************************
*******************************************************************************/

/*******************************************************************************
* Opens a config file and initializes the config file structure. This needs
* to be called before any other function from this library.
* @param[in]	pFileName		Path to the config file in the filesystem.
* @return									Config file structure on success, NULL on fail.
*******************************************************************************/
LWCFG_FILE_PTR lwcfg_open(D4D_CHAR* pFileName) {
  LWCFG_FILE_PTR  lwcfgFile = NULL;
  D4D_FILEPTR     pFile = NULL;
  LWord           size = 0;
	D4D_CHAR        *pBuf, *pBufTmp;

  // Initialize the structure
  lwcfgFile               = (LWCFG_FILE_PTR)D4D_MemAlloc(sizeof(LWCFG_FILE));
	lwcfgFile->firstRecord  = NULL;
	lwcfgFile->bChanged	    = 0;
  size                    = D4D_StrLen(pFileName);
  lwcfgFile->fileName     = (D4D_CHAR*)D4D_MemAlloc(size+1);
  memcpy(lwcfgFile->fileName, pFileName, size+1);

	// Open the config file
	pFile = D4D_FileOpen(lwcfgFile->fileName, "r");
	
	// File cannot be opened, return default (empty) config structure
  if(!pFile) return lwcfgFile;
	
	// Config files must be smaller then 10kB
  D4D_FileSeek(pFile, 0, D4D_SEEK_END);
  size = D4D_FileTell(pFile);
	if(size>10240) {
		D4D_FileClose(pFile);
    lwcfg_close(lwcfgFile);
		return NULL;
	}

	// Copy the file to buffer
	D4D_FileSeek(pFile, 0, D4D_SEEK_SET);    
	pBuf = D4D_MemAlloc(size + 2); 
	pBufTmp = pBuf;
	do {
		Byte buf512[512];
		LWord j;

		size = D4D_FileRead(pFile, buf512, 512);
		for(j = 0; j<size; j++)
			*(pBufTmp++) = buf512[j]; 
	} while(size==512);
	*((D4D_CHAR*)pBufTmp) = 0; // End char
	
	// Parse the buffer to the record structures
	parseText(&lwcfgFile, pBuf);

	// Close the file
	D4D_MemFree(pBuf);
  D4D_FileClose(pFile);
  return lwcfgFile;
}

/*******************************************************************************
* Closes a config file and destroys the config file structure.
* @param[in,out]	lwcfgFile		Config file structure to destroy.
* @return                     Zero on success, 1 for uninitialized argument,
*                             2 on stringification failure, 3 on file open
*                             failure, 4 on file write failure.
*******************************************************************************/
char lwcfg_close(LWCFG_FILE_PTR lwcfgFile) {
  LWCFG_RECORD_PTR *ppRecord = NULL;
  D4D_FILEPTR       pFile = NULL;
  D4D_CHAR          *buf = NULL;
  int               bufLen = 0;
  char              ret = 0;
  
	if(!lwcfgFile) return 1;
  
  // Write the changes to the config file on the filesystem
  if(lwcfgFile->bChanged) {
    if((bufLen=stringifyData(lwcfgFile, &buf))<=0) {
      ret=2;
    } else {
      pFile = D4D_FileOpen(lwcfgFile->fileName, "w");
      if(!pFile) {
        ret=3;
      } else {
        if(D4D_FileWrite(pFile, buf, bufLen) != bufLen) ret=4;
        D4D_FileClose(pFile);
      }
      D4D_MemFree(buf);
    }
  }
	
  // Free the memory
  ppRecord = &(lwcfgFile->firstRecord);
  while(*ppRecord) *ppRecord = freeRecord(ppRecord);
  D4D_MemFree(lwcfgFile->fileName);
	D4D_MemFree(lwcfgFile);
  return ret;
}

/*******************************************************************************
* Parses buffered text to the record structures.
* @param[out]	lwcfgFile		Initialized config file structure.
* @param[in]	buffer			Text to parse.
* @return									Number of records parsed. -1 in case of uninitialized
*													parameters.
*******************************************************************************/
static int parseText(LWCFG_FILE_PTR *lwcfgFile, D4D_CHAR *buffer) {
	unsigned short 	  keyLen=0, valueLen=0, cnt=0;
	D4D_CHAR				  *keyStart=NULL, *valueStart=NULL, *pos=NULL;
  LWCFG_RECORD_PTR  *newRecord=NULL;
	
	if(!lwcfgFile || !buffer) return -1;
	newRecord = &((*lwcfgFile)->firstRecord);
  
  pos = buffer;
	while(*pos) {
    //shift to the new non-empty line
    while(*pos && ((*pos=='\r') || (*pos=='\n'))) ++pos;
    keyLen = valueLen = 0;
    
    //shift to the '='
    keyStart = pos;
		while(*pos && (*pos!='=') && (*pos!='\r') && (*pos!='\n')) { ++pos; ++keyLen; }
		if(*pos!='=') continue;                                                     //syntax error on current line
    
    //shift to the end of line
    valueStart = ++pos;
		while(*pos && (*pos!='\r') && (*pos!='\n')) { ++pos; ++valueLen; }
    
    //create a new record
    initRecord(newRecord, keyLen+1, valueLen+1);
    memcpy((*newRecord)->key,    keyStart,   keyLen);    (*newRecord)->key[keyLen]='\0';
    memcpy((*newRecord)->value,  valueStart, valueLen);  (*newRecord)->value[valueLen]='\0';
    newRecord = &((*newRecord)->next);
    ++cnt;
	}
  return cnt;
}

/*******************************************************************************
* Stringifies the record structure and stores it to text buffer.
* @param[in]	lwcfgFile		Initialized config file structure.
* @param[out]	buffer			Pointer to UNinitialized memory. It is allocated in
*                         this function when needed memory space is known.
*                         User has to free the memory after use.
* @return									Number of output stored to the buffer,
*                         -1 for uninitialized input argument,
*                         -2 for invalid output argument. If a negative value
*                         is returned, buffer is not allocated, user should not
*                         try to free it.
*******************************************************************************/
static int stringifyData(LWCFG_FILE_PTR lwcfgFile, D4D_CHAR **buffer) {
  LWCFG_RECORD_PTR  pRecord = NULL;
  D4D_CHAR          *pos = NULL;
  int               size=0, len=0;
  
  if(!lwcfgFile)  return -1;
  if(!buffer)     return -2;
  
  // Firstly compute the needed size to alloc
  len = 20;                                                                     //"#lwcfg config file\r\n"
  pRecord = lwcfgFile->firstRecord;
  while(pRecord) {
    len += D4D_StrLen(pRecord->key);
    len += D4D_StrLen(pRecord->value);
    len += 3;                                                                   //'=' and "\r\n"
    pRecord = pRecord->next;
  }
  len += 1;                                                                     //'\0'
  
  // Allocate the memory
  size    = len;
  *buffer = (D4D_CHAR*)D4D_MemAlloc(size);
  
  // Initial comment
  pos = *buffer;
  len = 20;
  memcpy(pos, "#lwcfg config file\r\n", len);
  pos += len;
  
  // Copy the data from the record structure to the buffer
  pRecord = lwcfgFile->firstRecord;
  while(pRecord) {
    //key
    len = D4D_StrLen(pRecord->key);
    memcpy(pos, pRecord->key, len);
    pos += len;
    *(pos++) = '=';
    //value
    len = D4D_StrLen(pRecord->value);
    memcpy(pos, pRecord->value, len);
    pos += len;
    *(pos++) = '\r';
    *(pos++) = '\n';
    
    pRecord = pRecord->next;
  }
  *pos = '\0';                                                                  //termination character
  return size;
}

/*******************************************************************************
* Initializes the record structure, i.e. allocates its memory and initializes
* the member variables to NULLs.
* @param[out]	record	  Structure variable to initialize.
* @param[in]  keyLen    Number of bytes to allocate for the 'key' variable.
* @param[in]  valueLen  Number of bytes to allocate for the 'value' variable.
*******************************************************************************/
static void initRecord(LWCFG_RECORD_PTR *record, unsigned short keyLen,
                       unsigned short valueLen) {
	*record           = (LWCFG_RECORD_PTR)D4D_MemAlloc(sizeof(LWCFG_RECORD));
	(*record)->key    = (D4D_CHAR*)D4D_MemAlloc(keyLen);
  (*record)->value  = (D4D_CHAR*)D4D_MemAlloc(valueLen);
  (*record)->next   = NULL;
}

/*******************************************************************************
* Frees the memory allocated for given record and its variables.
* @param[out]	record	  Structure variable to free. It is set to NULL after.
* @return               Pointer to the next record ( (*record)->next ).
*******************************************************************************/
static LWCFG_RECORD_PTR freeRecord(LWCFG_RECORD_PTR *record) {
  LWCFG_RECORD_PTR next = NULL;
  
  if(!record || !*record) return NULL;
  next = (*record)->next;
	D4D_MemFree((*record)->key);
	D4D_MemFree((*record)->value);
  D4D_MemFree(*record);
  *record = NULL;
  return next;
}

/*******************************************************************************
* Finds location of the record specified by given key.
* @param[in]	lwcfgFile		Initialized config file structure.
* @param[in]	key					Key to locate its record.
* @return									Address of pointer to the found record. If the record
*                         is not found (unused key), address of the ending
*                         record pointer in the chain is returned, i.e. address
*                         of 'next' variable of the last allocated record
*                         is returned. Since for this pointer next==NULL, this
*                         might be easily recognizable by check *ret==NULL.
*                         Hence three situations may occur:
*                         1. [ret && (*ret)]  -> record found successfully
*                         2. [ret && !(*ret)] -> record not found
*                         3. [!ret]           -> invalid argument
*******************************************************************************/
static LWCFG_RECORD_PTR *getRecordAddr(LWCFG_FILE_PTR lwcfgFile, D4D_CHAR *key) {
  LWCFG_RECORD_PTR  *ppRecord = NULL;
  D4D_CHAR          *key1   = NULL;
  D4D_CHAR          *key2   = NULL;
  
	if(!lwcfgFile || !key) return NULL;
  
  ppRecord = &(lwcfgFile->firstRecord);
  while(*ppRecord) {
    key1 = key;
    key2 = (*ppRecord)->key;
    
    while((*key1==*key2) && *key1) { ++key1; ++key2; }
    if((*key1==*key2) && !(*key1)) break;
    
    ppRecord = &((*ppRecord)->next);
  }
  return ppRecord;
}

/*******************************************************************************
* Reads the string value given by the key.
* @param[in]	lwcfgFile		Initialized config file structure.
* @param[in]	key					Key to locate its value in the config records.
* @param[out]	value				Address to store the string value to. The memory space
*													has to be allocated before the call.
* @param[in]	maxLen			Maximal length of data to copy (including '\0').
* @return									If the key is found, its value is stored to the
*													address specified by the parameter "value" and 0
*													is returned.
*													1 is returned in case of uninitialized input parameters.
*													2 is returned in case of uninitialized output parameter.
*													3 is returned if given key is not found in the config
*													file. Parameter value is not modified in this case.
*******************************************************************************/
char lwcfg_getString(LWCFG_FILE_PTR lwcfgFile, D4D_CHAR *key, D4D_CHAR *value,
                     unsigned short maxLen) {

  LWCFG_RECORD_PTR *ppRecord = getRecordAddr(lwcfgFile, key);
  D4D_CHAR *pos = NULL;
  short int i=0;
  
  if(!ppRecord)   return 1;
  if(!value)      return 2;
  if(!*ppRecord)  return 3;
  
  // copy the string
  pos = (*ppRecord)->value;
	for(i=0; i<maxLen-1; ++i) {
		if(!*pos) break;
		*(value++) = *(pos++);
	}
	*value = '\0';																															  //termination character
  return 0;
}

/*******************************************************************************
* Reads the integer value given by the key.
* @param[in]	lwcfgFile		Initialized config file structure.
* @param[in]	key					Key to locate its value in the config records.
* @param[out]	value				Address to store the integer value to.
* @return									If the key is found, its value is stored to the
*													address specified by the parameter "value" and 0
*													is returned. This pays also for a value which cannot
*													be converted to any integer, value is set to zero then.
*													1 is returned in case of uninitialized input parameters.
*													2 is returned in case of uninitialized output parameter.
*													3 is returned if given key is not found in the config
*													file. Parameter value is not modified in this case.
*													4 is returned if the text cannot be converted.
*******************************************************************************/
char lwcfg_getInteger(LWCFG_FILE_PTR lwcfgFile, D4D_CHAR *key, long int *value) {
	D4D_CHAR buf[12];
	int ret = lwcfg_getString(lwcfgFile, key, buf, 12);
	D4D_CHAR *endptr = buf;
	
	if(ret) return ret;
	if(!value) return 2;
	*value = strtol(buf, &endptr, 10);
	return (*endptr=='\0' ? 0 : 4);
}

/*******************************************************************************
* Assigns a string value to the given key. The data is modified only in memory,
* config file structure needs to be closed first by calling lwcfg_close()
* to write the changes to the config file on the filesystem.
* @param[in,out]	lwcfgFile		Initialized config file structure.
* @param[in]	    key					Key to be assigned a value for.
* @param[in]	    value				Value to assign to the key.
* @return                     Zero on success, 1 on invalid argument, 2 on other
*                             failure.
*******************************************************************************/
char lwcfg_setString(LWCFG_FILE_PTR lwcfgFile, D4D_CHAR *key, D4D_CHAR *value) {
  LWCFG_RECORD_PTR *ppRecord = getRecordAddr(lwcfgFile, key);
  int keyLen=0, valueLen=0;
  
  if(!ppRecord || !value)   return 1;

  keyLen    = D4D_StrLen(key);
  valueLen  = D4D_StrLen(value);
  
  // record exists -> update value
  if(*ppRecord) {
    if(D4D_StrLen((*ppRecord)->value) < valueLen) {                             //realloc memory
      D4D_MemFree((*ppRecord)->value);
      (*ppRecord)->value = (D4D_CHAR*)D4D_MemAlloc(valueLen+1);
      if(!(*ppRecord)->value) return 2;
    }
    memcpy((*ppRecord)->value, value, valueLen);
    (*ppRecord)->value[valueLen] = '\0';
  
  // record does not exist -> create new one  
  } else {
    initRecord(ppRecord, keyLen+1, valueLen+1);
    memcpy((*ppRecord)->key,    key,    keyLen);    (*ppRecord)->key[keyLen]='\0';
    memcpy((*ppRecord)->value,  value,  valueLen);  (*ppRecord)->value[valueLen]='\0';
  }
  
  lwcfgFile->bChanged = 1;
  return 0;
}

/*******************************************************************************
* Assigns an integer value to the given key.
* @param[in,out]  lwcfgFile   Initialized config file structure.
* @param[in]      key         Key to be assigned a value for.
* @param[in]      value       Value to be assigned to the key.
* @return                     Zero on success, 1 on invalid argument, 2 on other
*                             failure.
*******************************************************************************/
char lwcfg_setInteger(LWCFG_FILE_PTR lwcfgFile, D4D_CHAR *key, long int value) {
  D4D_CHAR  pBuf[12];
  D4D_CHAR  *pos=pBuf;
  long int  divider = 1000000000;
  char      digit = 0;
  
  if(!lwcfgFile || !key) return 1;
  
  //"itoa(value, buf, 10);"
  if(value<0) {
    *(pos++)='-';
    divider *= -1;
  }
  while(divider) {
    digit = (char)(value/divider);
    *(pos++)='0'+digit;
    value %= divider;
    divider /= 10;
  }
  *pos = '\0';
  
  return lwcfg_setString(lwcfgFile, key, pBuf);
}


