/******************************************************************************
* 
* Copyright (c) 2012 Elnico s.r.o.
* All Rights Reserved                       
*
*******************************************************************************
*
* THIS SOFTWARE IS PROVIDED BY ELNICO "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL ELNICO OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************//*!
*
* @file      lwcfg.h
*
* @author    Petr Kubiznak
* 
* @version   1.0
* 
* @date      Aug-24-2012
* 
* @brief     Lightweight configuration files access library.
*
*******************************************************************************/

#ifndef __LWCFG_H__940189647031348705043110
#define __LWCFG_H__940189647031348705043110

/*******************************************************************************
*       Includes header files
*******************************************************************************/
#include "d4d.h"

/*******************************************************************************
*       Type definitions
*******************************************************************************/
typedef struct lwcfg_record {
  D4D_CHAR						*key;
	D4D_CHAR						*value;
	struct lwcfg_record	*next;
} LWCFG_RECORD, *LWCFG_RECORD_PTR;

typedef struct lwcfg_file {
  D4D_CHAR          *fileName;
	LWCFG_RECORD_PTR	firstRecord;
	char							bChanged;
} LWCFG_FILE, *LWCFG_FILE_PTR;


/*******************************************************************************
*       Functions declarations
*******************************************************************************/

/*******************************************************************************
* Opens a config file and initializes the config file structure. This needs
* to be called before any other function from this library.
* @param[in]	pFileName		Path to the config file in the filesystem.
* @return									Config file structure on success, NULL on fail.
*******************************************************************************/
LWCFG_FILE_PTR	lwcfg_open	(D4D_CHAR* pFileName);

/*******************************************************************************
* Closes a config file and destroys the config file structure.
* @param[in,out]	lwcfgFile		Config file structure to destroy.
* @return                     Zero on success, 1 for uninitialized argument,
*                             2 on stringification failure, 3 on file open
*                             failure, 4 on file write failure.
*******************************************************************************/
char lwcfg_close(LWCFG_FILE_PTR lwcfgFile);

/*******************************************************************************
* Reads the string value given by the key.
* @param[in]	lwcfgFile		Initialized config file structure.
* @param[in]	key					Key to locate its value in the config records.
* @param[out]	value				Address to store the string value to. The memory space
*													has to be allocated before the call.
* @param[in]	maxLen			Maximal length of data to copy.
* @return									If the key is found, its value is stored to the
*													address specified by the parameter "value" and 0
*													is returned.
*													1 is returned in case of uninitialized input parameters.
*													2 is returned in case of uninitialized output parameter.
*													3 is returned if given key is not found in the config
*													file. Parameter value is not modified in this case.
*******************************************************************************/
char lwcfg_getString(LWCFG_FILE_PTR lwcfgFile, D4D_CHAR *key, D4D_CHAR *value,
                     unsigned short maxLen);

/*******************************************************************************
* Reads the integer value given by the key.
* @param[in]	lwcfgFile		Initialized config file structure.
* @param[in]	key					Key to locate its value in the config records.
* @param[out]	value				Address to store the integer value to.
* @return									If the key is found, its value is stored to the
*													address specified by the parameter "value" and 0
*													is returned. This pays also for a value which cannot
*													be converted to any integer, value is set to zero then.
*													1 is returned in case of uninitialized input parameters.
*													2 is returned in case of uninitialized output parameter.
*													3 is returned if given key is not found in the config
*													file. Parameter value is not modified in this case.
*													4 is returned if the text cannot be converted.
*******************************************************************************/
char lwcfg_getInteger(LWCFG_FILE_PTR lwcfgFile, D4D_CHAR *key, long int *value);

/*******************************************************************************
* Assigns a string value to the given key. The data is modified only in memory,
* config file structure needs to be closed first by calling lwcfg_close()
* to write the changes to the config file on the filesystem.
* @param[in,out]	lwcfgFile		Initialized config file structure.
* @param[in]	    key					Key to be assigned a value for.
* @param[in]	    value				Value to assign to the key.
* @return                     Zero on success, 1 on invalid argument, 2 on other
*                             failure.
*******************************************************************************/
char lwcfg_setString(LWCFG_FILE_PTR lwcfgFile, D4D_CHAR *key, D4D_CHAR *value);

/*******************************************************************************
* Assigns an integer value to the given key.
* @param[in,out]  lwcfgFile   Initialized config file structure.
* @param[in]      key         Key to be assigned a value for.
* @param[in]      value       Value to be assigned to the key.
* @return                     Zero on success, 1 on invalid argument, 2 on other
*                             failure.
*******************************************************************************/
char lwcfg_setInteger(LWCFG_FILE_PTR lwcfgFile, D4D_CHAR *key, long int value);

/******************************************************************************/
#endif /* __LWCFG_H__940189647031348705043110 */
