/******************************************************************************
* 
* Copyright (c) 2012 Elnico s.r.o.
* All Rights Reserved                       
*
*******************************************************************************
*
* THIS SOFTWARE IS PROVIDED BY ELNICO "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL ELNICO OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************//*!
*
* @file      utils.c
*
* @author    Petr Kubiznak
* 
* @version   1.0
* 
* @date      Sep-04-2012
* 
* @brief     Utility functions and macro definitions.
*
*******************************************************************************/

/*******************************************************************************
*       Includes header files
*******************************************************************************/
#include "utils.h"

/*******************************************************************************
*       Function definitions
*******************************************************************************/

/*******************************************************************************
* Reads data from a file to the memory.
* @param[out] dst         Destination pointer. If NULL, function allocates new
*                         memory (of exactly the needed size). Else it assumes
*                         there is enough space allocated to read the required
*                         length of data.
* @param[in]  sFilename   Path to the file to read.
* @param[in]  maxlen      Maximum number of bytes to read. If 0, function reads
*                         the whole file.
* @return                 Number of bytes read or 0 in case of failure.
*******************************************************************************/
LWord readFile(void **dst, D4D_CHAR *sFilename, unsigned int maxlen) {
  D4D_FILEPTR     pFile = NULL;
  LWord           size = 0;
  D4D_CHAR        *pBuf = NULL;

  if(!dst || !sFilename) return 0;
  
  // Open the file
  pFile = D4D_FileOpen(sFilename, "r");
  if(!pFile) return 0;
	
  // Count the length
  D4D_FileSeek(pFile, 0, D4D_SEEK_END);
  size = D4D_FileTell(pFile);
  if(maxlen && (size>maxlen)) size=maxlen;

  // Copy the file to buffer
  D4D_FileSeek(pFile, 0, D4D_SEEK_SET);
  pBuf = *dst;
  if(!pBuf) pBuf = D4D_MemAlloc(size + 4); 
  size = D4D_FileRead(pFile, pBuf, size);
	
  // Close the file
  D4D_FileClose(pFile);
  
  *dst = pBuf;
  return size;
}

/*******************************************************************************
* Loads a bitmap from a file to RAM.
* @param[out] dst         Destination pointer. If dst->pData is NULL, function
*                         allocates new memory (of exactly the needed size).
*                         Else it assumes there is enough space allocated
*                         to read the whole bitmap.
* @param[in]  sFilename   Path to the file to read.
* @return                 0 on success, 1 on invalid argument, 2 on file read
*                         error.
*******************************************************************************/
Byte loadBitmap(D4D_BMP *dst, D4D_CHAR *sFilename) {
  char *pData = NULL;
  LWord ret;
  
  if(!dst || !sFilename) return 1;
  
  ret = readFile((void**)(&pData), sFilename, 0);
  if(!ret) return 2;
  
  // there are 32 bytes of file header, only 6 bytes have to be included in the
  // bitmap memory (W,H,mode), right before the real data; 26 bytes is a mess
  memcpy(pData+26, pData+8, 6);
  pData += 26;
  dst->pData = (void*)pData;
  
  return 0;
}

/*******************************************************************************
* Frees the memory allocated for a bitmap by the loadBitmap() call.
* @param[out] bmp         Bitmap to free its pData member variable.
*******************************************************************************/
void freeBitmap(D4D_BMP *bmp) {
  char *pData = NULL;
  if(!bmp) return;
  
  // we need to shift back by the 26 bytes we skiped after allocation
  pData = (char*)bmp->pData;
  pData -= 26;
  D4D_MemFree(pData);
  
  bmp->pData = NULL;
}
